/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QList>
#include <QDate>

#include <assert.h>

#include "ucbudgetdata.h"

TUCBudgetData::TUCBudgetData(void)
{
}

TUCBudgetData::~TUCBudgetData(void)
{
}

QString TUCBudgetData::Component_Name(
	const int							index) const
{
	assert(static_cast<unsigned int>(index) < d_components.size());
	
	return d_components[index].name;
}

TUCBudgetData::TComponent TUCBudgetData::Component(
	const int							index) const
{
	assert(static_cast<unsigned int>(index) < d_components.size());
	
	return d_components[index];
}

int TUCBudgetData::Component_Index(
	const QString						&name) const
{
	std::vector<TUCBudgetData::TComponent>::const_iterator iter;
	
	for(iter = d_components.begin();iter != d_components.end();++iter)
	{
		if(name == (*iter).name)
		{
			return static_cast<int>(iter - d_components.begin());
		}
	}
	
	return -1;
}

int TUCBudgetData::Component_Item_Count(
	const int							component_id) const
{
	assert(static_cast<unsigned int>(component_id) < d_components.size());
	
	return static_cast<int>(d_components[component_id].items.size());
}

QString TUCBudgetData::Component_Item_Name(
	const int							component_id,
	const int							index) const
{
	assert(static_cast<unsigned int>(component_id) < d_components.size());
	assert(static_cast<unsigned int>(index) < d_components[component_id].items.size());
	
	return d_components[component_id].items[index].name;
}

TUCBudgetData::TComponentItem TUCBudgetData::Component_Item(
	const int							component_id,
	const int							index) const
{
	assert(static_cast<unsigned int>(component_id) < d_components.size());
	assert(static_cast<unsigned int>(index) < d_components[component_id].items.size());
	
	return d_components[component_id].items[index];
}

int TUCBudgetData::Component_Item_Index(
	const int							component_id,
	const QString						&name) const
{
	std::vector<TUCBudgetData::TComponentItem>::const_iterator iter;

	assert(static_cast<unsigned int>(component_id) < d_components.size());
	
	for(iter = d_components[component_id].items.begin();iter != d_components[component_id].items.end();++iter)
	{
		if(name == (*iter).name)
		{
			return static_cast<int>(iter - d_components[component_id].items.begin());
		}
	}
	
	return -1;

}

TUCBudgetData::TComponentItem TUCBudgetData::Component_Item(
	const QString						&component_id,
	const QString						&component_item_id) const
{
	int									component_index;
	int									component_item_index;
	
	component_index = this->Component_Index(component_id);
	component_item_index = this->Component_Item_Index(component_index,component_item_id);
	
	return this->Component_Item(component_index,component_item_index);
}

int TUCBudgetData::Component_Identifier_Count(
	const int							component_id) const
{
	assert(static_cast<unsigned int>(component_id) < d_components.size());
	
	return static_cast<int>(d_components[component_id].instances.size());
}

QString TUCBudgetData::Component_Identifier_Name(
	const int							component_id,
	const int							index) const
{
	assert(static_cast<unsigned int>(component_id) < d_components.size());
	assert(static_cast<unsigned int>(index) < d_components[component_id].instances.size());
	
	return d_components[component_id].instances[index].name;
}

TUCBudgetData::TComponentInstance TUCBudgetData::Component_Instance(
	const int							component_id,
	const int							index) const
{
	assert(static_cast<unsigned int>(component_id) < d_components.size());
	assert(static_cast<unsigned int>(index) < d_components[component_id].instances.size());
	
	return d_components[component_id].instances[index];
}

int TUCBudgetData::Component_Instance_Index(
	const int							component_id,
	const QString						&name) const
{
	std::vector<TUCBudgetData::TComponentInstance>::const_iterator iter;
	
	assert(static_cast<unsigned int>(component_id) < d_components.size());
	
	for(iter = d_components[component_id].instances.begin();iter != d_components[component_id].instances.end();++iter)
	{
		if((*iter).name == name)
		{
			return static_cast<int>(iter - d_components[component_id].instances.begin());
		}
	}
	
	return -1;
}

bool TUCBudgetData::Add_Component(
	const QString						&component_name)
{
	TUCBudgetData::TComponent			component;
	std::vector<TUCBudgetData::TComponent>::iterator iter;
	
	for(iter = d_components.begin();iter != d_components.end();++iter)
	{
		if((*iter).name == component_name)
		{
			return false;			
		}
	}
	
	component.name = component_name;
	component.mode = TUCBudgetData::SUM_ITEMS;
	component.length = 0.0;
	component.expansion_coefficient = 0.01;
	
	component.export_equipment = false;
	component.limit_uncertainty_length = false;

	d_components.push_back(component);
	
	return true;
}

bool TUCBudgetData::Remove_Component(
	const QString						&component)
{
	std::vector<TUCBudgetData::TComponent>::iterator iter;
	
	for(iter = d_components.begin();iter != d_components.end();++iter)
	{
		if((*iter).name == component)
		{
			d_components.erase(iter);
			return true;
		}
	}
	
	return false;
}

bool TUCBudgetData::Rename_Component(
	const QString						&old_component_name,
	const QString						&new_component_name)
{	
	std::vector<TUCBudgetData::TComponent>::iterator iter;
	
	for(iter = d_components.begin();iter != d_components.end();++iter)
	{
		if((*iter).name == new_component_name)
		{
			return false;
		}
	}
	
	for(iter = d_components.begin();iter != d_components.end();++iter)
	{
		if((*iter).name == old_component_name)
		{
			(*iter).name = new_component_name;
			return true;
		}
	}

	return false;

}

bool TUCBudgetData::Add_Component_Item(
	const QString						&component,
	const QString						&component_item)
{
	TUCBudgetData::TComponentItem		new_component_item;
	TUCBudgetData::TReplacementItem		new_replacement_item;
	std::vector<TUCBudgetData::TComponent>::iterator iter;
	std::vector<TUCBudgetData::TComponentInstance>::iterator id_iter;
	std::vector<TUCBudgetData::TComponentItem>::iterator item_iter;
	
	for(iter = d_components.begin();iter != d_components.end();++iter)
	{
		if((*iter).name == component)
		{
			for(item_iter = (*iter).items.begin();item_iter != (*iter).items.end();++item_iter)
			{
				if((*item_iter).name == component_item)
				{
					return false;
				}
			}
			
			new_component_item.name = component_item;
			
			new_component_item.expression = "0.0000";
			new_component_item.sensitivity = 1.0;
			new_component_item.distribution = TUCBudgetData::DISTRIBUTION_NORMAL;
			new_component_item.uctype = TUCBudgetData::TYPE_B;
			new_component_item.expression_type = TUCBudgetData::CONSTANT_LINEAR;
			
			// add replacement items
			
			for(id_iter = (*iter).instances.begin();id_iter != (*iter).instances.end();++id_iter)
			{
				new_replacement_item.identifier = (*id_iter).name;
				new_replacement_item.expression = new_component_item.expression;
				new_replacement_item.use_default_expression = true;
				
				new_component_item.replacement_items.push_back(new_replacement_item);
			}
			
			(*iter).items.push_back(new_component_item);
			return true;
		}
	}
	
	return false;
}

bool TUCBudgetData::Rename_Component_Item(
	const QString						&component,
	const QString						&old_item_name,
	const QString						&new_item_name)
{
	std::vector<TUCBudgetData::TComponent>::iterator iter;
	std::vector<TUCBudgetData::TComponentItem>::iterator item_iter;
	
	for(iter = d_components.begin();iter != d_components.end();++iter)
	{
		if((*iter).name == component)
		{
			for(item_iter = (*iter).items.begin();item_iter != (*iter).items.end();++item_iter)
			{
				if((*item_iter).name == new_item_name)
				{
					return false;
				}
			}
			
			for(item_iter = (*iter).items.begin();item_iter != (*iter).items.end();++item_iter)
			{
				if((*item_iter).name == old_item_name)
				{
					(*item_iter).name = new_item_name;
					
					return true;
				}
			}

			return false;
		}
	}
	
	return false;
}

bool TUCBudgetData::Remove_Component_Items(
	const QString						&component,
	const QStringList					&items)
{
	std::vector<TUCBudgetData::TComponent>::iterator iter;
	std::vector<TUCBudgetData::TComponentItem>::iterator item_iter;
	QList<QString>::const_iterator		str_iter;
	
	for(iter = d_components.begin();iter != d_components.end();++iter)
	{
		if((*iter).name == component)
		{
			for(str_iter = items.begin();str_iter != items.end();++str_iter)
			{
				for(item_iter = (*iter).items.begin();item_iter != (*iter).items.end();++item_iter)
				{
					if((*item_iter).name == (*str_iter))
					{
						(*iter).items.erase(item_iter);
						break;
					}
				}
			}
			
			return true;
		}
	}
	
	return false;
}

bool TUCBudgetData::Component_Add_Identifier(
	const QString						&component,
	const QString						&identifier)
{
	TUCBudgetData::TComponentInstance	new_component_identifier;
	TUCBudgetData::TReplacementItem		new_replacement_item;
	std::vector<TUCBudgetData::TComponent>::iterator iter;
	std::vector<TUCBudgetData::TComponentItem>::iterator item_iter;
	std::vector<TUCBudgetData::TComponentInstance>::iterator id_iter;
	
	for(iter = d_components.begin();iter != d_components.end();++iter)
	{
		if((*iter).name == component)
		{
			// check for existing name
			for(id_iter = (*iter).instances.begin();id_iter != (*iter).instances.end();++id_iter)
			{
				if((*id_iter).name == identifier)
				{
					return false;
				}
			}
			
			new_component_identifier.name = identifier;
			new_component_identifier.calibration_date = QDate::currentDate().toString();
			new_component_identifier.calibration_due_date = QDate::currentDate().toString();
			
			new_component_identifier.length = (*iter).length;
			new_component_identifier.expansion_coefficient = (*iter).expansion_coefficient;
			
			(*iter).instances.push_back(new_component_identifier);
			
			// update all compnent item replacements
			
			for(item_iter = (*iter).items.begin();item_iter != (*iter).items.end();++item_iter)
			{
				new_replacement_item.expression = (*item_iter).expression;
				new_replacement_item.use_default_expression = true;
				new_replacement_item.identifier = identifier;
				(*item_iter).replacement_items.push_back(new_replacement_item);
			}
			
			return true;
		}
	}
	
	return false;
}

bool TUCBudgetData::Rename_Component_Identifier(
	const QString						&component,
	const QString						&old_identifier,
	const QString						&new_identifier)
{
	std::vector<TUCBudgetData::TComponent>::iterator iter;
	std::vector<TUCBudgetData::TComponentItem>::iterator item_iter;
	std::vector<TUCBudgetData::TReplacementItem>::iterator replacement_iter;
	std::vector<TUCBudgetData::TComponentInstance>::iterator id_iter;
	
	for(iter = d_components.begin();iter != d_components.end();++iter)
	{
		if((*iter).name == component)
		{
			for(id_iter = (*iter).instances.begin();id_iter != (*iter).instances.end();++id_iter)
			{
				if((*id_iter).name == old_identifier)
				{
					(*id_iter).name = new_identifier;
				}
			}
			
			for(item_iter = (*iter).items.begin();item_iter != (*iter).items.end();++item_iter)
			{
				for(replacement_iter = (*item_iter).replacement_items.begin();replacement_iter != (*item_iter).replacement_items.end();++replacement_iter)
				{
					if((*replacement_iter).identifier == old_identifier)
					{
						(*replacement_iter).identifier = new_identifier;
					}
				}
			}
			
			return true;
		}
	}
	
	return false;
}

bool TUCBudgetData::Component_Remove_Identifier(
	const QString						&component,
	const QStringList					&identifiers)
{
	std::vector<TUCBudgetData::TComponent>::iterator iter;
	std::vector<TUCBudgetData::TComponentItem>::iterator item_iter;
	std::vector<TUCBudgetData::TReplacementItem>::iterator replacement_iter;
	std::vector<TUCBudgetData::TComponentInstance>::iterator id_iter;
	QList<QString>::const_iterator		str_iter;
	
	for(iter = d_components.begin();iter != d_components.end();++iter)
	{
		if((*iter).name == component)
		{
			for(str_iter = identifiers.begin();str_iter != identifiers.end();++str_iter)
			{
				for(id_iter = (*iter).instances.begin();id_iter != (*iter).instances.end();++id_iter)
				{
					if((*id_iter).name == (*str_iter))
					{
						(*iter).instances.erase(id_iter);
						break;
					}
				}
			}
			
			// remove all replacement items from all items under component
			for(item_iter = (*iter).items.begin();item_iter != (*iter).items.end();++item_iter)
			{
				for(str_iter = identifiers.begin();str_iter != identifiers.end();++str_iter)
				{
					for(replacement_iter = (*item_iter).replacement_items.begin();
						replacement_iter != (*item_iter).replacement_items.end();
						++replacement_iter)
					{
						if((*replacement_iter).identifier == (*str_iter))
						{
							(*item_iter).replacement_items.erase(replacement_iter);
							break;
						}
					}
				}
			}
			
			return true;
		}
	}
	
	return false;
}

void TUCBudgetData::Update_Component(
	const TUCBudgetData::TComponent		&component)
{
	std::vector<TUCBudgetData::TComponent>::iterator iter;
	
	for(iter = d_components.begin();iter != d_components.end();++iter)
	{
		if((*iter).name == component.name)
		{
			(*iter) = component;
			return;
		}
	}
}

void TUCBudgetData::Update_Component_Item(
	const int							component_id,
	const TComponentItem				&item)
{
	assert(static_cast<unsigned int>(component_id) < d_components.size());
	std::vector<TUCBudgetData::TComponentItem>::iterator iter;
	
	for(iter = d_components[component_id].items.begin();iter != d_components[component_id].items.end();++iter)
	{
		if((*iter).name == item.name)
		{
			(*iter) = item;
			return;
		}
	}
}

void TUCBudgetData::Update_Component_Identifier(
	const int							component_id,
	const TComponentInstance			&identifier)
{
	assert(static_cast<unsigned int>(component_id) < d_components.size());
	std::vector<TUCBudgetData::TComponentInstance>::iterator iter;
		
	for(iter = d_components[component_id].instances.begin();iter != d_components[component_id].instances.end();++iter)
	{
		if((*iter).name == identifier.name)
		{
			(*iter) = identifier;
			return;
		}
	}
	
	// set default length and expansion coefficient
	d_components[component_id].length = identifier.length;
	d_components[component_id].expansion_coefficient = identifier.expansion_coefficient;
}
