/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#ifndef UCBUDGETDATAHEADERFILE
#define UCBUDGETDATAHEADERFILE

#include <QString>
#include <QStringList>
#include <vector>


class TUCBudgetData
{
public:
	
enum TUCMode
{
	SUM_ITEMS=0,
	INDIVIUAL_ITEM
};
	
enum TDistribution
{
	DISTRIBUTION_NORMAL = 0,
	DISTRIBUTION_RECTANGULAR,
	DISTRUBUTION_TRIANGULAR,
	DISTRIBUTION_U,
	DISTRIBUTION_STEP
};

enum TUCType
{
	TYPE_A=0,
	TYPE_B
};
	
enum TExpressionType
{
	CONSTANT_LINEAR=0,
	CONSTANT_TEMPERATURE,
	LENGTH_DEPENDENT_LINEAR,
	LENGTH_DEPENDENT_TEMPERATURE,
};
	
struct TReplacementItem
{
	QString								identifier;
	QString								expression;
	bool								use_default_expression;
};
	
struct TComponentItem
{
	QString								name;
	QString								expression;
	TDistribution						distribution;
	TExpressionType						expression_type;
	TUCType								uctype;
	double								sensitivity;
	std::vector<TReplacementItem>		replacement_items;
};
	
struct TComponentInstance
{
	QString								name;
	QString								calibration_date;
	QString								calibration_due_date;
	double								length;
	double								expansion_coefficient;
};
	
struct TComponent
{
	QString								name;
	TUCMode								mode;
	double								length;
	double								expansion_coefficient;
	bool								export_equipment;
	bool								limit_uncertainty_length;
	std::vector<TComponentItem>			items;
	std::vector<TComponentInstance>		instances;
};
	
// CREATORS
	TUCBudgetData(void);
	~TUCBudgetData(void);
	
// ACCESSORS
	int Component_Count(void) const {return static_cast<int>(d_components.size());}
	QString Component_Name(const int index) const;
	TComponent Component(const int index) const;
	int Component_Index(const QString &name) const;
	
	int Component_Item_Count(const int component_id) const;
	QString Component_Item_Name(const int component_id,const int index) const;
	TComponentItem Component_Item(const int component_id,const int index) const;
	int Component_Item_Index(const int component_id,const QString &name) const;
	TComponentItem Component_Item(const QString &component_id,const QString &component_item_id) const;
	TComponentItem Component_Item(const int component_id,const QString &component_item_name) const 
				{return this->Component_Item(component_id,this->Component_Item_Index(component_id,component_item_name));}
	
	int Component_Identifier_Count(const int component_id) const;
	QString Component_Identifier_Name(const int component_id,const int index) const;
	TComponentInstance Component_Instance(const int component_id,const int index) const;
	int Component_Instance_Index(const int component_id,const QString &name) const;
	TComponentInstance Component_Instance(const int component_id,const QString &component_identifier_name) const 
		{return this->Component_Instance(component_id,this->Component_Instance_Index(component_id,component_identifier_name));}
	
															 
// MANIPULATORS
	bool Add_Component(const QString &component);
	bool Remove_Component(const QString &component);
	bool Rename_Component(const QString &old_component_name,const QString &new_component_name);

	bool Add_Component_Item(const QString &component,const QString &component_item);
	bool Rename_Component_Item(const QString &component_name,const QString &old_item_name,const QString &new_item_name);
	bool Remove_Component_Items(const QString &component,const QStringList &items);
	
	bool Component_Add_Identifier(const QString &component,const QString &identifier);
	bool Rename_Component_Identifier(const QString &component,const QString &old_identifier,const QString &new_identifier);
	bool Component_Remove_Identifier(const QString &component,const QStringList &identifiers);
	
	void Update_Component(const TComponent &component);
	void Update_Component_Item(const int component_id,const TComponentItem &item);
	void Update_Component_Identifier(const int component_id,const TComponentInstance &identifier);
	
	void Clear(void) {d_components.clear();}
	void Add_Component(const TComponent &component) {d_components.push_back(component);}

private:
	
	std::vector<TComponent>				d_components;

	// NOT IMPLEMENTED
	TUCBudgetData(const TUCBudgetData&);
	TUCBudgetData& operator=(const TUCBudgetData&);
};

#endif
