/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////



#include <QVariant>
#include <QAction>
#include <QApplication>
#include <QButtonGroup>
#include <QCheckBox>
#include <QComboBox>
#include <QFrame>
#include <QGridLayout>
#include <QHeaderView>
#include <QLabel>
#include <QLineEdit>
#include <QListWidget>
#include <QRadioButton>
#include <QSpacerItem>
#include <QVBoxLayout>
#include <QFont>

#include "componentitemwidget.h"

TComponentItemWidget::TComponentItemWidget(
	const QWidget						*parent,
	Qt::WindowFlags						flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	QLabel								*uctype_label;
	QLabel								*distribution_label;
	QLabel								*sensitivity_label;
	QLabel								*options_label;
	QFrame								*separator_line_bottom;
	QLabel								*expression_value_label;
	QLabel								*expression_description_label;
	QSpacerItem							*widget_vspacer;
	QWidget								*variable_type_widget;
	QVBoxLayout							*typewidget_vlayout;
	QWidget								*unit_type_widget;
	QVBoxLayout							*unitwidget_vlayout;
	QFont								title_font;

	this->setMinimumWidth(300);
	
	title_font.setPointSize(18);
	title_font.setUnderline(true);

	widget_layout = new QGridLayout(this);
	widget_layout->setContentsMargins(0,0,0,0);
		
	d_ucitem_label = new QLabel(this);
	d_ucitem_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	d_ucitem_label->setFont(title_font);
	widget_layout->addWidget(d_ucitem_label,0,0,1,4);

	uctype_label = new QLabel(this);
	widget_layout->addWidget(uctype_label,1,0,1,1);

	d_uctype_combo = new QComboBox(this);
	widget_layout->addWidget(d_uctype_combo,1,1,1,3);

	distribution_label = new QLabel(this);
	widget_layout->addWidget(distribution_label,2,0,1,1);

	d_distribution_combo = new QComboBox(this);
	widget_layout->addWidget(d_distribution_combo,2,1,1,3);

	sensitivity_label = new QLabel(this);
	widget_layout->addWidget(sensitivity_label,3,0,1,1);

	d_sensitivity_edit = new QLineEdit(this);
	widget_layout->addWidget(d_sensitivity_edit,3,1,1,3);

	expression_value_label = new QLabel(this);
	widget_layout->addWidget(expression_value_label,4,0,1,1);

	d_expression_edit = new QLineEdit(this);
	widget_layout->addWidget(d_expression_edit,4,1,1,3);

	expression_description_label = new QLabel(this);
	widget_layout->addWidget(expression_description_label,5,0,1,1);

	d_expression_description_label = new QLabel(this);
	widget_layout->addWidget(d_expression_description_label,5,1,1,3);
	
	separator_line_bottom = new QFrame(this);
	separator_line_bottom->setFrameShape(QFrame::HLine);
	separator_line_bottom->setFrameShadow(QFrame::Sunken);
	widget_layout->addWidget(separator_line_bottom,6,0,1,4);

	options_label = new QLabel(this);
	options_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	options_label->setFont(title_font);
	widget_layout->addWidget(options_label,7,0,1,4);

	d_use_default_check = new QCheckBox(this);
	widget_layout->addWidget(d_use_default_check,8,0,1,4);

	variable_type_widget = new QWidget(this);

	typewidget_vlayout = new QVBoxLayout(variable_type_widget);
	typewidget_vlayout->setContentsMargins(0,0,0,0);

	d_constant_radio = new QRadioButton(variable_type_widget);
	typewidget_vlayout->addWidget(d_constant_radio);

	d_length_dependent_radio = new QRadioButton(variable_type_widget);
	typewidget_vlayout->addWidget(d_length_dependent_radio);

	widget_layout->addWidget(variable_type_widget,9,0,1,4);

	unit_type_widget = new QWidget(this);

	unitwidget_vlayout = new QVBoxLayout(unit_type_widget);
	unitwidget_vlayout->setContentsMargins(0,0,0,0);

	d_linear_radio = new QRadioButton(unit_type_widget);
	unitwidget_vlayout->addWidget(d_linear_radio);

	d_temperature_radio = new QRadioButton(unit_type_widget);
	unitwidget_vlayout->addWidget(d_temperature_radio);

	widget_layout->addWidget(unit_type_widget,10,0,1,4);
	
	widget_vspacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	widget_layout->addItem(widget_vspacer,11,0,1,4);
	
	// defaults
	d_linear_radio->setChecked(true);
	d_constant_radio->setChecked(true);
	d_block_modifed_signal = false;
	this->Set_Idenfitier_Name(QStringLiteral("[Default]"));

	uctype_label->setText(QStringLiteral("UC Type:"));
	distribution_label->setText(QStringLiteral("Distribution:"));
	options_label->setText(QStringLiteral("Options:"));
	sensitivity_label->setText(QStringLiteral("Sensitivity:"));
	expression_value_label->setText(QStringLiteral("Expression:"));
	d_expression_description_label->setText(QStringLiteral("Uc = f(exp) mm"));
	d_use_default_check->setText(QStringLiteral("Use Default Expression"));
	d_constant_radio->setText(QStringLiteral("Expression result is independent of length"));
	d_length_dependent_radio->setText(QStringLiteral("Expression result is length dependent"));
	d_linear_radio->setText(QStringLiteral("Expression describes a linear dimension"));
	d_temperature_radio->setText(QStringLiteral("Expression describes temperature"));
	
	connect(d_use_default_check,&QCheckBox::toggled,this,&TComponentItemWidget::Use_Default_Toggled);
	connect(d_use_default_check,&QCheckBox::toggled,this,&TComponentItemWidget::Expression_Changed);
	
	connect(d_constant_radio,&QRadioButton::toggled,this,&TComponentItemWidget::Item_Modified);
	connect(d_linear_radio,&QRadioButton::toggled,this,&TComponentItemWidget::Item_Modified);
	connect(d_distribution_combo,static_cast<void (QComboBox::*)(int)>(&QComboBox::currentIndexChanged),this,&TComponentItemWidget::Item_Modified);
	connect(d_uctype_combo,static_cast<void (QComboBox::*)(int)>(&QComboBox::currentIndexChanged),this,&TComponentItemWidget::Item_Modified);

	connect(d_expression_edit,&QLineEdit::editingFinished,this,&TComponentItemWidget::Expression_Changed);
	connect(d_sensitivity_edit,&QLineEdit::editingFinished,this,&TComponentItemWidget::Item_Modified);
}

TComponentItemWidget::~TComponentItemWidget(void)
{
}

TComponentItemWidget::TRadioOptions TComponentItemWidget::Radio_Options(void) const
{
	if(d_constant_radio->isChecked() && d_linear_radio->isChecked())
	{
		return TComponentItemWidget::CONSTANT_LINEAR;
	}
	else if(d_constant_radio->isChecked() && d_temperature_radio->isChecked())
	{
		return TComponentItemWidget::CONSTANT_TEMPERATURE;
	}
	else if(d_length_dependent_radio->isChecked() && d_linear_radio->isChecked())
	{
		return TComponentItemWidget::LENGTH_DEPENDENT_LINEAR;
	}

	return TComponentItemWidget::LENGTH_DEPENDENT_TEMPERATURE;
}

QString TComponentItemWidget::UCType(void) const
{
	return d_uctype_combo->currentText();
}

QString TComponentItemWidget::Distribution_Type(void) const
{
	return d_distribution_combo->currentText();
}

double TComponentItemWidget::Sensitivity(void) const
{
	return d_sensitivity_edit->text().toDouble();
}

QString TComponentItemWidget::Expression(void) const
{
	return d_expression_edit->text();
}

bool TComponentItemWidget::Use_Default_Expression(void) const
{
	return d_use_default_check->isChecked();
}

void TComponentItemWidget::Reset_Distribution_Types(
	const QStringList					&list)
{
	d_block_modifed_signal = true;
	
	d_distribution_combo->clear();
	d_distribution_combo->addItems(list);
	
	d_block_modifed_signal = false;

}

void TComponentItemWidget::Reset_Uncertainty_Types(
	const QStringList					&list)
{
	d_block_modifed_signal = true;

	d_uctype_combo->clear();
	d_uctype_combo->addItems(list);
	
	d_block_modifed_signal = false;

}

void TComponentItemWidget::Set_Idenfitier_Name(
	const QString 						&identifier)
{
	d_ucitem_label->setText(QString("Identifier: %1").arg(identifier));
}

void TComponentItemWidget::Set_Radio_Options(
	const TComponentItemWidget::TRadioOptions options)
{
	d_block_modifed_signal = true;
	
	switch(options)
	{
		case TComponentItemWidget::CONSTANT_LINEAR:
			d_constant_radio->setChecked(true);
			d_linear_radio->setChecked(true);
			break;
			
		case TComponentItemWidget::CONSTANT_TEMPERATURE:
			d_constant_radio->setChecked(true);
			d_temperature_radio->setChecked(true);
			break;
			
		case TComponentItemWidget::LENGTH_DEPENDENT_LINEAR:
			d_length_dependent_radio->setChecked(true);
			d_linear_radio->setChecked(true);
			break;
			
		case TComponentItemWidget::LENGTH_DEPENDENT_TEMPERATURE:
			d_length_dependent_radio->setChecked(true);
			d_temperature_radio->setChecked(true);
			break;
	}
	
	d_block_modifed_signal = false;
}

void TComponentItemWidget::Set_UCType(
	const QString						&type)
{
	int									index;
	
	index = d_uctype_combo->findText(type,Qt::MatchExactly);
	
	if(!(index < 0))
	{
		d_block_modifed_signal = true;
		
		d_uctype_combo->setCurrentIndex(index);
		
		d_block_modifed_signal = false;
	}
}

void TComponentItemWidget::Set_Distribution_Type(
	const QString						&type)
{
	int									index;
	
	index = d_distribution_combo->findText(type,Qt::MatchExactly);
	
	if(!(index < 0))
	{
		d_block_modifed_signal = true;
		
		d_distribution_combo->setCurrentIndex(index);
		
		d_block_modifed_signal = false;
	}
}

void TComponentItemWidget::Set_Sensitivity(
	const double						&val)
{
	d_sensitivity_edit->setText(QString("%1").arg(val,0,'f',6));
}

void TComponentItemWidget::Set_Expression(
	const QString						&text)
{
	d_expression_edit->setText(text);
}

void TComponentItemWidget::Set_Option_Use_Default(
	const bool							state)
{
	bool								prev_state;
	
	prev_state = d_use_default_check->blockSignals(true);
	d_block_modifed_signal = true;
	
	d_use_default_check->setChecked(state);
	
	d_block_modifed_signal = false;
	d_use_default_check->blockSignals(prev_state);
	
	if(state)
	{
		d_expression_edit->setEnabled(false);
	}
	else
	{
		d_expression_edit->setEnabled(true);
	}
}

void TComponentItemWidget::Enable_Option_Use_Default(
	const bool 							state)
{
	d_use_default_check->setEnabled(state);
}

void TComponentItemWidget::Use_Default_Toggled(
	bool								state)
{
	d_expression_edit->setEnabled(!state);
}

void TComponentItemWidget::Expression_Changed(void)
{
	QString								id;
		
	if(!d_block_modifed_signal)
	{
		emit Component_Item_Changed();
	}
}

void TComponentItemWidget::Item_Modified(void)
{	
	if(d_linear_radio->isChecked())
	{
		if(d_constant_radio->isChecked())
		{
			d_expression_description_label->setText(QStringLiteral("Uc = f(expression) mm"));
		}
		else
		{
			d_expression_description_label->setText(QStringLiteral("Uc = f(expression) mm/m"));
		}
	}
	else
	{
		if(d_constant_radio->isChecked())
		{
			d_expression_description_label->setText(QStringLiteral("Uc = f(expression) C"));
		}
		else
		{
			d_expression_description_label->setText(QStringLiteral("Uc = f(expression) C/m"));
		}
	}
	
	if(!d_block_modifed_signal)
	{
		emit Component_Item_Changed();
	}
}


