/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////



#include <QVariant>
#include <QAction>
#include <QApplication>
#include <QButtonGroup>
#include <QComboBox>
#include <QDateEdit>
#include <QFrame>
#include <QGridLayout>
#include <QHeaderView>
#include <QLabel>
#include <QLineEdit>
#include <QListWidget>
#include <QRadioButton>
#include <QCheckBox>
#include <QSpacerItem>
#include <QVBoxLayout>
#include <QToolButton>
#include <QFont>
#include <assert.h>

#include "../../core/messagebox.h"

#include "componentwidget.h"
#include "namedialog.h"

TComponentWidget::TComponentWidget(
	const QWidget						*parent,
	Qt::WindowFlags						flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	QWidget								*identifier_container_widget;
	QLabel								*length_unit_label;
	QLabel								*length_label;
	QLabel								*calibration_due_label;
	QLabel								*expansion_coefficient_label;
	QLabel								*expansion_coefficient_unit_label;
	QLabel								*options_label;
	QLabel								*calibration_date_label;
	QLabel								*identifiers_label;
	QFrame								*separator_line;
	QVBoxLayout							*identifier_container_vlayout;
	QHBoxLayout							*identifier_container_hlayout;
	QSpacerItem							*identifier_container_hspacer;
	QSpacerItem							*control_vspacer;
	QFont								title_font;
	
	title_font.setPointSize(18);
	title_font.setUnderline(true);
	
	widget_layout = new QGridLayout(this);
	widget_layout->setContentsMargins(0,0,0,0);

	identifiers_label = new QLabel(this);
	identifiers_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	identifiers_label->setFont(title_font);
	widget_layout->addWidget(identifiers_label,0,0,1,3);

	identifier_container_widget = new QWidget(this);
	
	identifier_container_vlayout = new QVBoxLayout(identifier_container_widget);
	identifier_container_vlayout->setContentsMargins(0,0,0,0);
	identifier_container_vlayout->setSpacing(4);

	d_identifiers_list = new QListWidget(this);
	d_identifiers_list->setSelectionMode(QAbstractItemView::ExtendedSelection);
	d_identifiers_list->setSortingEnabled(true);
	identifier_container_vlayout->addWidget(d_identifiers_list);
	
	identifier_container_hlayout = new QHBoxLayout;
	
	d_add_identifier_button = new QToolButton(identifier_container_widget);
	d_add_identifier_button->setFixedSize(16,16);
	identifier_container_hlayout->addWidget(d_add_identifier_button);
	
	d_remove_identifiers_button = new QToolButton(identifier_container_widget);
	d_remove_identifiers_button->setFixedSize(16,16);
	identifier_container_hlayout->addWidget(d_remove_identifiers_button);
	
	identifier_container_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	identifier_container_hlayout->addItem(identifier_container_hspacer);
	
	identifier_container_vlayout->addLayout(identifier_container_hlayout);

	widget_layout->addWidget(identifier_container_widget,1,0,1,3);
	
	calibration_date_label = new QLabel(this);
	widget_layout->addWidget(calibration_date_label,2,0,1,1);
	
	d_calibration_date = new QDateEdit(this);
	d_calibration_date->setDisplayFormat("ddd MMM d yyyy");
	widget_layout->addWidget(d_calibration_date,2,1,1,2);
	
	calibration_due_label = new QLabel(this);
	widget_layout->addWidget(calibration_due_label,3,0,1,1);

	d_calibration_due_date = new QDateEdit(this);
	d_calibration_due_date->setDisplayFormat("ddd MMM d yyyy");
	widget_layout->addWidget(d_calibration_due_date,3,1,1,2);
	
	length_label = new QLabel(this);
	widget_layout->addWidget(length_label,4,0,1,1);
	
	d_length_edit = new QLineEdit(this);
	widget_layout->addWidget(d_length_edit,4,1,1,1);
	
	length_unit_label = new QLabel(this);
	widget_layout->addWidget(length_unit_label,4,2,1,1);
	
	expansion_coefficient_label = new QLabel(this);
	widget_layout->addWidget(expansion_coefficient_label,5,0,1,1);
	
	d_expansion_coefficient_edit = new QLineEdit(this);
	widget_layout->addWidget(d_expansion_coefficient_edit,5,1,1,1);

	expansion_coefficient_unit_label = new QLabel(this);
	widget_layout->addWidget(expansion_coefficient_unit_label,5,2,1,1);
	
	separator_line = new QFrame(this);
	separator_line->setFrameShape(QFrame::HLine);
	separator_line->setFrameShadow(QFrame::Sunken);
	widget_layout->addWidget(separator_line,6,0,1,3);

	options_label = new QLabel(this);
	options_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	options_label->setFont(title_font);
	widget_layout->addWidget(options_label,7,0,1,3);
	
	d_ucmode_individual_radio = new QRadioButton(this);
	widget_layout->addWidget(d_ucmode_individual_radio,8,0,1,3);

	d_ucmode_combine_radio = new QRadioButton(this);
	d_ucmode_combine_radio->setChecked(true);
	widget_layout->addWidget(d_ucmode_combine_radio,9,0,1,3);
		
	d_export_equipment_check = new QCheckBox(this);
	widget_layout->addWidget(d_export_equipment_check,10,0,1,3);
	
	d_limit_length_check = new QCheckBox(this);
	widget_layout->addWidget(d_limit_length_check,11,0,1,3);

	control_vspacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	widget_layout->addItem(control_vspacer,12,0,1,3);


	// defaults
	d_msg_box = new TMessageBox(this);
	d_bock_modifed_signal = false;
	d_remove_identifiers_button->setEnabled(false);

	d_add_identifier_button->setIcon(QIcon(QStringLiteral(":/gui/add.png")));
	d_remove_identifiers_button->setIcon(QIcon(QStringLiteral(":/gui/remove.png")));

	length_unit_label->setText(QStringLiteral("mm"));
	length_label->setText(QStringLiteral("Length:"));
	calibration_due_label->setText(QStringLiteral("Calibration Due Date:"));
	expansion_coefficient_label->setText(QStringLiteral("Expansion Coefficient:"));
	expansion_coefficient_unit_label->setText(QStringLiteral("um/m/˚C"));
	options_label->setText(QStringLiteral("Options:"));
	calibration_date_label->setText(QStringLiteral("Calibration Date:"));
	identifiers_label->setText(QStringLiteral("Identifiers:"));
	d_ucmode_individual_radio->setText(QStringLiteral("Report as individual uncertainty items"));
	d_ucmode_combine_radio->setText(QStringLiteral("Report as a combined uncertainty item"));
	d_export_equipment_check->setText(QStringLiteral("Report component hint"));
	d_limit_length_check->setText(QStringLiteral("Limit uncertainty to component length"));
	
	connect(d_ucmode_individual_radio,&QRadioButton::toggled,this,&TComponentWidget::Data_Changed);
	connect(d_ucmode_combine_radio,&QRadioButton::toggled,this,&TComponentWidget::Data_Changed);
	connect(d_export_equipment_check,&QCheckBox::toggled,this,&TComponentWidget::Data_Changed);
	connect(d_limit_length_check,&QCheckBox::toggled,this,&TComponentWidget::Data_Changed);
	connect(d_calibration_date,&QDateEdit::dateChanged,this,&TComponentWidget::Data_Changed);
	connect(d_calibration_due_date,&QDateEdit::dateChanged,this,&TComponentWidget::Data_Changed);
	connect(d_length_edit,&QLineEdit::editingFinished,this,&TComponentWidget::Data_Changed);
	connect(d_expansion_coefficient_edit,&QLineEdit::editingFinished,this,&TComponentWidget::Data_Changed);
	
	connect(d_identifiers_list,&QListWidget::itemSelectionChanged,this,&TComponentWidget::Identifier_Selected);
	connect(d_identifiers_list->itemDelegate(),&QAbstractItemDelegate::commitData,this,&TComponentWidget::Rename_Identifier);

	connect(d_add_identifier_button, &QToolButton::clicked,this, &TComponentWidget::Add_Identifier);
	connect(d_remove_identifiers_button, &QToolButton::clicked,this, &TComponentWidget::Remove_Identifiers);
}

TComponentWidget::~TComponentWidget(void)
{
}

QString TComponentWidget::Current_Identifier(void) const
{
	return d_current_identifier;
}

QString TComponentWidget::Calibration_Date(void) const
{
	return d_calibration_date->date().toString("MMM d yyyy");
}

QString TComponentWidget::Calibration_Due_Date(void) const
{
	return d_calibration_due_date->date().toString("MMM d yyyy");
}

double TComponentWidget::Expansion_Coefficient(void) const
{
	return d_expansion_coefficient_edit->text().toDouble() / 1000;
}

bool TComponentWidget::Mode_Single(void) const
{
	return d_ucmode_individual_radio->isChecked();
}

bool TComponentWidget::Mode_Sum(void) const
{
	return d_ucmode_combine_radio->isChecked();
}

bool TComponentWidget::Mode_Export_Equipment(void) const
{
	return d_export_equipment_check->isChecked();
}

bool TComponentWidget::Mode_Limit_Uncertainty_Length(void) const
{
	return d_limit_length_check->isChecked();
}

double TComponentWidget::Length(void) const
{
	return d_length_edit->text().toDouble();
}

void TComponentWidget::Identifier_Selected(void)
{
	QList<QListWidgetItem*>				sel_items;
			
	sel_items = d_identifiers_list->selectedItems();
	
	if(sel_items.size() == 1)
	{
		d_calibration_date->setEnabled(true);
		d_calibration_due_date->setEnabled(true);
		
		d_current_identifier = sel_items[0]->text();
		emit Component_Identifier_Selected(sel_items[0]->text());
	}
	else
	{
		d_calibration_date->setEnabled(false);
		d_calibration_due_date->setEnabled(false);
		
		d_current_identifier = QString();
		emit Component_Identifier_Selected(QString());
	}
	
	d_remove_identifiers_button->setEnabled(sel_items.size() > 0);
}

void TComponentWidget::Add_Identifier(void)
{
	TNameDialog							name_dialog(this);
	QString								name;
	QList<QListWidgetItem*>				items;
	QListWidgetItem						*item;

	name_dialog.Set_Title("Add Identifier");
	
	if(name_dialog.exec() == QDialog::Accepted)
	{
		name = name_dialog.Name();
		
		items = d_identifiers_list->findItems(name,Qt::MatchFixedString);
		
		if(items.size() == 0)
		{
			item = new QListWidgetItem(d_identifiers_list);
			item->setText(name);
			item->setFlags(item->flags() | Qt::ItemIsEditable);

			d_current_identifier_list.push_back(name);
			
			emit Add_Component_Identifier(name);
		}
	}
}

void TComponentWidget::Rename_Identifier(
	QWidget 							*editor)
{
	QString 							new_name;
	QStringList::const_iterator			iter;
	int									cntr;
	int									index(-1);
		
	new_name = reinterpret_cast<QLineEdit*>(editor)->text();
	
	if(new_name.length())
	{
		for(iter = d_current_identifier_list.begin(),cntr = 0;iter != d_current_identifier_list.end();++iter,++cntr)
		{
			if(new_name.compare((*iter)) == 0)
			{
				d_identifiers_list->item(cntr)->setText(d_current_identifier);
				return;
			}
			
			if(d_current_identifier.compare((*iter)) == 0)
			{
				index = cntr;
			}
		}
				
		emit Rename_Component_Identifiers(d_current_identifier,new_name);
		d_current_identifier = new_name;
		
		emit Component_Identifier_Selected(d_current_identifier);
		
		assert(!(index < 0));
		d_current_identifier_list[index] = new_name;
	}
}

void TComponentWidget::Remove_Identifiers(void)
{
	QList<QListWidgetItem*>				items;
	QStringList							list;
	QList<QListWidgetItem*>::const_iterator item_iter;
	
	items = d_identifiers_list->selectedItems();
	
	if(items.size())
	{
		d_msg_box->setText("Delete Selected Component Identifiers");
		d_msg_box->setInformativeText("Are you sure you want to delete the selected component identifier items?");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_msg_box->setDefaultButton(QMessageBox::No);
		d_msg_box->setIcon(QMessageBox::Question);
		
		if(d_msg_box->exec() == QMessageBox::No)
		{
			return;
		}

		for(item_iter = items.begin();item_iter != items.end();++item_iter)
		{
			list.push_back((*item_iter)->text());
			
			delete d_identifiers_list->takeItem(d_identifiers_list->row(*item_iter));
		}
		
		emit Remove_Component_Identifiers(list);
	}
}

void TComponentWidget::Data_Changed(void)
{	
	if(!d_bock_modifed_signal)
	{
		emit Component_Changed();
	}
}

void TComponentWidget::Set_Identifiers(
	const QStringList					&ids)
{
	QStringList::const_iterator			iter;
	QListWidgetItem						*item;
	
	d_current_identifier_list = ids;
	
	d_identifiers_list->clear();
	d_current_identifier = QString();
	
	for(iter = ids.begin();iter != ids.end();++iter)
	{
		item = new QListWidgetItem(d_identifiers_list);
		item->setText((*iter));
		item->setFlags(item->flags() | Qt::ItemIsEditable);
	}
}

void TComponentWidget::Set_Dates(
	const QString						&cal_date,
	const QString						&cal_due_date)
{
	QDate								date;
	
	d_bock_modifed_signal = true;
	
	if(cal_date.size())
	{
		date = QDate::fromString(cal_date,"MMM d yyyy");
		
		if(date.isValid())
		{
			d_calibration_date->setDate(date);
		}
		else
		{
			d_calibration_date->setDate(QDate::currentDate());
		}
	}
	else
	{
		d_calibration_date->setDate(QDate::currentDate());
	}
	
	if(cal_due_date.size())
	{
		date = QDate::fromString(cal_due_date,"MMM d yyyy");
		
		if(date.isValid())
		{
			d_calibration_due_date->setDate(date);
		}
		else
		{
			d_calibration_due_date->setDate(QDate::currentDate());
		}
	}
	else
	{
		d_calibration_due_date->setDate(QDate::currentDate());
	}
	
	d_bock_modifed_signal = false;
}

void TComponentWidget::Clear_Dates(void)
{
	d_calibration_date->setEnabled(false);
	d_calibration_due_date->setEnabled(false);
}

void TComponentWidget::Set_Expansion_Coefficient(
	const double						&value)
{
	d_bock_modifed_signal = true;
	
	d_expansion_coefficient_edit->setText(QString("%1").arg(value * 1000,0,'f',2));

	d_bock_modifed_signal = false;
}

void TComponentWidget::Set_Mode_Single(void)
{
	d_bock_modifed_signal = true;
	
	d_ucmode_individual_radio->setChecked(true);
	
	d_bock_modifed_signal = false;
}

void TComponentWidget::Set_Mode_Sum(void)
{
	d_bock_modifed_signal = true;
	
	d_ucmode_combine_radio->setChecked(true);
	
	d_bock_modifed_signal = false;
}

void TComponentWidget::Set_Mode_Export_Equipment(
	const bool							state)
{
	d_bock_modifed_signal = true;

	d_export_equipment_check->setChecked(state);
	
	d_bock_modifed_signal = false;
}
void TComponentWidget::Set_Mode_Limit_Uncertainty_Length(
	const bool							state)
{
	d_bock_modifed_signal = true;
	
	d_limit_length_check->setChecked(state);
	
	d_bock_modifed_signal = false;
}

void TComponentWidget::Set_Length(
	const double						&len)
{
	d_bock_modifed_signal = true;
	
	d_length_edit->setText(QString("%1").arg(len,0,'f',5));
	
	d_bock_modifed_signal = false;
}
