/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QApplication>
#include <QButtonGroup>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QHeaderView>
#include <QLabel>
#include <QLineEdit>
#include <QListWidget>
#include <QToolButton>
#include <QSpacerItem>
#include <QList>
#include <QSize>
#include <QVariant>
#include <QVBoxLayout>
#include <QFont>
#include <assert.h>

#include "namedialog.h"
#include "reportgroupwidget.h"

TReportGroupWidget::TReportGroupWidget(
	const QWidget						*parent,
	Qt::WindowFlags						flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	QVBoxLayout							*reportgroups_container_vlayout;
	QHBoxLayout							*reportgroups_button_hlayout;
	QLabel								*uccomponents_label;
	QSpacerItem							*reportgroup_button_hspacer;
	QLabel								*reportgroups_label;
	QWidget								*reportgroups_container_widget;
	QFont								title_font;
	
	title_font.setPointSize(18);
	title_font.setUnderline(true);

	widget_layout = new QGridLayout(this);
	
	reportgroups_label = new QLabel(this);
	reportgroups_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	reportgroups_label->setFont(title_font);
	widget_layout->addWidget(reportgroups_label,0,0,1,1);
	
	reportgroups_container_widget = new QWidget(this);
	
	reportgroups_container_vlayout = new QVBoxLayout(reportgroups_container_widget);
	reportgroups_container_vlayout->setContentsMargins(0,0,0,0);
	reportgroups_container_vlayout->setSpacing(4);

	d_reportgroups_list = new QListWidget(reportgroups_container_widget);
	d_reportgroups_list->setSelectionMode(QAbstractItemView::ExtendedSelection);
	reportgroups_container_vlayout->addWidget(d_reportgroups_list);
	
	reportgroups_button_hlayout = new QHBoxLayout();
	
	d_add_reportgroup_button = new QToolButton(reportgroups_container_widget);
	d_add_reportgroup_button->setFixedSize(16,16);
	reportgroups_button_hlayout->addWidget(d_add_reportgroup_button);
	
	d_remove_reportgroup_button = new QToolButton(reportgroups_container_widget);
	d_remove_reportgroup_button->setFixedSize(16,16);
	reportgroups_button_hlayout->addWidget(d_remove_reportgroup_button);
	
	reportgroup_button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	reportgroups_button_hlayout->addItem(reportgroup_button_hspacer);

	reportgroups_container_vlayout->addLayout(reportgroups_button_hlayout);

	widget_layout->addWidget(reportgroups_container_widget,1,0,1,1);

	uccomponents_label = new QLabel(this);
	uccomponents_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	uccomponents_label->setFont(title_font);
	widget_layout->addWidget(uccomponents_label,0,1,1,1);

	d_uccomponents_list = new QListWidget(this);
	d_uccomponents_list->setSelectionMode(QAbstractItemView::ExtendedSelection);
	widget_layout->addWidget(d_uccomponents_list, 1,1,1,1);
	
	d_add_reportgroup_button->setIcon(QIcon(QStringLiteral(":/gui/add.png")));
	d_remove_reportgroup_button->setIcon(QIcon(QStringLiteral(":/gui/remove.png")));
	
	uccomponents_label->setText(QStringLiteral("Available Budget Components:"));
	reportgroups_label->setText(QStringLiteral("Current Report Groups:"));

	// defaults
	d_uccomponents_list->setEnabled(false);
	
	connect(d_reportgroups_list,&QListWidget::itemSelectionChanged,this,&TReportGroupWidget::Report_Group_Selection_Changed);
	connect(d_reportgroups_list->itemDelegate(),&QAbstractItemDelegate::commitData,this,&TReportGroupWidget::Rename_Current_Group);
	connect(d_uccomponents_list,&QListWidget::itemSelectionChanged,this,&TReportGroupWidget::Uc_Item_Selection_Changed);
	
	connect(d_add_reportgroup_button,&QToolButton::clicked,this,&TReportGroupWidget::Create_New_Group);
	connect(d_remove_reportgroup_button,&QToolButton::clicked,this,&TReportGroupWidget::Remove_Selected_Groups);
}

TReportGroupWidget::~TReportGroupWidget(void)
{
}

TGroup TReportGroupWidget::Group(
	const int							index) const
{
	assert(static_cast<unsigned int>(index) < d_groups.size());
	
	return d_groups[index];
}

QStringList TReportGroupWidget::Group_Names(void) const
{
	QStringList							list;
	std::vector<TGroup>::const_iterator iter;
	
	for(iter = d_groups.begin();iter != d_groups.end();++iter)
	{
		list.push_back((*iter).name);
	}

	return list;
}

QStringList TReportGroupWidget::Group_List(
	const QString						&group_name) const
{
	QStringList							list;
	std::vector<TGroup>::const_iterator iter;
	
	for(iter = d_groups.begin();iter != d_groups.end();++iter)
	{
		if((*iter).name == group_name)
		{
			list = (*iter).items;
			break;
		}
	}
	
	return list;
}

void TReportGroupWidget::Uc_Item_Selection_Changed(void)
{
	QList<QListWidgetItem*>				sel_items;
	QList<QListWidgetItem*>::const_iterator item_iter;
	QStringList							components_list;
	QString								group_name;
	
	sel_items = d_uccomponents_list->selectedItems();

	if(sel_items.size())
	{
		for(item_iter = sel_items.begin();item_iter != sel_items.end();++item_iter)
		{
			components_list.push_back((*item_iter)->text());
		}
		
		if(components_list.size())
		{
			// modify selected group if active
			sel_items = d_reportgroups_list->selectedItems();
			
			if(sel_items.size() == 1)
			{
				group_name = sel_items[0]->text();
				this->Modify_Group(group_name,components_list);
				
				emit Report_Groups_Changed();
			}
		}
	}
	else
	{
		d_reportgroups_list->clearSelection();
	}
}

void TReportGroupWidget::Report_Group_Selection_Changed(void)
{
	QList<QListWidgetItem*>				sel_items;
	QStringList							items;
	QStringList::ConstIterator			str_iter;
	QListWidgetItem						*item;
	bool								prev_state;
	
	sel_items = d_reportgroups_list->selectedItems();
	
	d_remove_reportgroup_button->setEnabled(sel_items.size() != 0);
	
	if(sel_items.size() == 1)
	{
		d_uccomponents_list->setEnabled(true);
		
		d_current_report_group = sel_items[0]->text();
		
		items = this->Group_Items(d_current_report_group);
		
		prev_state = d_uccomponents_list->blockSignals(true);
		
		d_uccomponents_list->clearSelection();
		
		for(str_iter = items.begin();str_iter != items.end();++str_iter)
		{
			item = this->Find_Component_Item(*str_iter);
			
			if(item)
			{
				item->setSelected(true);
			}
		}
		
		d_uccomponents_list->blockSignals(prev_state);
	}
	else
	{
		d_uccomponents_list->setEnabled(false);
		d_current_report_group = QString();
	}
}

void TReportGroupWidget::Create_New_Group(void)
{
	QList<QListWidgetItem*>				sel_items;
	QList<QListWidgetItem*>::const_iterator item_iter;
	TNameDialog							name_dialog(this);
	TGroup								group;
	QListWidgetItem						*item;
	
	name_dialog.Set_Title("Add Report Group");
	
	if(name_dialog.exec() == QDialog::Accepted)
	{
		group.name = name_dialog.Name();
		
		sel_items = d_uccomponents_list->selectedItems();

		for(item_iter = sel_items.begin();item_iter != sel_items.end();++item_iter)
		{
			group.items.push_back((*item_iter)->text());
		}
		
		if(this->Add_Group(group))
		{			
			item = new QListWidgetItem(d_reportgroups_list);
			item->setFlags(item->flags() | Qt::ItemIsEditable);
			item->setText(group.name);
			
			emit Report_Groups_Changed();
		}
	}
}

void TReportGroupWidget::Remove_Selected_Groups(void)
{
	QList<QListWidgetItem*>				sel_items;
	QList<QListWidgetItem*>::const_iterator item_iter;
	QListWidgetItem						*item;
	QString								name;

	sel_items = d_reportgroups_list->selectedItems();
	
	if(sel_items.size())
	{
		for(item_iter = sel_items.begin();item_iter != sel_items.end();++item_iter)
		{
			name = (*item_iter)->text();
			this->Remove_Group(name);
			
			item = d_reportgroups_list->takeItem(d_reportgroups_list->row(*item_iter));
			delete item;
		}
		
		emit Report_Groups_Changed();
	}
}


void TReportGroupWidget::Rename_Current_Group(
	QWidget 							*editor)
{
	QString 							new_name;
	
	new_name = reinterpret_cast<QLineEdit*>(editor)->text();
	
	if(d_current_report_group.length() && d_current_report_group.compare(new_name) != 0)
	{
		this->Rename_Group(d_current_report_group,new_name);
		emit Report_Groups_Changed();
	}
}

void TReportGroupWidget::Update_Reportable_Components(
	const QStringList					&list)
{
	d_uccomponents_list->clear();
	d_uccomponents_list->addItems(list);
}

void TReportGroupWidget::Set_Groups(
	const std::vector<TGroup> &groups)
{
	std::vector<TGroup>::const_iterator iter;
	QListWidgetItem						*item;
	
	d_groups = groups;
	
	d_reportgroups_list->clear();
	
	for(iter = d_groups.begin();iter != d_groups.end();++iter)
	{
		item = new QListWidgetItem(d_reportgroups_list);
		item->setFlags(item->flags() | Qt::ItemIsEditable);
		item->setText((*iter).name);
	}
}

void TReportGroupWidget::Clear_Groups(void)
{
	d_groups.clear();
	d_reportgroups_list->clear();
}

bool TReportGroupWidget::Add_Group(
	const TGroup						&group)
{
	std::vector<TGroup>::iterator iter;
	
	for(iter = d_groups.begin();iter != d_groups.end();++iter)
	{
		if((*iter).name == group.name)
		{
			return false;
		}
	}

	d_groups.push_back(group);
	
	return true;
}

QListWidgetItem* TReportGroupWidget::Find_Component_Item(
	const QString						&name) const
{
	for(int cntr = 0;cntr < d_uccomponents_list->count();++cntr)
	{
		if(d_uccomponents_list->item(cntr)->text() == name)
		{
			return d_uccomponents_list->item(cntr);
		}
	}
	
	return 0;
}

void TReportGroupWidget::Remove_Group(
	const QString						&name)
{
	std::vector<TGroup>::iterator iter;
	
	for(iter = d_groups.begin();iter != d_groups.end();++iter)
	{
		if((*iter).name == name)
		{
			d_groups.erase(iter);
			return;
		}
	}
}

void TReportGroupWidget::Rename_Group(
	const QString						&old_name,
	const QString						&new_name)
{
	std::vector<TGroup>::iterator iter;
	
	for(iter = d_groups.begin();iter != d_groups.end();++iter)
	{
		if((*iter).name == old_name)
		{
			(*iter).name = new_name;
			return;
		}
	}
}

QStringList TReportGroupWidget::Group_Items(
	const QString						&name)
{
	std::vector<TGroup>::iterator iter;
	
	for(iter = d_groups.begin();iter != d_groups.end();++iter)
	{
		if((*iter).name == name)
		{
			return (*iter).items;
		}
	}
	
	return QStringList();
}

void TReportGroupWidget::Modify_Group(
	const QString						&group_name,
	const QStringList					&items)
{
	std::vector<TGroup>::iterator iter;

	for(iter = d_groups.begin();iter != d_groups.end();++iter)
	{
		if((*iter).name == group_name)
		{
			(*iter).items = items;
			return;
		}
	}
}


