/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////



#include <QVariant>
#include <QAction>
#include <QApplication>
#include <QButtonGroup>
#include <QComboBox>
#include <QFrame>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QHeaderView>
#include <QLabel>
#include <QPushButton>
#include <QSpacerItem>
#include <QBrush>
#include <QColor>
#include <QFile>
#include <QPainter>
#include <QPrinter>
#include <QRect>
#include <QIcon>
#include <QPen>
#include <math.h>

#include "../../core/optioneditor.h"
#include "../../core/interpolation.h"

#include "reportwidget.h"

TReportWidget::TReportWidget(
	const QWidget						*parent,
	Qt::WindowFlags						flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	QHBoxLayout							*button_hlayout;
	QLabel								*report_label;
	QSpacerItem							*widget_vspacer;
	QLabel								*output_label;
	QSpacerItem							*control_hspacer;
	QSpacerItem							*button_hspacer;
	QFrame								*separator_line;
	QLabel								*reportgroup_label;

	widget_layout = new QGridLayout(this);
	
	reportgroup_label = new QLabel(this);
	widget_layout->addWidget(reportgroup_label, 0, 0, 1, 1);

	d_reportgroup_combo = new QComboBox(this);
	d_reportgroup_combo->setSizeAdjustPolicy(QComboBox::AdjustToContents);
	widget_layout->addWidget(d_reportgroup_combo, 0, 1, 1, 1);
	
	control_hspacer = new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Minimum);
	widget_layout->addItem(control_hspacer, 0, 2, 1, 1);

	report_label = new QLabel(this);
	report_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	widget_layout->addWidget(report_label, 1, 0, 1, 1);

	d_report_table = new TOptionEditorWidget(this);
	d_report_table->setColumnCount(7);
	d_report_table->headerItem()->setText(0, QStringLiteral("Description"));
	d_report_table->headerItem()->setText(1, QStringLiteral("Identifier"));
	d_report_table->headerItem()->setText(2, QStringLiteral("Type"));
	d_report_table->headerItem()->setText(3, QStringLiteral("Expression"));
	d_report_table->headerItem()->setText(4, QStringLiteral("Sensitivity"));
	d_report_table->headerItem()->setText(5, QStringLiteral("Distribution Type"));
	d_report_table->headerItem()->setText(6, QStringLiteral("Value"));
	widget_layout->addWidget(d_report_table, 1, 1, 1, 2);
	
	d_status_label = new QLabel(this);
	widget_layout->addWidget(d_status_label, 2, 1, 1, 2);
	
	separator_line = new QFrame(this);
	separator_line->setFrameShape(QFrame::HLine);
	separator_line->setFrameShadow(QFrame::Sunken);
	widget_layout->addWidget(separator_line, 3, 0, 1, 3);

	output_label = new QLabel(this);
	widget_layout->addWidget(output_label, 4, 0, 1, 1);
	
	button_hlayout = new QHBoxLayout;

	d_export_pdf_button = new QPushButton(this);
	d_export_pdf_button->setEnabled(false);
	button_hlayout->addWidget(d_export_pdf_button);
	
	d_export_csv_button = new QPushButton(this);
	d_export_csv_button->setEnabled(false);
	button_hlayout->addWidget(d_export_csv_button);
	
	button_hspacer = new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Minimum);
	button_hlayout->addItem(button_hspacer);
	widget_layout->addLayout(button_hlayout,4,1,1,2);
	
	widget_vspacer = new QSpacerItem(0, 0, QSizePolicy::Minimum, QSizePolicy::Preferred);
	widget_layout->addItem(widget_vspacer, 5, 0, 1, 3);

	report_label->setText(QStringLiteral("Report Data:"));
	d_export_pdf_button->setText(QStringLiteral("Export PDF"));
	d_export_csv_button->setText(QStringLiteral("Export CSV"));
	output_label->setText(QStringLiteral("Output:"));
	reportgroup_label->setText(QStringLiteral("Report Group:"));
		
	// defaults
	d_reportgroup_combo->addItem("<no selection>");
	d_export_format = TReportWidget::EXPORT_PDF_FORMAT;
		
	connect(d_export_pdf_button,&QPushButton::clicked,this,&TReportWidget::Generate_PDF);
	connect(d_export_csv_button,&QPushButton::clicked,this,&TReportWidget::Generate_CSV);
	connect(d_reportgroup_combo,static_cast<void (QComboBox::*)(int)>(&QComboBox::activated),this,&TReportWidget::Generate_Report);
	connect(d_report_table,&TOptionEditorWidget::SelectionListItemChanged,this,&TReportWidget::Identifier_Selection_Changed);	
}

TReportWidget::~TReportWidget(void)
{
}

QString TReportWidget::Current_Identifier(
	const QString						&component_name) const
{
	std::vector<TReportWidget::TReportData>::const_iterator iter;
		
	for(iter = d_report_data.begin();iter != d_report_data.end();++iter)
	{
		if((*iter).component_name == component_name)
		{
			return (*iter).identifier;
		}
	}
	
	return QString();
}

bool TReportWidget::Write_Export_CSV_Data(
	const QString						&file_name) const
{
	QFile								file(file_name);
	std::vector<QString>::const_iterator str_iter;
	QString								line;
	QString								blank_line("\n");
	
    if(!file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		d_last_error = QString("ERR:  %1").arg(file.errorString());
        return false;
	}
	
	line = QString("Budget Definition - %1\n").arg(d_reportgroup_combo->currentText());
	file.write(line.toLatin1());
	file.write(blank_line.toLatin1());

	for(str_iter = d_item_data.begin();str_iter != d_item_data.end();++str_iter)
	{
		line = (*str_iter) + '\n';
		
		file.write(line.toLatin1());
	}
	
	line = "Global Variables\n";
	file.write(line.toLatin1());
	file.write(blank_line.toLatin1());
	
	for(str_iter = d_variable_data.begin();str_iter != d_variable_data.end();++str_iter)
	{
		line = (*str_iter) + '\n';
		
		file.write(line.toLatin1());
	}
	
	line = "Budget Data\n";
	file.write(line.toLatin1());
	file.write(blank_line.toLatin1());
	
	for(str_iter = d_value_csv_data.begin();str_iter != d_value_csv_data.end();++str_iter)
	{
		line = (*str_iter) + '\n';
		
		file.write(line.toLatin1());
	}
	
	file.write(blank_line.toLatin1());
	
	for(str_iter = d_uncertainty_csv_data.begin();str_iter != d_uncertainty_csv_data.end();++str_iter)
	{
		line = (*str_iter) + '\n';
		
		file.write(line.toLatin1());
	}
	
	file.close();

	return true;
}

bool TReportWidget::Write_Export_PDF_Data(
	const QString						&file_name) const
{
	std::vector<QString>::const_iterator str_iter;
	QList<QString>::iterator			str_item_iter;
	QString								footer_text;
	QString								group_title;
	QString								line;
	QStringList							line_item_list;
	QPainter							painter;
	QPrinter							printer(QPrinter::HighResolution);
	QPen								black_pen(QColor(0,0,0));
	QPen								grey_pen(QColor(164,164,164));
	QRect								page_rect;
	QRect								rect;
	QFont								text_font("verdana",8,QFont::Normal);
	QFont								title_font("verdana",10,QFont::Bold);
	QFont								heading_font("verdana",8,QFont::Bold);
	QFont								footer_font("verdana",6,QFont::Normal);
	QMarginsF							margins;
	int									line_height;
	int									column_width;
	int									max_lines;
	int									row,column;
	int									page_margin;
	int									data_index;
	int									last_data_index;
	int									page_number;
	int									column_positions[18];
	int									row_positions[4];
	bool								init_page;
	bool								next_line_bold(false);
	bool								data_fit_page;
	static const int					max_data_values(11);
	
	printer.setPageOrientation(QPageLayout::Landscape);
	printer.setOutputFormat(QPrinter::PdfFormat);
	printer.setOutputFileName(file_name);
	
	margins.setLeft(0.0);
	margins.setTop(0.0);
	margins.setRight(0.0);
	margins.setBottom(0.0);
	printer.setPageMargins(margins,QPageLayout::Millimeter);
	
	if(!painter.begin(&printer))
	{
		return false;
	}
	
	footer_text = QString("Select Calibration Incorporated.  www.selectcalibration.ca");
	group_title = d_reportgroup_combo->currentText();

	painter.setFont(text_font);
	page_rect = printer.pageLayout().paintRectPixels(printer.resolution());
	
	page_margin = static_cast<int>(static_cast<double>(page_rect.width()) * 0.03);	// estimation based on 8.5x11 and a small margin
	page_rect.adjust(page_margin,2 * page_margin,-page_margin,-page_margin);
	
	rect = painter.boundingRect(0,0,0,0,Qt::AlignLeft,"10000.0000j");
	line_height = rect.height();
	
	// calculate initial column and row positions.
	// title | description | source | expression | sensitivity | distribution type | distribution divisor
	
	column_width = static_cast<int>((static_cast<double>(page_rect.width()) / 18.0) + 0.5);
	column_positions[0] = page_rect.left();

	for(column = 1;column < 18;++column)
	{
		column_positions[column] = column_positions[column - 1] + column_width;
	}
		
	// 0 title (2 rows)
	// 1 top of data header
	// 2 top of data
	// 3 footer
	
	row_positions[0] = page_rect.top();
	row_positions[1] = row_positions[0] + 2 * line_height;
	row_positions[2] = row_positions[1] + line_height;
	row_positions[3] = page_rect.bottom() - line_height;
	
	max_lines = ((row_positions[3] - row_positions[2]) / line_height) - 2;
	
	
	page_number = 1;
	
	row = 1;
	init_page = true;
	
	painter.setPen(black_pen);
			
	for(str_iter = d_item_data.begin();str_iter != d_item_data.end();++str_iter)
	{
		if(init_page)
		{
			painter.setFont(title_font);
			rect.setRect(column_positions[0],row_positions[0],column_positions[17] - column_positions[0] + column_width,row_positions[1] - row_positions[0]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignTop,QString("Budget Definition - %1").arg(group_title));

			painter.setFont(footer_font);
			
			rect.setRect(column_positions[0],row_positions[3],10 * column_width, line_height);
			painter.drawText(rect,Qt::AlignLeft,footer_text);
			
			rect.setRect(column_positions[16],row_positions[3],2 * column_width, line_height);
			painter.drawText(rect,Qt::AlignRight,QString("Page %1").arg(page_number));
			
			painter.setFont(text_font);
			init_page = false;
		}
		
		line = (*str_iter);
		line_item_list = this->Split_CSV_Data(line);
		
		for(str_item_iter = line_item_list.begin();str_item_iter != line_item_list.end();++str_item_iter)
		{
			(*str_item_iter).remove('\"');
		}
		
		while(line_item_list.size() < 8)
		{
			line_item_list.push_back(QString());
		}
		
		if(line_item_list[0].length())
		{
			painter.setFont(heading_font);
			
			rect.setRect(column_positions[0],row_positions[2] + row * line_height,column_width * 18, row_positions[2] - row_positions[1]);
			
			painter.setPen(grey_pen);
			painter.drawLine(rect.bottomLeft(),rect.bottomRight());
			painter.setPen(black_pen);

			next_line_bold = true;

		}
		else
		{
			if(next_line_bold)
			{
				painter.setFont(heading_font);
				next_line_bold = false;
			}
			else
			{
				painter.setFont(text_font);
			}
		}

		rect.setRect(column_positions[0],row_positions[2] + row * line_height,column_width * 3, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[0]));
		
		rect.setRect(column_positions[1],row_positions[2] + row * line_height,column_width * 3, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[1]));
		
		rect.setRect(column_positions[5],row_positions[2] + row * line_height,column_width * 3, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[2]));
		
		rect.setRect(column_positions[7],row_positions[2] + row * line_height,column_width * 3, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[3]));
		
		rect.setRect(column_positions[11],row_positions[2] + row * line_height,column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[4]));

		rect.setRect(column_positions[12],row_positions[2] + row * line_height,column_width * 2, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[5]));
		
		rect.setRect(column_positions[14],row_positions[2] + row * line_height,column_width * 2, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[6]));
		
		rect.setRect(column_positions[16],row_positions[2] + row * line_height,column_width * 3, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[7]));
		
		row++;
		
		if(row > max_lines)
		{
			printer.newPage();
			page_number++;
			
			painter.setFont(text_font);
			page_rect = printer.pageLayout().paintRectPixels(printer.resolution());
			
			page_margin = static_cast<int>(static_cast<double>(page_rect.width()) * 0.03);
			page_rect.adjust(page_margin,2 * page_margin,-page_margin,-page_margin);
			
			init_page = true;
			row = 1;
		}
	}
	
	
	
	row++;
	if(row > (max_lines -4))	// minimum number of items on this page
	{
		printer.newPage();
		page_number++;
		
		painter.setFont(text_font);
		page_rect = printer.pageLayout().paintRectPixels(printer.resolution());
		
		page_margin = static_cast<int>(static_cast<double>(page_rect.width()) * 0.03);
		page_rect.adjust(page_margin,2 * page_margin,-page_margin,-page_margin);
		
		init_page = true;
		row = 1;
	}

	painter.setFont(heading_font);
	
	rect.setRect(column_positions[0],row_positions[2] + row * line_height,column_width * 18, line_height);

	painter.setPen(grey_pen);
	painter.drawLine(rect.bottomLeft(),rect.bottomRight());
	painter.setPen(black_pen);


	rect.setRect(column_positions[0],row_positions[2] + row * line_height,column_width * 3, line_height);
	painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("Global Variables"));
	
	row++;
	painter.setFont(text_font);

	
	// global variables
	for(str_iter = d_variable_data.begin();str_iter != d_variable_data.end();++str_iter)
	{
		if(init_page)
		{
			painter.setFont(title_font);
			rect.setRect(column_positions[0],row_positions[0],column_positions[17] - column_positions[0] + column_width,row_positions[1] - row_positions[0]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignTop,QString("Budget Definition - %1").arg(group_title));

			painter.setFont(footer_font);
			
			rect.setRect(column_positions[0],row_positions[3],10 * column_width, line_height);
			painter.drawText(rect,Qt::AlignLeft,footer_text);
			
			rect.setRect(column_positions[16],row_positions[3],2 * column_width, line_height);
			painter.drawText(rect,Qt::AlignRight,QString("Page %1").arg(page_number));
			
			painter.setFont(text_font);
			init_page = false;
		}
		
		line = (*str_iter);
		line_item_list = this->Split_CSV_Data(line);
		
		for(str_item_iter = line_item_list.begin();str_item_iter != line_item_list.end();++str_item_iter)
		{
			(*str_item_iter).remove('\"');
		}
		
		while(line_item_list.size() < 3)
		{
			line_item_list.push_back(QString());
		}
				
		rect.setRect(column_positions[0],row_positions[2] + row * line_height,column_width * 3, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[0]));
		
		rect.setRect(column_positions[1],row_positions[2] + row * line_height,column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[1]));
		
		rect.setRect(column_positions[2],row_positions[2] + row * line_height,column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[2]));
		
		row++;
		
		if(row > max_lines && (d_variable_data.end() - str_iter) > 1)	// only create a new page if there is more data
		{
			printer.newPage();
			page_number++;
			
			painter.setFont(text_font);
			page_rect = printer.pageLayout().paintRectPixels(printer.resolution());
			
			page_margin = static_cast<int>(static_cast<double>(page_rect.width()) * 0.03);
			page_rect.adjust(page_margin,2 * page_margin,-page_margin,-page_margin);
			
			init_page = true;
			row = 1;
		}
	}
		
	// uncertainty data
	last_data_index = 0;

	do
	{
		// budget data.  New page automatically
		printer.newPage();
		page_number++;

		painter.setFont(text_font);
		
		page_rect = printer.pageLayout().paintRectPixels(printer.resolution());
		
		page_margin = static_cast<int>(static_cast<double>(page_rect.width()) * 0.03);
		page_rect.adjust(page_margin,2 * page_margin,-page_margin,-page_margin);
		
		init_page = true;
		row = 1;
	
		data_fit_page = true;
		for(str_iter = d_value_pdf_data.begin();str_iter != d_value_pdf_data.end();++str_iter)
		{
			if(init_page)
			{
				painter.setFont(title_font);
				rect.setRect(column_positions[0],row_positions[0],column_positions[17] - column_positions[0] + column_width,row_positions[1] - row_positions[0]);
				painter.drawText(rect,Qt::AlignLeft | Qt::AlignTop,QString("Budget Data - %1").arg(group_title));

				painter.setFont(footer_font);
				
				rect.setRect(column_positions[0],row_positions[3],10 * column_width, line_height);
				painter.drawText(rect,Qt::AlignLeft,footer_text);
				
				rect.setRect(column_positions[16],row_positions[3],2 * column_width, line_height);
				painter.drawText(rect,Qt::AlignRight,QString("Page %1").arg(page_number));
				
				painter.setFont(text_font);
				init_page = false;
			}
			
			line = (*str_iter);
			line_item_list = this->Split_CSV_Data(line);
			
			for(str_item_iter = line_item_list.begin();str_item_iter != line_item_list.end();++str_item_iter)
			{
				(*str_item_iter).remove('\"');
			}
			
			if((line_item_list.size() - last_data_index - 3) > max_data_values)
			{
				data_fit_page = false;
			}
					
			while(line_item_list.size() < (max_data_values + last_data_index) + 3)
			{
				line_item_list.push_back(QString());
			}
			
			if(line_item_list[0].length())
			{
				painter.setFont(heading_font);
				
				rect.setRect(column_positions[0],row_positions[2] + row * line_height,column_width * 18, row_positions[2] - row_positions[1]);

				painter.setPen(grey_pen);
				painter.drawLine(rect.bottomLeft(),rect.bottomRight());
				painter.setPen(black_pen);

				next_line_bold = false;
				
			}
			else
			{
				if(next_line_bold)
				{
					painter.setFont(heading_font);
					next_line_bold = false;
				}
				else
				{
					painter.setFont(text_font);
				}
			}
			
			rect.setRect(column_positions[0],row_positions[2] + row * line_height,column_width * 3, line_height);
			painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[0]));
			
			rect.setRect(column_positions[1],row_positions[2] + row * line_height,column_width * 2, line_height);
			painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[1]));
			
			rect.setRect(column_positions[3],row_positions[2] + row * line_height,column_width * 3, line_height);
			painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[2]));
			
			for(data_index = 3;data_index < (max_data_values + 3);++data_index)
			{
				rect.setRect(column_positions[data_index + 4],row_positions[2] + row * line_height,column_width, line_height);
				painter.drawText(rect,Qt::AlignCenter | Qt::TextDontClip,QString("%1").arg(line_item_list[last_data_index + data_index]));
			}
					
			row++;
			
			if(row > max_lines)
			{
				printer.newPage();
				page_number++;
				
				painter.setFont(text_font);
				page_rect = printer.pageLayout().paintRectPixels(printer.resolution());
				
				page_margin = static_cast<int>(static_cast<double>(page_rect.width()) * 0.03);
				page_rect.adjust(page_margin,2 * page_margin,-page_margin,-page_margin);
				
				init_page = true;
				row = 1;
			}
		}
		
		last_data_index += max_data_values;
		
	}while(!data_fit_page);

	// uncertainty statement
	row++;
	if(row > (max_lines - 6))	// minimum number of items on this page
	{
		printer.newPage();
		page_number++;
		
		painter.setFont(text_font);
		page_rect = printer.pageLayout().paintRectPixels(printer.resolution());
		
		page_margin = static_cast<int>(static_cast<double>(page_rect.width()) * 0.03);
		page_rect.adjust(page_margin,2 * page_margin,-page_margin,-page_margin);
		
		init_page = true;
		row = 1;
	}
	
	painter.setFont(heading_font);
	
	rect.setRect(column_positions[0],row_positions[2] + row * line_height,column_width * 18, line_height);

	painter.setPen(grey_pen);
	painter.drawLine(rect.bottomLeft(),rect.bottomRight());
	painter.setPen(black_pen);

	rect.setRect(column_positions[0],row_positions[2] + row * line_height,column_width * 3, line_height);
	painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("Uncertainty Expression"));
	
	row++;
	row++;
	
	for(str_iter = d_uncertainty_pdf_data.begin();str_iter != d_uncertainty_pdf_data.end();++str_iter)
	{
		if(init_page)
		{
			painter.setFont(title_font);
			rect.setRect(column_positions[0],row_positions[0],column_positions[17] - column_positions[0] + column_width,row_positions[1] - row_positions[0]);
			painter.drawText(rect,Qt::AlignLeft | Qt::AlignTop,QString("Budget Data - %1").arg(group_title));

			painter.setFont(footer_font);
			
			rect.setRect(column_positions[0],row_positions[3],10 * column_width, line_height);
			painter.drawText(rect,Qt::AlignLeft,footer_text);
			
			rect.setRect(column_positions[16],row_positions[3],2 * column_width, line_height);
			painter.drawText(rect,Qt::AlignRight,QString("Page %1").arg(page_number));
			
			init_page = false;
		}
		
		painter.setFont(text_font);

		line = (*str_iter);
		line_item_list = this->Split_CSV_Data(line);
		
		for(str_item_iter = line_item_list.begin();str_item_iter != line_item_list.end();++str_item_iter)
		{
			(*str_item_iter).remove('\"');
		}
		
		while(line_item_list.size() < 6)
		{
			line_item_list.push_back(QString());
		}
		
		rect.setRect(column_positions[0],row_positions[2] + row * line_height,column_width * 3, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[0]));
		
		rect.setRect(column_positions[1],row_positions[2] + row * line_height,column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[2]));

		rect.setRect(column_positions[2],row_positions[2] + row * line_height,column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[3]));

		rect.setRect(column_positions[3],row_positions[2] + row * line_height,column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[4]));
		
		rect.setRect(column_positions[6],row_positions[2] + row * line_height,column_width, line_height);
		painter.drawText(rect,Qt::AlignLeft | Qt::TextDontClip,QString("%1").arg(line_item_list[5]));
		
		row++;
		
		if(row > max_lines)
		{
			printer.newPage();
			page_number++;
			
			painter.setFont(text_font);
			page_rect = printer.pageLayout().paintRectPixels(printer.resolution());
			
			page_margin = static_cast<int>(static_cast<double>(page_rect.width()) * 0.03);
			page_rect.adjust(page_margin,2 * page_margin,-page_margin,-page_margin);
			
			init_page = true;
			row = 1;
		}
	}
	
	painter.end();
	
	return true;
}

void TReportWidget::Generate_PDF(void)
{
	QString								report_group;
	
	report_group = d_reportgroup_combo->currentText();
	d_export_format = TReportWidget::EXPORT_PDF_FORMAT;
	
	emit Create_Export_Report(report_group);
}

void TReportWidget::Generate_CSV(void)
{
	QString								report_group;
	
	report_group = d_reportgroup_combo->currentText();
	d_export_format = TReportWidget::EXPORT_CSV_FORMAT;

	emit Create_Export_Report(report_group);
}

void TReportWidget::Generate_Report(void)
{
	int									sel_index;
	
	sel_index = d_reportgroup_combo->currentIndex();
	
	if(sel_index == 0)
	{
		this->Clear_Report();
		return;
	}
	
	emit Create_Report(d_reportgroup_combo->currentText());
}

void TReportWidget::Identifier_Selection_Changed(
	int									index,
	const QString						&text)
{
	std::vector<TReportWidget::TReportData>::iterator iter;
		
	for(iter = d_report_data.begin();iter != d_report_data.end();++iter)
	{
		if((*iter).identifier_list_index == index)
		{
			(*iter).identifier = text;
			break;
		}
	}
	
	this->Generate_Report();
}

void TReportWidget::Update_Report_Groups(
	const QStringList					&list)
{
	QString								text;
	int									index;
	
	text = d_reportgroup_combo->currentText();
	
	d_reportgroup_combo->clear();
	d_reportgroup_combo->addItem("<no selection>");
	d_reportgroup_combo->addItems(list);
	
	index = d_reportgroup_combo->findText(text,Qt::MatchExactly);
	
	if(!(index < 0))
	{
		d_reportgroup_combo->setCurrentIndex(index);
	}
}

void TReportWidget::Clear_Report(void)
{
	d_report_data.clear();
	d_report_table->clear();
	d_export_pdf_button->setEnabled(false);
	d_export_csv_button->setEnabled(false);
}

void TReportWidget::Set_Default_Report_Group(void)
{
	if(d_reportgroup_combo->count() > 0)
	{
		d_reportgroup_combo->setCurrentIndex(0);
	}
}

void TReportWidget::Start_Report_Update(void)
{
	d_block_signals_report_update = d_report_table->blockSignals(true);
	
	d_export_pdf_button->setEnabled(true);
	d_export_csv_button->setEnabled(true);
}

void TReportWidget::Add_Report_Data(
	const TReportData					&report_data)
{
	QTreeWidgetItem						*item;
	QFont								title_font(this->font());
	TReportWidget::TReportData			report_data_local;
	int									identifier_index;
	std::vector<TReportWidget::TReportItemData>::const_iterator item_iter;
	bool								init_loop;
	QBrush								error_brush(QColor(255,0,0));
	
	title_font.setBold(true);
	
	report_data_local = report_data;
		
	identifier_index = report_data_local.identifier_list.indexOf(report_data_local.identifier);
	
	if(identifier_index < 0)
	{
		identifier_index = 0;
	}
	
	init_loop = true;
	for(item_iter = report_data_local.report_items.begin();item_iter != report_data_local.report_items.end();++item_iter)
	{
		if(init_loop)
		{
			report_data_local.identifier_list_index = d_report_table->AddSelectionList(
																				   report_data_local.component_name,
																				   report_data_local.identifier_list,
																				   identifier_index);
			
			d_report_table->topLevelItem(d_report_table->topLevelItemCount()-1)->setFont(0,title_font);
			
			init_loop = false;
		}

		item = new QTreeWidgetItem;
			
		item->setText(0,(*item_iter).component_item_name);
		item->setText(1,QString());
		item->setText(2,(*item_iter).value_type);
		item->setText(3,(*item_iter).expression);
		item->setText(4,(*item_iter).sensitivity);
		item->setText(5,(*item_iter).distribution);
		item->setText(6,(*item_iter).value_string);
		
		if((*item_iter).value_string == "ERR")
		{
			d_export_pdf_button->setEnabled(false);
			d_export_csv_button->setEnabled(false);
			
			item->setForeground(0,error_brush);
			item->setForeground(1,error_brush);
			item->setForeground(2,error_brush);
			item->setForeground(3,error_brush);
			item->setForeground(4,error_brush);
			item->setForeground(5,error_brush);
			item->setForeground(6,error_brush);
		}
		
		d_report_table->addTopLevelItem(item);
	}

	d_report_data.push_back(report_data_local);
}

void TReportWidget::End_Report_Update(void)
{
	d_report_table->resizeColumnToContents(0);
	d_report_table->resizeColumnToContents(1);
	d_report_table->resizeColumnToContents(2);
	d_report_table->resizeColumnToContents(3);
	d_report_table->resizeColumnToContents(4);
	d_report_table->resizeColumnToContents(5);
	d_report_table->resizeColumnToContents(6);

	d_report_table->blockSignals(d_block_signals_report_update);
}

void TReportWidget::Clear_Export(void)
{
	d_export_data.clear();
	d_export_variable_data.clear();
}

void TReportWidget::Add_Export_Data(
	const TReportWidget::TReportData	&report_data,
	const double						&position)
{
	TReportWidget::TExportData			export_data;
	std::vector<TReportWidget::TReportItemData>::const_iterator item_iter;
	bool								init_loop;
		
	init_loop = true;
	for(item_iter = report_data.report_items.begin();item_iter != report_data.report_items.end();++item_iter)
	{
		if(init_loop)
		{
			export_data.component_name = report_data.component_name;
			export_data.identifier = report_data.identifier;
			export_data.is_variable = report_data.is_variable;
			init_loop = false;
		}
		
		export_data.component_item = (*item_iter);
				
		this->Add_Export_Data(export_data,position);
	}
}

void TReportWidget::Add_Export_Variable_Data(
	const TReportWidget::TExportVariableData &data)
{
	std::vector<TReportWidget::TExportVariableData>::iterator iter;
	
	for(iter = d_export_variable_data.begin();iter != d_export_variable_data.end();++iter)
	{
		if(data.component_name.length())
		{
			if(data.component_name == (*iter).component_name)
			{
				if(data.variable_name == (*iter).variable_name)
				{
					(*iter).values.insert((*iter).values.end(),data.values.begin(),data.values.end());
					return;
				}
			}
		}
		else
		{
			if((*iter).component_name.length() == 0)
			{
				if(data.variable_name == (*iter).variable_name)
				{
					(*iter).values.insert((*iter).values.end(),data.values.begin(),data.values.end());
					return;
				}
			}
		}
	}
	
	d_export_variable_data.push_back(data);
}

void TReportWidget::Format_Export_Data(void)
{
	std::vector<TReportWidget::TExportData>::const_iterator data_iter;
	std::vector<TReportWidget::TExportVariableData>::const_iterator var_iter;
	std::vector<double>::const_iterator value_iter;
	std::vector<TVector2>::iterator		uc_iter;
	QString								line,line2;
	QString								last_component_name;
	int									precision;
	bool								first_line;
	static const int					default_precision(6);
	static const int					pdf_precision(5);
	static const int					csv_precision(9);
	double								a,b,c;

	d_item_data.clear();
	d_variable_data.clear();
	d_value_csv_data.clear();
	d_value_pdf_data.clear();
	d_uncertainty_csv_data.clear();
	d_uncertainty_pdf_data.clear();
	
	// data map
	for(data_iter = d_export_data.begin();data_iter != d_export_data.end();++data_iter)
	{
		if((*data_iter).component_name != last_component_name)
		{
			// start new section
			d_item_data.push_back(QString());

			if((*data_iter).is_variable)
			{
				line = QString("%1 Component %2 Variable").arg((*data_iter).component_name).arg((*data_iter).identifier);
			}
			else
			{
				line = QString("%1 Component %2").arg((*data_iter).component_name).arg((*data_iter).identifier);
			}

			d_item_data.push_back(line);
			
			line = QString(",Source,Value,Expression,Type,Sensitivity,Distribution,Divisor");
			d_item_data.push_back(line);
			
			last_component_name = (*data_iter).component_name;
		}
		
		line = QString(",%1").arg((*data_iter).component_item.component_item_name);
		line += QString(",%1").arg((*data_iter).component_item.value_type);
		line += QString(",\"%1\"").arg((*data_iter).component_item.expression);
		line += QString(",%1").arg((*data_iter).component_item.type);
		line += QString(",%1").arg((*data_iter).component_item.sensitivity);
		line += QString(",%1").arg((*data_iter).component_item.distribution);
		line += QString(",%1").arg((*data_iter).component_item.distribution_divisor,0,'f',default_precision);
		
		d_item_data.push_back(line);
	}
	
	// global variables
	
	for(var_iter = d_export_variable_data.begin();var_iter != d_export_variable_data.end();++var_iter)
	{
		if((*var_iter).component_name.length() == 0 && (*var_iter).values.size() > 0)
		{
			if((*var_iter).variable_name != QStringLiteral("CTEc")
			   && (*var_iter).variable_name != QStringLiteral("CTEd")
			   && (*var_iter).variable_name != QStringLiteral("Lc")
			   && (*var_iter).variable_name != QStringLiteral("L"))		// These are updated locally
			{
				line = QString(",%1,%2").arg((*var_iter).variable_name).arg((*var_iter).values[0],0,'f',default_precision);
				d_variable_data.push_back(line);
			}
		}
	}
		
	// data values

	last_component_name = QString();
	first_line = true;
	for(data_iter = d_export_data.begin();data_iter != d_export_data.end();++data_iter)
	{
		if(first_line)
		{
			line = QString(",,Position (mm)");
			line2 = QString(",,Position (mm)");
			for(value_iter = (*data_iter).positions.begin();value_iter != (*data_iter).positions.end();++value_iter)
			{
				line += QString(",%1").arg((*value_iter),0,'f',3);
				line2 += QString(",%1").arg((*value_iter),0,'f',1);
			}
			d_value_csv_data.push_back(line);
			d_value_pdf_data.push_back(line2);

			first_line = false;
		}
		
		if((*data_iter).component_name != last_component_name)
		{
			// start new section
			d_value_csv_data.push_back(QString());
			d_value_pdf_data.push_back(QString());
			
			if((*data_iter).is_variable)
			{
				line = QString("%1 Component %2 Variable").arg((*data_iter).component_name).arg((*data_iter).identifier);
			}
			else
			{
				line = QString("%1 Component %2").arg((*data_iter).component_name).arg((*data_iter).identifier);
			}
			
			d_value_csv_data.push_back(line);
			d_value_pdf_data.push_back(line);
						
			if((*data_iter).is_variable)
			{
				line = QString(",Export Variables");
			}
			else
			{
				line = QString(",Local Variables");
			}
			
			line2 = line;

			for(var_iter = d_export_variable_data.begin();var_iter != d_export_variable_data.end();++var_iter)
			{
				if((*var_iter).component_name == (*data_iter).component_name)
				{
					line += QString(",%1").arg((*var_iter).variable_name);
					line2 += QString(",%1").arg((*var_iter).variable_name);
					
					
					for(value_iter = (*var_iter).values.begin();value_iter != (*var_iter).values.end();++value_iter)
					{
						// reduce PDF precision if value doesn't fit
						
						if(fabs((*value_iter)) < 10.0)
						{
							precision = pdf_precision;
						}
						else if(fabs((*value_iter)) < 100.0)
						{
							precision = pdf_precision - 1;
						}
						else if(fabs((*value_iter)) < 1000.0)
						{
							precision = pdf_precision - 2;
						}
						else
						{
							precision = pdf_precision - 3;
						}

						line += QString(",%1").arg((*value_iter),0,'f',csv_precision);
						line2 += QString(",%1").arg((*value_iter),0,'f',precision);
					}
					
					d_value_csv_data.push_back(line);
					d_value_pdf_data.push_back(line2);
					
					line = QString(",");
					line2 = line;
				}
			}
			
			line = QString(",Source");
			line2 = line;
			
			last_component_name = (*data_iter).component_name;
		}
		
		line += QString(",%1").arg((*data_iter).component_item.component_item_name);
		line2 += QString(",%1").arg((*data_iter).component_item.component_item_name);

		for(value_iter = (*data_iter).values.begin();value_iter != (*data_iter).values.end();++value_iter)
		{
			line += QString(",%1").arg((*value_iter),0,'f',csv_precision);
			line2 += QString(",%1").arg((*value_iter),0,'f',pdf_precision);
		}
		
		d_value_csv_data.push_back(line);
		d_value_pdf_data.push_back(line2);
		
		line = QString(",");
		line2 = line;
	}
	
	d_value_csv_data.push_back(QString());
	d_value_pdf_data.push_back(QString());
	
	line = QString("Uncertainty");
	d_value_csv_data.push_back(line);
	d_value_pdf_data.push_back(line);
	
	line = QString(",,Combined");
	line2 = QString(",,Combined");
	
	for(uc_iter = d_uc_data.begin();uc_iter != d_uc_data.end();++uc_iter)
	{
		line += QString(",%1").arg((*uc_iter).y,0,'f',csv_precision);
		line2 += QString(",%1").arg((*uc_iter).y,0,'f',pdf_precision);
	}
	
	d_value_csv_data.push_back(line);
	d_value_pdf_data.push_back(line2);

	if(LeastSquares::Fit_Curve(d_uc_data,&c,&b,&a))	// default format is ax^2 + bx + c
	{
		
		line = QString(",,Uc(K=1):,%1").arg(a,0,'f',default_precision);
		line2 = QString(",,Uc(K=2):,%1").arg(a*2,0,'f',default_precision);

		if(b < 0)
		{
			line += QString("%1L").arg(b*1000,0,'f',default_precision);
			line2 += QString("%1L").arg(b*2000,0,'f',default_precision);
		}
		else
		{
			line += QString("+%1L").arg(b*1000,0,'f',default_precision);
			line2 += QString("+%1L").arg(b*2000,0,'f',default_precision);
		}
		
		if(c < 0)
		{
			line += QString("%1L^2 mm  (L is length in meters)").arg(c*1000000,0,'f',default_precision);
			line2 += QString("%1L^2 mm  (L is length in meters)").arg(c*2000000,0,'f',default_precision);
		}
		else
		{
			line += QString("+%1L^2 mm  (L is length in meters)").arg(c*1000000,0,'f',default_precision);
			line2 += QString("+%1L^2 mm  (L is length in meters)").arg(c*2000000,0,'f',default_precision);
		}
		
		d_uncertainty_csv_data.push_back(line);
		d_uncertainty_csv_data.push_back(line2);
						
		line = QString(",,Uc(K=1):,%1").arg(a,0,'f',pdf_precision);
		line2 = QString(",,Uc(K=2):,%1").arg(a*2,0,'f',pdf_precision);
		
		if(b < 0)
		{
			line += QString("%1L").arg(b*1000,0,'f',pdf_precision);
			line2 += QString("%1L").arg(b*2000,0,'f',pdf_precision);
		}
		else
		{
			line += QString("+%1L").arg(b*1000,0,'f',pdf_precision);
			line2 += QString("+%1L").arg(b*2000,0,'f',pdf_precision);
		}
		
		if(c < 0)
		{
			line += QString("%1L^2 mm  (L is length in meters)").arg(c*1000000,0,'f',pdf_precision);
			line2 += QString("%1L^2 mm  (L is length in meters)").arg(c*2000000,0,'f',pdf_precision);
		}
		else
		{
			line += QString("+%1L^2 mm  (L is length in meters)").arg(c*1000000,0,'f',pdf_precision);
			line2 += QString("+%1L^2 mm  (L is length in meters)").arg(c*2000000,0,'f',pdf_precision);
		}
		
		d_uncertainty_pdf_data.push_back(line);
		d_uncertainty_pdf_data.push_back(line2);
		
	}
	else
	{
		line = QString(",,,Uc(K=1):, ERR");
		d_uncertainty_csv_data.push_back(line);
		d_uncertainty_pdf_data.push_back(line);
		
		line = QString(",,,Uc(K=2):, ERR");
		d_uncertainty_csv_data.push_back(line);
		d_uncertainty_pdf_data.push_back(line);
	}
}

void TReportWidget::Update_Status_Display(
	const double						&length)
{
	std::vector<TReportWidget::TReportData>::const_iterator data_iter;
	std::vector<TReportWidget::TReportItemData>::const_iterator data_item_iter;
	double								sum_constant;
	double								sum_length_dependent;
	double								sum;
	double								dval;
	
	sum_constant = 0.0;
	sum_length_dependent = 0.0;
	for(data_iter = d_report_data.begin();data_iter != d_report_data.end();++data_iter)
	{
		if(!(*data_iter).is_variable)
		{
			for(data_item_iter = (*data_iter).report_items.begin();data_item_iter != (*data_iter).report_items.end();++data_item_iter)
			{
				if((*data_item_iter).value_string == "ERR")
				{
					d_status_label->setText(QString("Uc at length %1 mm: ERR").arg(length,0,'f',3));
					return;
				}
				
				dval = (*data_item_iter).value;
				
				if((*data_item_iter).value_type == "Constant")
				{
					sum_constant += (dval * dval);
				}
				else
				{
					sum_length_dependent += (dval * dval);
				}
			}
		}
	}
	
	sum_constant = sqrt(sum_constant);
	sum_length_dependent = sqrt(sum_length_dependent);
	
	sum = sqrt(sum_constant*sum_constant + sum_length_dependent*sum_length_dependent);
		
	d_status_label->setText(QString("Uc at length %1 mm: %2 mm").arg(length,0,'f',3).arg(sum,0,'f',4));
}

void TReportWidget::Update_Export_Uncertainty_Data(void)
{
	std::vector<TReportWidget::TExportData>::const_iterator data_iter;
	std::vector<TReportWidget::TExportVariableData>::const_iterator var_iter;
	TVector2							uc_value;
	double								sum_constant;
	double								sum_length_dependent;
	double								dval;
	int									position_index(0);
	bool								eod(false);
	QString								line1,line2;
	
	d_uc_data.clear();
	
	do
	{
		sum_constant = 0.0;
		sum_length_dependent = 0.0;

		for(data_iter = d_export_data.begin();data_iter != d_export_data.end();++data_iter)
		{
			if(!(static_cast<unsigned int>(position_index) < (*data_iter).positions.size()))
			{
				eod = true;
				break;
			}
			
			if(!(static_cast<unsigned int>(position_index) < (*data_iter).values.size()))
			{
				eod = true;
				break;
			}

			uc_value.x = (*data_iter).positions[position_index];

			if(!(*data_iter).is_variable)
			{
				dval = (*data_iter).values[position_index];
				
				if((*data_iter).component_item.value_type == "Constant")
				{
					sum_constant += (dval * dval);
				}
				else
				{
					sum_length_dependent += (dval * dval);
				}
			}
		}
		
		if(!eod)
		{			
			sum_constant = sqrt(sum_constant);
			sum_length_dependent = sqrt(sum_length_dependent);

			uc_value.y = sqrt(sum_constant*sum_constant + sum_length_dependent*sum_length_dependent);
						
			d_uc_data.push_back(uc_value);
		}
		
		position_index++;
		
		
	}while(!eod);
	
}

void TReportWidget::Refresh(void)
{
	this->Generate_Report();
}

void TReportWidget::Add_Export_Data(
	const TReportWidget::TExportData	&data,
	const double						&position)
{
	std::vector<TReportWidget::TExportData>::iterator iter;
	
	for(iter = d_export_data.begin();iter != d_export_data.end();++iter)
	{
		if((*iter).is_variable == data.is_variable)
		{
			if((*iter).component_name == data.component_name)
			{
				if((*iter).component_item.component_item_name == data.component_item.component_item_name)
				{
					(*iter).values.push_back(data.component_item.value);
					(*iter).positions.push_back(position);
					
					return;
				}
			}
		}
	}
	
	d_export_data.push_back(data);
	
	// add first value to values
	(*(d_export_data.end() - 1)).values.clear();
	(*(d_export_data.end() - 1)).values.push_back(data.component_item.value);
	
	(*(d_export_data.end() - 1)).positions.clear();
	(*(d_export_data.end() - 1)).positions.push_back(position);
	
}

QStringList TReportWidget::Split_CSV_Data(
	const QString						&data) const
{
	QStringList							list;
	QChar								character;
	int									cntr,index;
	bool								inside_quote;
	QString								item;
	
	index = 0;
	inside_quote = false;
		
	for(cntr = 0;cntr < data.size();++cntr)
	{
		character = data[cntr].toLatin1();
		
		if(inside_quote == false && character == ',')
		{
			if(cntr > index)
			{
				item = data.mid(index,cntr - index);
			}
			else
			{
				item = QString();
			}
			list.push_back(item);
			index = cntr + 1;
		}
		
		if(character == '\"')
		{
			inside_quote = !inside_quote;
		}
	}
	
	if(cntr > index)
	{
		item = data.mid(index);
		list.push_back(item);
	}
	
	return list;
}
