/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////



#ifndef REPORTWIDGETHEADERFILE
#define REPORTWIDGETHEADERFILE

#include <QWidget>
#include <QStringList>
#include <QString>
#include <vector>

#include "../../core/vector2.h"

class QComboBox;
class QPushButton;
class QLabel;
class TOptionEditorWidget;

class TReportWidget : public QWidget
{
Q_OBJECT
public:

struct TReportItemData
{
	QString								component_item_name;
	QString								value_type;
	QString								expression;
	QString								sensitivity;
	QString								distribution;
	QString								type;
	double								distribution_divisor;
	QString								value_string;
	double								value;
};
	
struct TReportVariable
{
	QString								name;
	double								value;
};
	
struct TReportData
{
	QString								component_name;
	QString								identifier;
	QStringList							identifier_list;
	int									identifier_list_index;
	bool								is_variable;
	std::vector<TReportItemData>		report_items;
	std::vector<TReportVariable>		report_variables;
};
	
struct TExportData
{
	QString								component_name;
	TReportItemData						component_item;
	QString								identifier;
	bool								is_variable;
	std::vector<double>					values;
	std::vector<double>					positions;
};
	
struct TExportVariableData
{
	QString								component_name;
	QString								variable_name;
	std::vector<double>					values;
};
	
enum TExportFormat
{
	EXPORT_PDF_FORMAT = 0,
	EXPORT_CSV_FORMAT
};
	
// CREATORS
	TReportWidget(const QWidget *parent=0,Qt::WindowFlags flags=Qt::WindowFlags());
	~TReportWidget(void);
	
// ACCESSORS
	QString Current_Identifier(const QString &component_name) const;
	bool Write_Export_CSV_Data(const QString &file_name) const;
	bool Write_Export_PDF_Data(const QString &file_name) const;
	
	TExportFormat Export_Format(void) const {return d_export_format;}
	
	QString Last_Error(void) const {return d_last_error;}

// MANIPULATORS
	void Update_Report_Groups(const QStringList &list);
	
	void Clear_Report(void);
	void Set_Default_Report_Group(void);
	
	void Start_Report_Update(void);
	void Add_Report_Data(const TReportData &data);
	void End_Report_Update(void);
	
	void Clear_Export(void);
	void Add_Global_Variable(const TReportVariable &variable);
	void Add_Export_Data(const TReportData &data,const double &position);
	void Add_Export_Variable_Data(const TExportVariableData &data);
	void Format_Export_Data(void);
	void Update_Status_Display(const double &length);
	void Update_Export_Uncertainty_Data(void);
	
public slots:
	void Refresh(void);
	
signals:
	void Create_Report(const QString&);		// name of report group
	void Create_Export_Report(const QString&);		// name of report group
	
private slots:
	void Generate_PDF(void);
	void Generate_CSV(void);
	void Generate_Report(void);
	void Identifier_Selection_Changed(int,const QString&);
	
private:
	
	QComboBox							*d_reportgroup_combo;
	TOptionEditorWidget					*d_report_table;
	QPushButton							*d_export_pdf_button;
	QPushButton							*d_export_csv_button;
	QLabel								*d_status_label;
	
	bool								d_block_signals_report_update;
	
	std::vector<TReportData>			d_report_data;
	std::vector<TExportData>			d_export_data;
	std::vector<TExportVariableData>	d_export_variable_data;
	
	std::vector<TVector2>				d_uc_data;
	
	std::vector<QString>				d_item_data;
	std::vector<QString>				d_variable_data;
	std::vector<QString>				d_value_csv_data;
	std::vector<QString>				d_value_pdf_data;
	std::vector<QString>				d_uncertainty_csv_data;
	std::vector<QString>				d_uncertainty_pdf_data;
	
	TExportFormat						d_export_format;
	
	mutable QString						d_last_error;

	void Add_Export_Data(const TExportData &data,const double &position);
	QStringList Split_CSV_Data(const QString &data) const;
	
// NOT IMPLEMENTED
	TReportWidget(const TReportWidget&);
	TReportWidget& operator=(const TReportWidget&);
};


#endif
