/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////



#include <QDate>
#include <math.h>
#include <assert.h>

#include "../../core/xmlfile.h"

#include "ucbudget.h"

TUCBudget::TUCBudget(void)
{
}

TUCBudget::~TUCBudget(void)
{
}

QStringList TUCBudget::Enum_Components(void) const
{
	QStringList							list;
	int									count;
	
	count = d_budget_data.Component_Count();

	for(int i=0;i< count;++i)
	{
		list.push_back(d_budget_data.Component_Name(i));
	}
	
	return list;
}

QStringList TUCBudget::Enum_Component_Items(
	const QString						&component_name) const
{
	QStringList							list;
	int									component_index;
	int									item_count;
	
	component_index = d_budget_data.Component_Index(component_name);
	
	if(!(component_index < 0))
	{
		item_count = d_budget_data.Component_Item_Count(component_index);
		
		for(int i = 0;i < item_count;++i)
		{
			list.push_back(d_budget_data.Component_Item_Name(component_index,i));
		}
	}
	
	return list;
}

QStringList TUCBudget::Enum_Component_Identifiers(
	const QString						&component_name) const
{
	QStringList							list;
	int									component_index;
	int									item_count;
	
	component_index = d_budget_data.Component_Index(component_name);
	
	if(!(component_index < 0))
	{
		item_count = d_budget_data.Component_Identifier_Count(component_index);
		
		for(int i = 0;i < item_count;++i)
		{
			list.push_back(d_budget_data.Component_Identifier_Name(component_index,i));
		}
	}
	
	return list;
}

QStringList TUCBudget::Enum_Reportable_Components(void) const
{
	QStringList							list;
	int									count;
	TUCBudgetData::TComponent			component;
	std::vector<TUCBudgetData::TComponentItem>::const_iterator item_iter;

	count = d_budget_data.Component_Count();
	
	for(int i=0;i< count;++i)
	{
		component = d_budget_data.Component(i);
		
		if(component.mode == TUCBudgetData::SUM_ITEMS)
		{
			list.push_back(component.name);
		}
		else
		{
			for(item_iter = component.items.begin();item_iter != component.items.end();++item_iter)
			{
				list.push_back(QString("%1 - %2").arg(component.name).arg((*item_iter).name));			// string format used in 'Convert_From_Reportable'
			}
		}		
	}
	
	return list;
}

bool TUCBudget::Component_Item_Default_Expression(
	const QString						&component_name,
	const QString						&component_item_name,
	QString								*default_expression) const
{
	int									component_id;
	int									component_item_id;
	TUCBudgetData::TComponentItem		component_item;
	
	assert(default_expression);
	
	component_id = d_budget_data.Component_Index(component_name);
	
	if(component_id < 0)
	{
		return false;
	}
	
	component_item_id = d_budget_data.Component_Item_Index(component_id,component_item_name);	
	
	if(component_item_id < 0)
	{
		return false;
	}
	component_item = d_budget_data.Component_Item(component_id,component_item_id);
	
	
	*default_expression = component_item.expression;
	
	
	return true;
}

bool TUCBudget::Component_Item_Expression(
	const QString						&component_name,
	const QString						&component_item_name,
	const QString						&identifier,
	bool								*use_default,
	QString								*replacement_expression) const
{
	int									component_id;
	int									component_item_id;
	TUCBudgetData::TComponentItem		component_item;
	TUCBudgetData::TReplacementItem		replacement_item;
	int									cntr;
	
	assert(use_default);
	assert(replacement_expression);
	
	component_id = d_budget_data.Component_Index(component_name);
	
	if(component_id < 0)
	{
		return false;
	}

	component_item_id = d_budget_data.Component_Item_Index(component_id,component_item_name);
	
	if(component_item_id < 0)
	{
		return false;
	}

	component_item = d_budget_data.Component_Item(component_id,component_item_id);
	
	*replacement_expression = component_item.expression;
	*use_default = false;
	
	for(cntr = 0;cntr < static_cast<int>(component_item.replacement_items.size());++cntr)
	{
		if(component_item.replacement_items[cntr].identifier == identifier)
		{
			*use_default = component_item.replacement_items[cntr].use_default_expression;
			*replacement_expression = component_item.replacement_items[cntr].expression;
			break;
		}
	}
		
		
	return true;
}

double TUCBudget::Component_Length(
	const QString						&component_name,
	const QString						&identifier) const
{
	TUCBudgetData::TComponent			component;
	int									index;
	std::vector<TUCBudgetData::TComponentInstance>::const_iterator id_iter;
	
	index = d_budget_data.Component_Index(component_name);
	component = d_budget_data.Component(index);
	
	if(identifier.length())
	{
		for(id_iter = component.instances.begin();id_iter != component.instances.end();++id_iter)
		{
			if((*id_iter).name == identifier)
			{
				return (*id_iter).length;
			}
		}
	}
	
	return component.length;
}

double TUCBudget::Component_Expansion_Coefficient(
	const QString						&component_name,
	const QString						&identifier) const
{
	TUCBudgetData::TComponent			component;
	int									index;
	std::vector<TUCBudgetData::TComponentInstance>::const_iterator id_iter;
	
	index = d_budget_data.Component_Index(component_name);
	component = d_budget_data.Component(index);
		
	if(identifier.length())
	{
		for(id_iter = component.instances.begin();id_iter != component.instances.end();++id_iter)
		{
			if((*id_iter).name == identifier)
			{
				return (*id_iter).expansion_coefficient;
			}
		}
	}
	
	return component.expansion_coefficient;

}

QString TUCBudget::Component_Item_Expression(
	const QString						&component_name,
	const QString						&component_item_name,
	const QString						&identifier,
	bool								*valid) const
{
	bool								use_default;
	QString								expression;
	QString								replacement_expression;

	if(valid)
	{
		*valid = false;
	}
	
	if(this->Component_Item_Default_Expression(component_name,
													component_item_name,
													&expression))
	{
		if(this->Component_Item_Expression(component_name,
										component_item_name,
										identifier,
										&use_default,
										&replacement_expression))
		{
			if(!use_default)
			{
				expression = replacement_expression;
			}
			
			if(valid)
			{
				*valid = true;
			}
		}
	}

	return expression;
}


bool TUCBudget::Convert_From_Reportable(
	const QString						&name,
	QString								*component_name,
	QString								*component_item_name) const
{
	TUCBudgetData::TComponent			component;
	QString								text;
	int									index;
	int									item_index;
	
	assert(component_name);
	assert(component_item_name);
	
	for(index = 0;index < d_budget_data.Component_Count();++index)
	{
		component = d_budget_data.Component(index);
		
		if(component.mode == TUCBudgetData::INDIVIUAL_ITEM)
		{
			for(item_index = 0;item_index < d_budget_data.Component_Item_Count(index);++item_index)
			{
				text = QString("%1 - %2").arg(component.name).arg(d_budget_data.Component_Item_Name(index,item_index));
				
				if(text == name)
				{
					*component_item_name = d_budget_data.Component_Item_Name(index,item_index);
					*component_name = component.name;
					
					return true;
				}
			}
		}
		else
		{
			if(component.name == name)
			{
				*component_item_name = QString();
				*component_name = name;
				
				return true;
			}
		}
	}
	
	return false;
}

QStringList TUCBudget::Distribution_Type_List(void) const
{
	QStringList							list;
	
	list.push_back("Normal");
	list.push_back("Rectangular");
	list.push_back("Triangular");
	list.push_back("U");
	list.push_back("Step");
	
	return list;
}

QStringList TUCBudget::Uncertainty_Type_List(void) const
{
	QStringList							list;
	
	list.push_back("Type A");
	list.push_back("Type B");
	
	return list;
}

QString TUCBudget::Distribution_Type(
	const TUCBudgetData::TDistribution	type) const
{
	switch(type)
	{
		case TUCBudgetData::DISTRIBUTION_NORMAL:
			default:
			break;
		case TUCBudgetData::DISTRIBUTION_RECTANGULAR:
			return QString("Rectangular");
		case TUCBudgetData::DISTRUBUTION_TRIANGULAR:
			return QString("Triangular");
		case TUCBudgetData::DISTRIBUTION_U:
			return QString("U");
		case TUCBudgetData::DISTRIBUTION_STEP:
			return QString("Step");
	}
	
	return QString("Normal");
}

TUCBudgetData::TDistribution TUCBudget::Distribution_Type(
	const QString						&type) const
{
	if(type == "Rectangular")
	{
		return TUCBudgetData::DISTRIBUTION_RECTANGULAR;
	}
	else if(type == "Triangular")
	{
		return TUCBudgetData::DISTRUBUTION_TRIANGULAR;
	}
	else if(type == "U")
	{
		return TUCBudgetData::DISTRIBUTION_U;
	}
	else if(type == "Step")
	{
		return TUCBudgetData::DISTRIBUTION_STEP;
	}
	
	return TUCBudgetData::DISTRIBUTION_NORMAL;
}

QString TUCBudget::UCType(
	const TUCBudgetData::TUCType		type) const
{
	switch(type)
	{
		case TUCBudgetData::TYPE_A:
			default:
			break;
		case TUCBudgetData::TYPE_B:
			return QString("Type B");
	}
	
	return QString("Type A");
}

TUCBudgetData::TUCType TUCBudget::UCType(
	const QString						&type) const
{
	if(type == "Type B")
	{
		return TUCBudgetData::TYPE_B;
	}

	return TUCBudgetData::TYPE_A;
}

TUCBudgetData::TUCMode TUCBudget::UCMode(
	const QString						&mode) const
{
	if(mode == "Individual")
	{
		return TUCBudgetData::INDIVIUAL_ITEM;
	}
	
	return TUCBudgetData::SUM_ITEMS;

}

QString TUCBudget::UCMode(
	const TUCBudgetData::TUCMode		mode) const
{
	if(mode == TUCBudgetData::INDIVIUAL_ITEM)
	{
		return QString("Individual");
	}
	
	return QString("Sum");
}


QString TUCBudget::Expression_Type(
	const TUCBudgetData::TExpressionType	type) const
{
	switch(type)
	{
		case TUCBudgetData::CONSTANT_LINEAR:
			default:
			break;
		case TUCBudgetData::CONSTANT_TEMPERATURE:
			return QString("Constant Temperature");
		case TUCBudgetData::LENGTH_DEPENDENT_LINEAR:
			return QString("Length Dependent Linear");
		case TUCBudgetData::LENGTH_DEPENDENT_TEMPERATURE:
			return QString("Length Dependent Temperature");
	}
	
	return QString("Constant Linear");
}

TUCBudgetData::TExpressionType TUCBudget::Expression_Type(
	const QString						&type) const
{
	if(type == "Constant Temperature")
	{
		return TUCBudgetData::CONSTANT_TEMPERATURE;
	}
	else if(type == "Length Dependent Linear")
	{
		return TUCBudgetData::LENGTH_DEPENDENT_LINEAR;
	}
	else if(type == "Length Dependent Temperature")
	{
		return TUCBudgetData::LENGTH_DEPENDENT_TEMPERATURE;
	}
	
	return TUCBudgetData::CONSTANT_LINEAR;
}

bool TUCBudget::Boolean_Type(
	const QString						&yes_no) const
{
	if(yes_no == "Yes")
		return true;
	
	return false;
}

QString TUCBudget::Boolean_Type(
	const bool							yes_no) const
{
	if(yes_no)
		return QString("Yes");
	
	return QString("No");
}

double TUCBudget::Distribution_Divisor(
	const TUCBudgetData::TDistribution	type) const
{
	switch(type)
	{
		case TUCBudgetData::DISTRIBUTION_NORMAL:
			return 1.0;
		case TUCBudgetData::DISTRIBUTION_RECTANGULAR:
			return 1.732050807568877;
		case TUCBudgetData::DISTRUBUTION_TRIANGULAR:
			return 2.449489742783178;
		case TUCBudgetData::DISTRIBUTION_U:
			return 1.414213562373095;
		case TUCBudgetData::DISTRIBUTION_STEP:
			return 3.46410161513775;
	}
	return 1.0;
}

bool TUCBudget::Save(
	const QString						&file_name,
	const std::vector<TVariable>		&variables,
	const std::vector<TGroup>			&groups,
	const TInputData					&input_data) const
{
	TXmlFile							xml_file;
	QDomElement							node;
	QDomElement							node_level_a;
	QDomElement							node_level_b;
	QDomElement							node_level_c;
	QDomElement							node_level_d;
	QDomElement							node_level_e;
	QDomElement							node_level_f;
	QString								text;
	int									index;
	int									item_index;
	int									repl_index;
	TUCBudgetData::TComponent			component;
	TUCBudgetData::TComponentItem		component_item;
	TUCBudgetData::TComponentInstance	component_instance;
	TUCBudgetData::TReplacementItem		replacement_item;
	QStringList::ConstIterator			str_iter;


	try
	{
		xml_file.Set_File_Name(file_name);
		xml_file.Reset("UCBudget_Data",3);
		
		xml_file.Write_Attribute(xml_file.Root_Node(),"File_Revision",QDate::currentDate().toString("MMM dd yyyy"));
		
		node = xml_file.Create_Node(xml_file.Root_Node(),"Components");

		for(index = 0;index < d_budget_data.Component_Count();++index)
		{
			component = d_budget_data.Component(index);
			
			node_level_a = xml_file.Create_Node(&node,"Component");
			
			node_level_b = xml_file.Create_Node(&node_level_a,"Name");
			xml_file.Write_Text_Node(&node_level_b,component.name.toHtmlEscaped());
			node_level_b = xml_file.Create_Node(&node_level_a,"Mode");
			xml_file.Write_Text_Node(&node_level_b,this->UCMode(component.mode).toHtmlEscaped());
			node_level_b = xml_file.Create_Node(&node_level_a,"Export_Equipment");
			xml_file.Write_Text_Node(&node_level_b,this->Boolean_Type(component.export_equipment).toHtmlEscaped());
			node_level_b = xml_file.Create_Node(&node_level_a,"Limit_Uncertainty_Length");
			xml_file.Write_Text_Node(&node_level_b,this->Boolean_Type(component.limit_uncertainty_length).toHtmlEscaped());
			node_level_b = xml_file.Create_Node(&node_level_a,"Length");
			xml_file.Write_Text_Node(&node_level_b,QString("%1").arg(component.length,0,'f',6).toHtmlEscaped());
			node_level_b = xml_file.Create_Node(&node_level_a,"Expansion_Coefficient");
			xml_file.Write_Text_Node(&node_level_b,QString("%1").arg(component.expansion_coefficient * 1000,0,'f',2).toHtmlEscaped());
			
			node_level_b = xml_file.Create_Node(&node_level_a,"Identifiers");

			for(item_index = 0;item_index < d_budget_data.Component_Identifier_Count(index);++item_index)
			{
				component_instance = d_budget_data.Component_Instance(index,item_index);
				
				node_level_c = xml_file.Create_Node(&node_level_b,"Identifier");
				
				node_level_d = xml_file.Create_Node(&node_level_c,"Name");
				xml_file.Write_Text_Node(&node_level_d,component_instance.name.toHtmlEscaped());
				node_level_d = xml_file.Create_Node(&node_level_c,"Calibration_Date");
				xml_file.Write_Text_Node(&node_level_d,component_instance.calibration_date.toHtmlEscaped());
				node_level_d = xml_file.Create_Node(&node_level_c,"Calibration_Due_Date");
				xml_file.Write_Text_Node(&node_level_d,component_instance.calibration_due_date.toHtmlEscaped());
				node_level_d = xml_file.Create_Node(&node_level_c,"Length");
				xml_file.Write_Text_Node(&node_level_d,QString("%1").arg(component_instance.length,0,'f',6).toHtmlEscaped());
				node_level_d = xml_file.Create_Node(&node_level_c,"Expansion_Coefficient");
				xml_file.Write_Text_Node(&node_level_d,QString("%1").arg(component_instance.expansion_coefficient * 1000,0,'f',2).toHtmlEscaped());
			}


			node_level_b = xml_file.Create_Node(&node_level_a,"Component_Items");
			
			for(item_index = 0;item_index < d_budget_data.Component_Item_Count(index);++item_index)
			{
				component_item = d_budget_data.Component_Item(index,item_index);
				
				node_level_c = xml_file.Create_Node(&node_level_b,"Item");
				
				node_level_d = xml_file.Create_Node(&node_level_c,"Name");
				xml_file.Write_Text_Node(&node_level_d,component_item.name.toHtmlEscaped());
				node_level_d = xml_file.Create_Node(&node_level_c,"Expression");
				xml_file.Write_Text_Node(&node_level_d,component_item.expression.toHtmlEscaped());
				node_level_d = xml_file.Create_Node(&node_level_c,"Distribution");
				xml_file.Write_Text_Node(&node_level_d,this->Distribution_Type(component_item.distribution).toHtmlEscaped());
				node_level_d = xml_file.Create_Node(&node_level_c,"Expression_Type");
				xml_file.Write_Text_Node(&node_level_d,this->Expression_Type(component_item.expression_type).toHtmlEscaped());
				node_level_d = xml_file.Create_Node(&node_level_c,"Uncertainty_Type");
				xml_file.Write_Text_Node(&node_level_d,this->UCType(component_item.uctype).toHtmlEscaped());
				node_level_d = xml_file.Create_Node(&node_level_c,"Sensitivity");
				xml_file.Write_Text_Node(&node_level_d,QString("%1").arg(component_item.sensitivity,0,'f',6).toHtmlEscaped());

				node_level_d = xml_file.Create_Node(&node_level_c,"Replacement_Items");
				
				for(repl_index = 0;repl_index < static_cast<int>(component_item.replacement_items.size());++repl_index)
				{
					node_level_e = xml_file.Create_Node(&node_level_d,"Item");

					node_level_f = xml_file.Create_Node(&node_level_e,"Identifier");
					xml_file.Write_Text_Node(&node_level_f,component_item.replacement_items[repl_index].identifier.toHtmlEscaped());
					node_level_f = xml_file.Create_Node(&node_level_e,"Expression");
					xml_file.Write_Text_Node(&node_level_f,component_item.replacement_items[repl_index].expression.toHtmlEscaped());
					node_level_f = xml_file.Create_Node(&node_level_e,"Use_Default_Expression");
					xml_file.Write_Text_Node(&node_level_f,this->Boolean_Type(component_item.replacement_items[repl_index].use_default_expression).toHtmlEscaped());

				}
			}
		}
		
		node = xml_file.Create_Node(xml_file.Root_Node(),"Variables");
		
		for(index = 0;index < static_cast<int>(variables.size());++index)
		{
			
			node_level_a = xml_file.Create_Node(&node,"Variable");
			
			node_level_b = xml_file.Create_Node(&node_level_a,"Name");
			xml_file.Write_Text_Node(&node_level_b,variables[index].name.toHtmlEscaped());
			node_level_b = xml_file.Create_Node(&node_level_a,"Description");
			xml_file.Write_Text_Node(&node_level_b,variables[index].description.toHtmlEscaped());
		}
		
		node = xml_file.Create_Node(xml_file.Root_Node(),"Report_Groups");
		
		for(index = 0;index < static_cast<int>(groups.size());++index)
		{
			
			node_level_a = xml_file.Create_Node(&node,"Group");
			
			node_level_b = xml_file.Create_Node(&node_level_a,"Name");
			xml_file.Write_Text_Node(&node_level_b,groups[index].name.toHtmlEscaped());
			
			node_level_b = xml_file.Create_Node(&node_level_a,"Items");
			
			for(str_iter = groups[index].items.begin();str_iter != groups[index].items.end();++str_iter)
			{
				node_level_c = xml_file.Create_Node(&node_level_b,"Item");
				xml_file.Write_Text_Node(&node_level_c,(*str_iter).toHtmlEscaped());
			}
		}
		
		node = xml_file.Create_Node(xml_file.Root_Node(),"User_Input");
		
		node_level_a = xml_file.Create_Node(&node,"Machine_Expansion_Coefficient");
		xml_file.Write_Text_Node(&node_level_a,QString("%1").arg(input_data.machine_expansion_coefficient,0,'f',2).toHtmlEscaped());
		node_level_a = xml_file.Create_Node(&node,"Machine_Scale");
		xml_file.Write_Text_Node(&node_level_a,QString("%1").arg(input_data.machine_scale,0,'f',6).toHtmlEscaped());
		node_level_a = xml_file.Create_Node(&node,"Temperature_Minimum");
		xml_file.Write_Text_Node(&node_level_a,QString("%1").arg(input_data.temp_min,0,'f',2).toHtmlEscaped());
		node_level_a = xml_file.Create_Node(&node,"Temperature_Maximum");
		xml_file.Write_Text_Node(&node_level_a,QString("%1").arg(input_data.temp_max,0,'f',2).toHtmlEscaped());
		node_level_a = xml_file.Create_Node(&node,"Temperature_Gradient");
		xml_file.Write_Text_Node(&node_level_a,QString("%1").arg(input_data.temp_gradient,0,'f',2).toHtmlEscaped());
		node_level_a = xml_file.Create_Node(&node,"Report_Start");
		xml_file.Write_Text_Node(&node_level_a,QString("%1").arg(0.0,0,'f',3).toHtmlEscaped());
		node_level_a = xml_file.Create_Node(&node,"Report_Increment");
		xml_file.Write_Text_Node(&node_level_a,QString("%1").arg(input_data.report_increment,0,'f',3).toHtmlEscaped());
		node_level_a = xml_file.Create_Node(&node,"Report_End");
		xml_file.Write_Text_Node(&node_level_a,QString("%1").arg(input_data.report_end,0,'f',3).toHtmlEscaped());
		
		xml_file.Write_File();
	}
	catch(const TXmlFileException &xml_file_exception)
	{
		d_last_error = xml_file_exception.ErrorText();
		return false;
	}
	
	return true;
}

bool TUCBudget::Update_Component_Item(
	const QString						&component_name,
	const QString						&component_item_name,
	const QString						&identifier,
	const QString						&expression,
	const bool							use_default)
{
	int									component_id;
	int									component_item_id;
	TUCBudgetData::TComponentItem		component_item;
	int									cntr;
	bool								found_item;
	
	component_id = d_budget_data.Component_Index(component_name);
	
	if(component_id < 0)
	{
		return false;
	}
	
	component_item_id = d_budget_data.Component_Item_Index(component_id,component_item_name);	
	
	if(component_item_id < 0)
	{
		return false;
	}
	component_item = d_budget_data.Component_Item(component_id,component_item_id);
	
	found_item = false;
	if(identifier.length())
	{
		for(cntr = 0;cntr < static_cast<int>(component_item.replacement_items.size());++cntr)
		{
			if(component_item.replacement_items[cntr].identifier == identifier)
			{
				component_item.replacement_items[cntr].expression = expression;
				component_item.replacement_items[cntr].use_default_expression = use_default;
				
				found_item = true;
				
				break;
			}
		}
	}
	
	if(!found_item)
	{
		// update default
		component_item.expression = expression;
	}
	
	d_budget_data.Update_Component_Item(component_id,component_item);
		
	return true;
}

bool TUCBudget::Load(
	const QString						&file_name,
	std::vector<TVariable>				*variables,
	std::vector<TGroup>					*groups,
	TInputData							*input_data)
{
	TXmlFile							xml_file;
	int									version;
	
	assert(variables);
	assert(groups);
	assert(input_data);

	try
	{		
		xml_file.Set_File_Name(file_name);
		xml_file.Open("UCBudget_Data",&version);
		
		switch(version)
		{
			case 1:
				this->LoadV1(&xml_file,variables,groups,input_data);
				break;
				
			case 2:
				this->LoadV2(&xml_file,variables,groups,input_data);
				break;
				
			case 3:
				this->LoadV3(&xml_file,variables,groups,input_data);
				break;
				
			default:
				d_last_error = "UCBudget_Data file version not supported"; 
				return false;
		}
		
	}
	catch(const TXmlFileException &xml_file_exception)
	{
		d_last_error = xml_file_exception.ErrorText();
		return false;
	}
	
	return true;
}

void TUCBudget::Reset_Expression_Variables(void)
{
	d_expression.Clear_Variables();
}

void TUCBudget::Add_Expression_Variable(
	const QString						&name,
	const double						&value)
{
	d_expression.Add_Variable(name,value);
}

double TUCBudget::Evaluate_Reportable_Component(
	const QString						&description,
	const QString						&identifier,
	const double						&length,
	bool								*valid)
{
	QStringList							reportable_components;
	QStringList							components;
	QStringList::ConstIterator			str_iter;
	QString								component_name;
	QString								component_item_name;
	bool								result;
	QString								expression;
	int									component_id;
	int									item_count;
	int									item_index;
	double								dval;
	double								constant;
	double								length_dependent;
	double								component_length;
	double								distribution_divisor;
	TUCBudgetData::TComponentItem		component_item;
	TUCBudgetData::TComponent			component;
	
	assert(valid);
	
	*valid = false;
	
	if(this->Convert_From_Reportable(description,&component_name,&component_item_name))
	{
		if(component_item_name.length())
		{
			return this->Evaluate_Component_Item(component_name,component_item_name,identifier,length,valid);
		}
		else
		{
			component_id = d_budget_data.Component_Index(component_name);
			
			item_count = d_budget_data.Component_Item_Count(component_id);
			
			constant = 0;
			length_dependent = 0;

			if(item_count)
			{
				component = d_budget_data.Component(component_id);

				for(item_index = 0;item_index < item_count;++item_index)
				{
					expression = this->Component_Item_Expression(component_name,d_budget_data.Component_Item_Name(component_id,item_index),identifier,&result);
					
					if(result)
					{
						dval = d_expression.Evaluate(expression,&result);
												
						if(result)
						{
							component_item = d_budget_data.Component_Item(component_id,item_index);
							
							dval *= component_item.sensitivity;
							
							distribution_divisor = this->Distribution_Divisor(component_item.distribution);
							dval /= distribution_divisor;

							if(component_item.expression_type == TUCBudgetData::CONSTANT_LINEAR || component_item.expression_type == TUCBudgetData::CONSTANT_TEMPERATURE)
							{
								constant += (dval * dval);
							}
							else
							{
								if(component.limit_uncertainty_length)
								{
									component_length = this->Component_Length(component_name,identifier);
									
									if(component_length < length)
									{
										dval *= (component_length / 1000);
									}
									else
									{
										dval *= (length / 1000);
									}
								}
								else
								{
									dval *= (length / 1000);
								}
								
								length_dependent += (dval * dval);
							}
						}
					}
				}
				
				constant = sqrt(constant);
				length_dependent = sqrt(length_dependent);
				
				dval = sqrt(constant * constant + length_dependent * length_dependent);
				
				*valid = true;
				return dval;
			}
		}
	}
	
	*valid = false;
	return 0.0;
}

double TUCBudget::Evaluate_Component_Item(
	const QString						&component_name,
	const QString						&component_item_name,
	const QString						&identifier,
	const double						&length,
	bool								*valid)
{
	bool								result;
	QString								expression;
	int									component_id;
	double								dval;
	double								constant;
	double								length_dependent;
	double								distribution_divisor;
	double								component_length;
	TUCBudgetData::TComponentItem		component_item;
	TUCBudgetData::TComponent			component;
	
	assert(valid);
	
	*valid = false;
	
	component_id = d_budget_data.Component_Index(component_name);

	constant = 0;
	length_dependent = 0;

	expression = this->Component_Item_Expression(component_name,component_item_name,identifier,&result);

	if(result)
	{
		dval = d_expression.Evaluate(expression,&result);

		if(result)
		{
			component = d_budget_data.Component(component_id);
			component_item = d_budget_data.Component_Item(component_id,component_item_name);
			
			dval *= component_item.sensitivity;

			distribution_divisor = this->Distribution_Divisor(component_item.distribution);
			dval /= distribution_divisor;

			if(component_item.expression_type == TUCBudgetData::CONSTANT_LINEAR || component_item.expression_type == TUCBudgetData::CONSTANT_TEMPERATURE)
			{
				constant = dval;	// only one value is expected.  No reason to square -> square_root
			}
			else
			{
				if(component.limit_uncertainty_length)
				{
					component_length = this->Component_Length(component_name,identifier);
					
					if(component_length < length)
					{
						length_dependent = dval * component_length / 1000;
					}
					else
					{
						length_dependent = dval * length / 1000;
					}
				}
				else
				{
					length_dependent = dval * length / 1000;
				}
			}

			dval = constant + length_dependent;	// only one of the two values are expected

			*valid = true;
			return dval;
		}
	}

	return 0.0;
}


void TUCBudget::LoadV1(
	TXmlFile							*xml_file,
	std::vector<TVariable>				*variables,
	std::vector<TGroup>					*groups,
	TInputData							*input_data)
{
	QDomElement							node;
	QDomElement							node_lev_a;
	QDomElement							node_lev_b;
	QDomElement							node_lev_c;
	QDomElement							node_lev_d;
	QDomElement							node_lev_e;
	QDomElement							node_lev_f;
	TUCBudgetData::TComponent			component;
	TUCBudgetData::TComponentItem		component_item;
	TUCBudgetData::TComponentInstance	component_instance;
	TUCBudgetData::TReplacementItem		replacement_item;
	TVariable							variable;
	TGroup								group;
	
	assert(xml_file);
	
	d_budget_data.Clear();
	
	node = xml_file->Get_Node(xml_file->Root_Node(),"Components");

	node_lev_a = xml_file->Get_Node(&node,"Component",true);

	while(!node_lev_a.isNull())
	{		
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Name");
		component.name = xml_file->Read_Text_Node(&node_lev_b);
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Mode");
		component.mode = this->UCMode(xml_file->Read_Text_Node(&node_lev_b));
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Export_Equipment");
		component.export_equipment = this->Boolean_Type(xml_file->Read_Text_Node(&node_lev_b));
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Limit_Uncertainty_Length");
		component.limit_uncertainty_length = this->Boolean_Type(xml_file->Read_Text_Node(&node_lev_b));		
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Length");
		component.length = xml_file->Read_Text_Node(&node_lev_b).toDouble();
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Expansion_Coefficient");
		component.expansion_coefficient = xml_file->Read_Text_Node(&node_lev_b).toDouble() / 1000;
		
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Identifiers");
		node_lev_c = xml_file->Get_Node(&node_lev_b,"Identifier",true);
		
		component.instances.clear();

		while(!node_lev_c.isNull())
		{
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Name");
			component_instance.name = xml_file->Read_Text_Node(&node_lev_d);
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Calibration_Date");
			component_instance.calibration_date = xml_file->Read_Text_Node(&node_lev_d);
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Calibration_Due_Date");
			component_instance.calibration_due_date = xml_file->Read_Text_Node(&node_lev_d);
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Length");
			component_instance.length = xml_file->Read_Text_Node(&node_lev_d).toDouble();
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Expansion_Coefficient");
			component_instance.expansion_coefficient = xml_file->Read_Text_Node(&node_lev_d).toDouble() / 1000;
						
			component.instances.push_back(component_instance);

			node_lev_c = xml_file->Get_Sibling_Node(&node_lev_c,"Identifier",true);
		}

		node_lev_b = xml_file->Get_Node(&node_lev_a,"Component_Items");
		node_lev_c = xml_file->Get_Node(&node_lev_b,"Item",true);
		
		component.items.clear();

		while(!node_lev_c.isNull())
		{
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Name");
			component_item.name = xml_file->Read_Text_Node(&node_lev_d);
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Expression");
			component_item.expression = xml_file->Read_Text_Node(&node_lev_d);
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Distribution");
			component_item.distribution = this->Distribution_Type(xml_file->Read_Text_Node(&node_lev_d));
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Expression_Type");
			component_item.expression_type = this->Expression_Type(xml_file->Read_Text_Node(&node_lev_d));
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Uncertainty_Type");
			component_item.uctype = this->UCType(xml_file->Read_Text_Node(&node_lev_d));
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Sensitivity");
			component_item.sensitivity = xml_file->Read_Text_Node(&node_lev_d).toDouble();
			
		
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Replacement_Items");
			node_lev_e = xml_file->Get_Node(&node_lev_d,"Item",true);
			
			component_item.replacement_items.clear();
			
			while(!node_lev_e.isNull())
			{
				node_lev_f = xml_file->Get_Node(&node_lev_e,"Identifier");
				replacement_item.identifier = xml_file->Read_Text_Node(&node_lev_f);
				node_lev_f = xml_file->Get_Node(&node_lev_e,"Expression");
				replacement_item.expression = xml_file->Read_Text_Node(&node_lev_f);
				node_lev_f = xml_file->Get_Node(&node_lev_e,"Use_Default_Expression");
				replacement_item.use_default_expression = this->Boolean_Type(xml_file->Read_Text_Node(&node_lev_f));
				
				component_item.replacement_items.push_back(replacement_item);
				
				node_lev_e = xml_file->Get_Sibling_Node(&node_lev_e,"Item",true);
			}
	
			component.items.push_back(component_item);
			
			node_lev_c = xml_file->Get_Sibling_Node(&node_lev_c,"Item",true);
		}
		
		d_budget_data.Add_Component(component);

		node_lev_a = xml_file->Get_Sibling_Node(&node_lev_a,"Component",true);
	}
	
	node = xml_file->Get_Node(xml_file->Root_Node(),"Variables");
	variables->clear();
	
	node_lev_a = xml_file->Get_Node(&node,"Variable",true);
	
	while(!node_lev_a.isNull())
	{		
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Name");
		variable.name = xml_file->Read_Text_Node(&node_lev_b);
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Description");
		variable.description = xml_file->Read_Text_Node(&node_lev_b,true);
		variable.editor_index = -1;
		
		variables->push_back(variable);
		
		node_lev_a = xml_file->Get_Sibling_Node(&node_lev_a,"Variable",true);
	}

	node = xml_file->Get_Node(xml_file->Root_Node(),"Report_Groups");
	groups->clear();
	
	node_lev_a = xml_file->Get_Node(&node,"Group",true);

	while(!node_lev_a.isNull())
	{
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Name");
		group.name = xml_file->Read_Text_Node(&node_lev_b);
		
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Items");
		group.items.clear();
		
		node_lev_c = xml_file->Get_Node(&node_lev_b,"Item",true);
		
		while(!node_lev_c.isNull())
		{
			group.items.push_back(xml_file->Read_Text_Node(&node_lev_c));

			node_lev_c = xml_file->Get_Sibling_Node(&node_lev_c,"Item",true);
		}

		groups->push_back(group);
		
		node_lev_a = xml_file->Get_Sibling_Node(&node_lev_a,"Group",true);
	}
	
	node = xml_file->Get_Node(xml_file->Root_Node(),"User_Input");
	
	node_lev_a = xml_file->Get_Node(&node,"Machine_Expansion_Coefficient");
	input_data->machine_expansion_coefficient = xml_file->Read_Text_Node(&node_lev_a).toDouble();
	node_lev_a = xml_file->Get_Node(&node,"Machine_Scale");
	input_data->machine_scale = xml_file->Read_Text_Node(&node_lev_a).toDouble();
	node_lev_a = xml_file->Get_Node(&node,"Temperature_Minimum");
	input_data->temp_min = xml_file->Read_Text_Node(&node_lev_a).toDouble();
	node_lev_a = xml_file->Get_Node(&node,"Temperature_Maximum");
	input_data->temp_max = xml_file->Read_Text_Node(&node_lev_a).toDouble();
	node_lev_a = xml_file->Get_Node(&node,"Report_Increment");
	input_data->report_increment = xml_file->Read_Text_Node(&node_lev_a).toDouble();
	node_lev_a = xml_file->Get_Node(&node,"Report_End");
	input_data->report_end = xml_file->Read_Text_Node(&node_lev_a).toDouble();
	
	// default value for unsupported fields
	// ------------------------------------
	input_data->temp_gradient = 0.0;
}

void TUCBudget::LoadV2(
	TXmlFile							*xml_file,
	std::vector<TVariable>				*variables,
	std::vector<TGroup>					*groups,
	TInputData							*input_data)
{
	// no difference in format between V1 and V2 files
	// V2 expressions may include variables that didn't exist in V1
	// safe to load V2 file as V1 content
	this->LoadV1(xml_file,variables,groups,input_data);
}

void TUCBudget::LoadV3(
	TXmlFile							*xml_file,
	std::vector<TVariable>				*variables,
	std::vector<TGroup>					*groups,
	TInputData							*input_data)
{
	QDomElement							node;
	QDomElement							node_lev_a;
	QDomElement							node_lev_b;
	QDomElement							node_lev_c;
	QDomElement							node_lev_d;
	QDomElement							node_lev_e;
	QDomElement							node_lev_f;
	TUCBudgetData::TComponent			component;
	TUCBudgetData::TComponentItem		component_item;
	TUCBudgetData::TComponentInstance	component_instance;
	TUCBudgetData::TReplacementItem		replacement_item;
	TVariable							variable;
	TGroup								group;
	
	assert(xml_file);
	
	d_budget_data.Clear();
	
	node = xml_file->Get_Node(xml_file->Root_Node(),"Components");
	
	node_lev_a = xml_file->Get_Node(&node,"Component",true);
	
	while(!node_lev_a.isNull())
	{
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Name");
		component.name = xml_file->Read_Text_Node(&node_lev_b);
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Mode");
		component.mode = this->UCMode(xml_file->Read_Text_Node(&node_lev_b));
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Export_Equipment");
		component.export_equipment = this->Boolean_Type(xml_file->Read_Text_Node(&node_lev_b));
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Limit_Uncertainty_Length");
		component.limit_uncertainty_length = this->Boolean_Type(xml_file->Read_Text_Node(&node_lev_b));
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Length");
		component.length = xml_file->Read_Text_Node(&node_lev_b).toDouble();
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Expansion_Coefficient");
		component.expansion_coefficient = xml_file->Read_Text_Node(&node_lev_b).toDouble() / 1000;
		
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Identifiers");
		node_lev_c = xml_file->Get_Node(&node_lev_b,"Identifier",true);
		
		component.instances.clear();
		
		while(!node_lev_c.isNull())
		{
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Name");
			component_instance.name = xml_file->Read_Text_Node(&node_lev_d);
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Calibration_Date");
			component_instance.calibration_date = xml_file->Read_Text_Node(&node_lev_d);
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Calibration_Due_Date");
			component_instance.calibration_due_date = xml_file->Read_Text_Node(&node_lev_d);
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Length");
			component_instance.length = xml_file->Read_Text_Node(&node_lev_d).toDouble();
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Expansion_Coefficient");
			component_instance.expansion_coefficient = xml_file->Read_Text_Node(&node_lev_d).toDouble() / 1000;
			
			component.instances.push_back(component_instance);
			
			node_lev_c = xml_file->Get_Sibling_Node(&node_lev_c,"Identifier",true);
		}
		
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Component_Items");
		node_lev_c = xml_file->Get_Node(&node_lev_b,"Item",true);
		
		component.items.clear();
		
		while(!node_lev_c.isNull())
		{
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Name");
			component_item.name = xml_file->Read_Text_Node(&node_lev_d);
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Expression");
			component_item.expression = xml_file->Read_Text_Node(&node_lev_d);
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Distribution");
			component_item.distribution = this->Distribution_Type(xml_file->Read_Text_Node(&node_lev_d));
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Expression_Type");
			component_item.expression_type = this->Expression_Type(xml_file->Read_Text_Node(&node_lev_d));
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Uncertainty_Type");
			component_item.uctype = this->UCType(xml_file->Read_Text_Node(&node_lev_d));
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Sensitivity");
			component_item.sensitivity = xml_file->Read_Text_Node(&node_lev_d).toDouble();
			
			
			node_lev_d = xml_file->Get_Node(&node_lev_c,"Replacement_Items");
			node_lev_e = xml_file->Get_Node(&node_lev_d,"Item",true);
			
			component_item.replacement_items.clear();
			
			while(!node_lev_e.isNull())
			{
				node_lev_f = xml_file->Get_Node(&node_lev_e,"Identifier");
				replacement_item.identifier = xml_file->Read_Text_Node(&node_lev_f);
				node_lev_f = xml_file->Get_Node(&node_lev_e,"Expression");
				replacement_item.expression = xml_file->Read_Text_Node(&node_lev_f);
				node_lev_f = xml_file->Get_Node(&node_lev_e,"Use_Default_Expression");
				replacement_item.use_default_expression = this->Boolean_Type(xml_file->Read_Text_Node(&node_lev_f));
				
				component_item.replacement_items.push_back(replacement_item);
				
				node_lev_e = xml_file->Get_Sibling_Node(&node_lev_e,"Item",true);
			}
			
			component.items.push_back(component_item);
			
			node_lev_c = xml_file->Get_Sibling_Node(&node_lev_c,"Item",true);
		}
		
		d_budget_data.Add_Component(component);
		
		node_lev_a = xml_file->Get_Sibling_Node(&node_lev_a,"Component",true);
	}
	
	node = xml_file->Get_Node(xml_file->Root_Node(),"Variables");
	variables->clear();
	
	node_lev_a = xml_file->Get_Node(&node,"Variable",true);
	
	while(!node_lev_a.isNull())
	{
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Name");
		variable.name = xml_file->Read_Text_Node(&node_lev_b);
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Description");
		variable.description = xml_file->Read_Text_Node(&node_lev_b,true);
		variable.editor_index = -1;
		
		variables->push_back(variable);
		
		node_lev_a = xml_file->Get_Sibling_Node(&node_lev_a,"Variable",true);
	}
	
	node = xml_file->Get_Node(xml_file->Root_Node(),"Report_Groups");
	groups->clear();
	
	node_lev_a = xml_file->Get_Node(&node,"Group",true);
	
	while(!node_lev_a.isNull())
	{
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Name");
		group.name = xml_file->Read_Text_Node(&node_lev_b);
		
		node_lev_b = xml_file->Get_Node(&node_lev_a,"Items");
		group.items.clear();
		
		node_lev_c = xml_file->Get_Node(&node_lev_b,"Item",true);
		
		while(!node_lev_c.isNull())
		{
			group.items.push_back(xml_file->Read_Text_Node(&node_lev_c));
			
			node_lev_c = xml_file->Get_Sibling_Node(&node_lev_c,"Item",true);
		}
		
		groups->push_back(group);
		
		node_lev_a = xml_file->Get_Sibling_Node(&node_lev_a,"Group",true);
	}
	
	node = xml_file->Get_Node(xml_file->Root_Node(),"User_Input");
	
	node_lev_a = xml_file->Get_Node(&node,"Machine_Expansion_Coefficient");
	input_data->machine_expansion_coefficient = xml_file->Read_Text_Node(&node_lev_a).toDouble();
	node_lev_a = xml_file->Get_Node(&node,"Machine_Scale");
	input_data->machine_scale = xml_file->Read_Text_Node(&node_lev_a).toDouble();
	node_lev_a = xml_file->Get_Node(&node,"Temperature_Minimum");
	input_data->temp_min = xml_file->Read_Text_Node(&node_lev_a).toDouble();
	node_lev_a = xml_file->Get_Node(&node,"Temperature_Maximum");
	input_data->temp_max = xml_file->Read_Text_Node(&node_lev_a).toDouble();
	node_lev_a = xml_file->Get_Node(&node,"Temperature_Gradient");
	input_data->temp_gradient = xml_file->Read_Text_Node(&node_lev_a).toDouble();
	node_lev_a = xml_file->Get_Node(&node,"Report_Increment");
	input_data->report_increment = xml_file->Read_Text_Node(&node_lev_a).toDouble();
	node_lev_a = xml_file->Get_Node(&node,"Report_End");
	input_data->report_end = xml_file->Read_Text_Node(&node_lev_a).toDouble();
}

