/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////



#include <QVariant>
#include <QAction>
#include <QApplication>
#include <QButtonGroup>
#include <QGridLayout>
#include <QHeaderView>
#include <QMenuBar>
#include <QToolBar>
#include <QLineEdit>
#include <QTreeWidget>
#include <QTreeWidgetItem>
#include <QTabWidget>
#include <QGridLayout>
#include <QVBoxLayout>
#include <QPlainTextEdit>
#include <QList>
#include <QFont>
#include <QFileDialog>
#include <QCloseEvent>
#include <QToolButton>
#include <QIcon>
#include <QSpacerItem>
#include <QFrame>
#include <QSize>
#include <QLabel>
#include <QDragEnterEvent>
#include <QDropEvent>
#include <QMimeData>
#include <QList>
#include <QUrl>
#include <cmath>
#include <assert.h>

#include "../../core/messagebox.h"

#include "componentwidget.h"
#include "componentitemwidget.h"
#include "variablewidget.h"
#include "reportgroupwidget.h"
#include "reportwidget.h"
#include "namedialog.h"

#include "ucbudgeteditor.h"

TUCBudgetEditor::TUCBudgetEditor(
	const QWidget						*parent,
	Qt::WindowFlags						flags)
:QMainWindow(const_cast<QWidget*>(parent),flags)
{
    QWidget								*central_widget;
	QWidget								*budget_tree_container_widget;
	QWidget								*budget_widget;
	QToolBar							*toolbar;
	QMenuBar							*menu_bar;
	QMenu								*file_menu;
	QMenu								*view_menu;
	QMenu								*edit_menu;
    QGridLayout							*central_widget_layout;
	QGridLayout							*budget_layout;
	QVBoxLayout							*budget_tree_container_vlayout;
	QHBoxLayout							*budget_tree_control_hlayout;
	QSpacerItem							*budget_tree_control_hspacer;
	QFrame								*separator_vline1;
	QFrame								*separator_vline2;
	QLabel								*copyright_label;
	QLabel								*logo_label;
	QFont								log_font;
	QFont								copyright_font;

	this->resize(1101,801);
	this->setAcceptDrops(true);
		
#ifdef Q_OS_WIN
	d_data_path = qApp->applicationDirPath();
#else
	d_data_path = QDir::homePath();
#endif
	
	log_font.setFamily(QStringLiteral("Courier New"));
	log_font.setFixedPitch(true);
	
	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);

	central_widget = new QWidget(this);
	central_widget_layout = new QGridLayout(central_widget);
	
	d_tab_widget = new QTabWidget(central_widget);
	
	budget_widget = new QWidget;
	budget_layout = new QGridLayout(budget_widget);
	budget_layout->setContentsMargins(0,0,0,0);

	budget_tree_container_widget = new QWidget(this);
	budget_tree_container_vlayout = new QVBoxLayout(budget_tree_container_widget);
	budget_tree_container_vlayout->setContentsMargins(0,0,0,0);
	budget_tree_container_vlayout->setSpacing(4);

	d_budget_tree = new QTreeWidget(budget_tree_container_widget);
	d_budget_tree->setSelectionMode(QAbstractItemView::ExtendedSelection);
	d_budget_tree->setAnimated(true);
 	d_budget_tree->header()->setVisible(false);
	d_budget_tree->setSortingEnabled(true);
	d_budget_tree->sortByColumn(0,Qt::AscendingOrder);
	d_budget_tree->setExpandsOnDoubleClick(false);
	budget_tree_container_vlayout->addWidget(d_budget_tree);
	
	budget_tree_control_hlayout = new QHBoxLayout();
	
	d_add_budgetitem_button = new QToolButton(budget_tree_container_widget);
	d_add_budgetitem_button->setFixedSize(16,16);
	budget_tree_control_hlayout->addWidget(d_add_budgetitem_button);
	
	d_remove_budgetitem_button = new QToolButton(budget_tree_container_widget);
	d_remove_budgetitem_button->setFixedSize(16,16);
	budget_tree_control_hlayout->addWidget(d_remove_budgetitem_button);
	
	budget_tree_control_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	budget_tree_control_hlayout->addItem(budget_tree_control_hspacer);

	budget_tree_container_vlayout->addLayout(budget_tree_control_hlayout);

	budget_layout->addWidget(budget_tree_container_widget,0,0,1,1);

	separator_vline1 = new QFrame(budget_widget);
	separator_vline1->setFrameShadow(QFrame::Sunken);
	separator_vline1->setFrameShape(QFrame::VLine);
	budget_layout->addWidget(separator_vline1,0,1,1,1);
	
	d_component_widget = new TComponentWidget(budget_widget);
	budget_layout->addWidget(d_component_widget,0,2,1,1);
	
	separator_vline2 = new QFrame(budget_widget);
	separator_vline2->setFrameShadow(QFrame::Sunken);
	separator_vline2->setFrameShape(QFrame::VLine);
	budget_layout->addWidget(separator_vline2,0,3,1,1);

	d_component_item_widget = new TComponentItemWidget(budget_widget);
	budget_layout->addWidget(d_component_item_widget,0,4,1,1);

	d_tab_widget->addTab(budget_widget,QStringLiteral("Budget Components"));
	
	d_variable_widget = new TVariableWidget;
	d_tab_widget->addTab(d_variable_widget,QStringLiteral("Variables"));
	
	d_reportgroup_widget = new TReportGroupWidget;
	d_tab_widget->addTab(d_reportgroup_widget,QStringLiteral("Report Groups"));
	
	d_report_widget = new TReportWidget;
	d_tab_widget->addTab(d_report_widget,QStringLiteral("Report"));
	
	d_log_edit = new QPlainTextEdit;
	d_log_edit->setReadOnly(true);
	d_log_edit->setWordWrapMode(QTextOption::NoWrap);
	d_log_edit->setFont(log_font);
	d_tab_widget->addTab(d_log_edit,QStringLiteral("Log"));
	
	central_widget_layout->addWidget(d_tab_widget,0,0,1,2);
	
	copyright_label = new QLabel(this);
	copyright_label->setFont(copyright_font);
	copyright_label->setAlignment(Qt::AlignBottom|Qt::AlignLeading|Qt::AlignLeft);
	central_widget_layout->addWidget(copyright_label,1,0,1,1);
	
	logo_label = new QLabel(this);
	logo_label->setMinimumSize(100,40);
	logo_label->setMaximumSize(100,40);
	central_widget_layout->addWidget(logo_label,1,1,1,1);

	this->setCentralWidget(central_widget);
	
	d_new_action = new QAction(this);
	d_new_action->setText(QStringLiteral("&New..."));
	d_new_action->setShortcut(tr("Ctrl+N"));
	d_new_action->setWhatsThis(tr("Create a new uncertainty budget"));
	d_new_action->setIcon(QIcon(":/actions/new.png"));
	
	d_open_action = new QAction(this);
	d_open_action->setText(QStringLiteral("&Open..."));
	d_open_action->setShortcut(tr("Ctrl+O"));
	d_open_action->setWhatsThis(tr("Open an existing uncertainty budget"));
	d_open_action->setIcon(QIcon(":/actions/open.png"));
	
	d_save_action = new QAction(this);
	d_save_action->setText(QStringLiteral("&Save"));
	d_save_action->setShortcut(tr("Ctrl+S"));
	d_save_action->setWhatsThis(tr("Saves the current uncertainty budget"));
	d_save_action->setIcon(QIcon(":/actions/save.png"));
	d_save_action->setEnabled(false);
	
	d_saveas_action = new QAction(this);
	d_saveas_action->setText(QStringLiteral("&Save As"));
	d_saveas_action->setShortcut(tr("Ctrl+S"));
	d_saveas_action->setWhatsThis(tr("Saves the current uncertainty budget with a different file name"));
	d_saveas_action->setIcon(QIcon(":/actions/save_as.png"));
	d_saveas_action->setEnabled(false);
	
	d_exit_action = new QAction(this);
	d_exit_action->setText(QStringLiteral("Exit Program"));
	
	d_add_budgetitem_action = new QAction(this);
	d_add_budgetitem_action->setText(QStringLiteral("Add Budget Item"));
	d_add_budgetitem_action->setIcon(QIcon(":/gui/add.png"));
	
	d_remove_budgetitem_action = new QAction(this);
	d_remove_budgetitem_action->setText(QStringLiteral("Remove Budget Item"));
	d_remove_budgetitem_action->setEnabled(false);
	d_remove_budgetitem_action->setIcon(QIcon(":/gui/remove.png"));

	d_components_action = new QAction(this);
	d_components_action->setText(QStringLiteral("Budget &Components"));
	d_components_action->setShortcut(tr("Alt+C"));
	d_components_action->setWhatsThis(tr("Display the page for uncertainty components"));
	d_components_action->setCheckable(true);
	d_components_action->setIcon(QIcon(":/actions/components.png"));

	d_variables_action = new QAction(this);
	d_variables_action->setText(QStringLiteral("&Variables"));
	d_variables_action->setShortcut(tr("Alt+V"));
	d_variables_action->setWhatsThis(tr("Display the page for report variables"));
	d_variables_action->setCheckable(true);
	d_variables_action->setIcon(QIcon(":/actions/variables.png"));

	d_groups_action = new QAction(this);
	d_groups_action->setText(QStringLiteral("&Groups"));
	d_groups_action->setShortcut(tr("Alt+G"));
	d_groups_action->setWhatsThis(tr("Display the page for the report groups"));
	d_groups_action->setCheckable(true);
	d_groups_action->setIcon(QIcon(":/actions/group.png"));

	d_report_action = new QAction(this);
	d_report_action->setText(QStringLiteral("&Report"));
	d_report_action->setShortcut(tr("Alt+R"));
	d_report_action->setWhatsThis(tr("Display the page for the report"));
	d_report_action->setCheckable(true);
	d_report_action->setIcon(QIcon(":/actions/report.png"));

	d_log_action = new QAction(this);
	d_log_action->setText(QStringLiteral("&Log"));
	d_log_action->setShortcut(tr("Alt+L"));
	d_log_action->setWhatsThis(tr("Display the log page"));
	d_log_action->setCheckable(true);
	d_log_action->setIcon(QIcon(":/actions/log.png"));
	
	toolbar = new QToolBar(this);
	toolbar->setOrientation(Qt::Horizontal);
	toolbar->setIconSize(QSize(32,32));
	
	toolbar->addAction(d_new_action);
	toolbar->addSeparator();
	toolbar->addAction(d_open_action);
	toolbar->addAction(d_save_action);
	toolbar->addAction(d_saveas_action);
	toolbar->addSeparator();
	toolbar->addAction(d_components_action);
	toolbar->addAction(d_variables_action);
	toolbar->addAction(d_groups_action);
	toolbar->addAction(d_report_action);
	toolbar->addAction(d_log_action);
	
	this->addToolBar(toolbar);
	
	menu_bar = new QMenuBar;
	
	file_menu = new QMenu("File");
	file_menu->addAction(d_new_action);
	file_menu->addSeparator();
	file_menu->addAction(d_save_action);
	file_menu->addAction(d_saveas_action);
	file_menu->addSeparator();
	file_menu->addAction(d_exit_action);
	menu_bar->addMenu(file_menu);
	
	edit_menu = new QMenu("Edit");
	edit_menu->addAction(d_add_budgetitem_action);
	edit_menu->addAction(d_remove_budgetitem_action);
	edit_menu->addSeparator();
	menu_bar->addMenu(edit_menu);

	view_menu = new QMenu("View");
	view_menu->addAction(d_components_action);
	view_menu->addAction(d_variables_action);
	view_menu->addAction(d_groups_action);
	view_menu->addAction(d_report_action);
	view_menu->addAction(d_log_action);
	menu_bar->addMenu(view_menu);

	this->setMenuBar(menu_bar);
	
	d_budget_tree->setContextMenuPolicy(Qt::ActionsContextMenu);
	d_budget_tree->addAction(d_add_budgetitem_action);
	d_budget_tree->addAction(d_remove_budgetitem_action);

	// defaults
	d_data_modified = false;
	d_save_file_name = QString();
	d_tab_widget->setCurrentIndex(0);
	d_components_action->setChecked(true);
	d_current_item_index = -1;
	d_component_widget->setEnabled(false);
	d_component_item_widget->setEnabled(false);
	d_default_component_identifier_name = QStringLiteral("[Default]");
	d_remove_budgetitem_action->setEnabled(false);
	d_remove_budgetitem_button->setEnabled(false);
	d_add_budgetitem_action->setEnabled(true);
	d_add_budgetitem_button->setEnabled(true);
	d_msg_box = new TMessageBox(this);

	d_component_item_widget->Reset_Distribution_Types(d_ucbudget.Distribution_Type_List());
	d_component_item_widget->Reset_Uncertainty_Types(d_ucbudget.Uncertainty_Type_List());
		
	this->Update_BudgetTree();
	this->Select_BudgetTree_Item();
	
	logo_label->setPixmap(QPixmap(QStringLiteral(":/images/logo.png")));
	
	d_add_budgetitem_button->setIcon(QIcon(QStringLiteral(":/gui/add.png")));
	d_remove_budgetitem_button->setIcon(QIcon(QStringLiteral(":/gui/remove.png")));

	this->setWindowTitle(QStringLiteral("UC Budget Editor - Version 7.1"));
	copyright_label->setText(QStringLiteral("Copyright (C) 2024 Select Calibration Incorporated.  www.selectcalibration.ca"));
	
	connect(d_new_action, &QAction::triggered,this, &TUCBudgetEditor::New);
	connect(d_open_action, &QAction::triggered,this, &TUCBudgetEditor::Open);
	connect(d_save_action, &QAction::triggered,this, &TUCBudgetEditor::Save);
	connect(d_saveas_action, &QAction::triggered,this, &TUCBudgetEditor::SaveAs);
	connect(d_exit_action, &QAction::triggered,this, &TUCBudgetEditor::Exit);
	connect(d_log_action, &QAction::triggered,this, &TUCBudgetEditor::View_Log);
	connect(d_report_action, &QAction::triggered,this, &TUCBudgetEditor::View_Report);
	connect(d_groups_action, &QAction::triggered,this, &TUCBudgetEditor::View_Groups);
	connect(d_variables_action, &QAction::triggered,this, &TUCBudgetEditor::View_Variables);
	connect(d_components_action, &QAction::triggered,this, &TUCBudgetEditor::View_Budget_Data);
	connect(d_remove_budgetitem_action, &QAction::triggered,this, &TUCBudgetEditor::BudgetTree_Remove_Item);
	connect(d_add_budgetitem_action, &QAction::triggered,this, &TUCBudgetEditor::BudgetTree_Add_Item);

	connect(d_tab_widget,&QTabWidget::currentChanged,this,&TUCBudgetEditor::Tab_Selection_Changed);

	connect(d_budget_tree,&QTreeWidget::itemSelectionChanged,this,&TUCBudgetEditor::BudgetTree_Selection_Changed);
	connect(d_budget_tree->itemDelegate(),&QAbstractItemDelegate::commitData,this,&TUCBudgetEditor::BudgetTree_Rename_Item);

	connect(d_component_widget,&TComponentWidget::Add_Component_Identifier,this,&TUCBudgetEditor::Component_Add_Identifier);
	connect(d_component_widget,&TComponentWidget::Rename_Component_Identifiers,this,&TUCBudgetEditor::Component_Rename_Identifier);
	connect(d_component_widget,&TComponentWidget::Remove_Component_Identifiers,this,&TUCBudgetEditor::Component_Remove_Identifiers);
	connect(d_component_widget,&TComponentWidget::Component_Changed,this,&TUCBudgetEditor::Component_Data_Changed);
	connect(d_component_widget,&TComponentWidget::Component_Identifier_Selected,this,&TUCBudgetEditor::Component_Identifier_Selected);

	connect(d_component_item_widget,&TComponentItemWidget::Component_Item_Changed,this,&TUCBudgetEditor::Component_Item_Data_Changed);

	connect(d_variable_widget,&TVariableWidget::Variable_Changed,this,&TUCBudgetEditor::Variable_Data_Chanaged);

	connect(d_reportgroup_widget,&TReportGroupWidget::Report_Groups_Changed,this,&TUCBudgetEditor::Report_Groups_Changed);

	connect(d_report_widget,&TReportWidget::Create_Report,this,&TUCBudgetEditor::Update_Report_View);
	connect(d_report_widget,&TReportWidget::Create_Export_Report,this,&TUCBudgetEditor::Generate_Export_Report);
	
	connect(d_add_budgetitem_button, &QToolButton::clicked,this, &TUCBudgetEditor::BudgetTree_Add_Item);
	connect(d_remove_budgetitem_button, &QToolButton::clicked,this, &TUCBudgetEditor::BudgetTree_Remove_Item);
}

TUCBudgetEditor::~TUCBudgetEditor(void)
{
}
			
void TUCBudgetEditor::BudgetTree_Selection_Changed(void)
{
	QList<QTreeWidgetItem*>				sel_items;
	QTreeWidgetItem						*item;
	QTreeWidgetItem						*parent;
	QStringList							list;
	int									item_level;
	int									index;
	TUCBudgetData::TComponent			budget_component;
	TUCBudgetData::TComponentItem		budget_component_item;
				
	sel_items = d_budget_tree->selectedItems();
		
	d_add_budgetitem_action->setEnabled(sel_items.size() == 1);
	d_add_budgetitem_button->setEnabled(sel_items.size() == 1);
	
	d_remove_budgetitem_action->setEnabled(sel_items.size() > 0);
	d_remove_budgetitem_button->setEnabled(sel_items.size() > 0);

	if(sel_items.size() == 1)
	{
		item = sel_items[0];

		// determine depth of tree widget
		item_level = 0;
		parent = item->parent();
		
		while(parent != 0)
		{
			item_level++;
			parent = parent->parent();			
		}
				
		d_current_item_index = item_level;
		parent = item->parent();
		
		if(item_level == 1)
		{
			d_current_component_name = item->text(0);
			d_current_component_item_name = QString();
			d_current_component_identifier_name = QString();
			
			d_component_item_widget->Set_Idenfitier_Name(d_default_component_identifier_name);
			d_component_item_widget->Enable_Option_Use_Default(false);
			
			list = d_ucbudget.Enum_Component_Identifiers(d_current_component_name);
			d_component_widget->Set_Identifiers(list);
			
			index = d_ucbudget.Budget_Data()->Component_Index(d_current_component_name);
			budget_component = d_ucbudget.Budget_Data()->Component(index);
			
			d_component_widget->Clear_Dates();
			d_component_widget->Set_Expansion_Coefficient(budget_component.expansion_coefficient);
			d_component_widget->Set_Length(budget_component.length);
			
			if(budget_component.mode == TUCBudgetData::SUM_ITEMS)
			{
				d_component_widget->Set_Mode_Sum();
			}
			else
			{
				d_component_widget->Set_Mode_Single();
			}
			
			d_component_widget->Set_Mode_Export_Equipment(budget_component.export_equipment);
			d_component_widget->Set_Mode_Limit_Uncertainty_Length(budget_component.limit_uncertainty_length);
			
			d_remove_budgetitem_action->setEnabled(true);
			d_remove_budgetitem_button->setEnabled(true);

			d_component_widget->setEnabled(true);
			d_component_item_widget->setEnabled(false);

		}
		else if(item_level == 2)
		{
			if(d_current_component_name.compare(parent->text(0)) != 0 )
			{
				d_current_component_name = parent->text(0);
						
				list = d_ucbudget.Enum_Component_Identifiers(d_current_component_name);
				d_component_widget->Set_Identifiers(list);
				
				index = d_ucbudget.Budget_Data()->Component_Index(d_current_component_name);
				budget_component = d_ucbudget.Budget_Data()->Component(index);

				if(d_current_component_item_name.compare(item->text(0)) != 0)
				{
					d_current_component_item_name = item->text(0);
	
					d_component_item_widget->Set_Idenfitier_Name(d_default_component_identifier_name);
					d_component_item_widget->Enable_Option_Use_Default(false);
				}
				
				d_component_widget->Clear_Dates();
				d_component_widget->Set_Expansion_Coefficient(budget_component.expansion_coefficient);
				d_component_widget->Set_Length(budget_component.length);
				
				if(budget_component.mode == TUCBudgetData::SUM_ITEMS)
				{
					d_component_widget->Set_Mode_Sum();
				}
				else
				{
					d_component_widget->Set_Mode_Single();
				}
				
				d_component_widget->Set_Mode_Export_Equipment(budget_component.export_equipment);
				d_component_widget->Set_Mode_Limit_Uncertainty_Length(budget_component.limit_uncertainty_length);
			}
			
			d_current_component_item_name = item->text(0);

			d_remove_budgetitem_action->setEnabled(true);
			d_remove_budgetitem_button->setEnabled(true);

			d_component_widget->setEnabled(true);
			d_component_item_widget->setEnabled(false);
				
			budget_component_item = d_ucbudget.Budget_Data()->Component_Item(d_current_component_name,d_current_component_item_name);
			
			if(budget_component_item.expression_type == TUCBudgetData::CONSTANT_LINEAR)
			{
				d_component_item_widget->Set_Radio_Options(TComponentItemWidget::CONSTANT_LINEAR);
			}
			else if(budget_component_item.expression_type == TUCBudgetData::CONSTANT_TEMPERATURE)
			{
				d_component_item_widget->Set_Radio_Options(TComponentItemWidget::CONSTANT_TEMPERATURE);
			}
			else if(budget_component_item.expression_type == TUCBudgetData::LENGTH_DEPENDENT_LINEAR)
			{
				d_component_item_widget->Set_Radio_Options(TComponentItemWidget::LENGTH_DEPENDENT_LINEAR);
			}
			else if(budget_component_item.expression_type == TUCBudgetData::LENGTH_DEPENDENT_TEMPERATURE)
			{
				d_component_item_widget->Set_Radio_Options(TComponentItemWidget::LENGTH_DEPENDENT_TEMPERATURE);
			}
			else
			{
				assert(false);
			}
			
			d_component_item_widget->Set_Distribution_Type(d_ucbudget.Distribution_Type(budget_component_item.distribution));
			d_component_item_widget->Set_UCType(d_ucbudget.UCType(budget_component_item.uctype));
			d_component_item_widget->Set_Sensitivity(budget_component_item.sensitivity);
			
			if(d_current_component_identifier_name.length())
			{
				this->Component_Identifier_Selected(d_current_component_identifier_name);
			}
			else
			{
				d_component_item_widget->Set_Expression(budget_component_item.expression); // default expression if no identifier selected
			}
		
			d_remove_budgetitem_action->setEnabled(true);
			d_remove_budgetitem_button->setEnabled(true);

			d_component_widget->setEnabled(true);
			d_component_item_widget->setEnabled(true);
		}
		else
		{
			d_remove_budgetitem_action->setEnabled(false);
			d_remove_budgetitem_button->setEnabled(false);

			d_current_component_name = QString();
			d_current_component_item_name = QString();
			d_current_component_identifier_name = QString();

			d_component_widget->setEnabled(false);
			d_component_item_widget->setEnabled(false);
			
			d_component_item_widget->Set_Idenfitier_Name(d_default_component_identifier_name);
		}
	}
	else
	{
		d_component_widget->setEnabled(false);
		d_component_item_widget->setEnabled(false);
	}
}

void TUCBudgetEditor::Tab_Selection_Changed(
	int									index)
{
	switch(index)
	{
		case 0:
			d_components_action->setChecked(true);
			d_variables_action->setChecked(false);
			d_groups_action->setChecked(false);
			d_report_action->setChecked(false);
			d_log_action->setChecked(false);
			break;
			
		case 1:
			d_components_action->setChecked(false);
			d_variables_action->setChecked(true);
			d_groups_action->setChecked(false);
			d_report_action->setChecked(false);
			d_log_action->setChecked(false);
			break;
			
		case 2:
			d_components_action->setChecked(false);
			d_variables_action->setChecked(false);
			d_groups_action->setChecked(true);
			d_report_action->setChecked(false);
			d_log_action->setChecked(false);
			break;
			
		case 3:
			d_components_action->setChecked(false);
			d_variables_action->setChecked(false);
			d_groups_action->setChecked(false);
			d_report_action->setChecked(true);
			d_log_action->setChecked(false);
			
			d_report_widget->Refresh();
			
			break;
			
		case 4:
			d_components_action->setChecked(false);
			d_variables_action->setChecked(false);
			d_groups_action->setChecked(false);
			d_report_action->setChecked(false);
			d_log_action->setChecked(true);
			break;
			
		case 5:
			d_components_action->setChecked(false);
			d_variables_action->setChecked(false);
			d_groups_action->setChecked(false);
			d_report_action->setChecked(false);
			d_log_action->setChecked(false);
			break;
	}
}

void TUCBudgetEditor::BudgetTree_Rename_Item(
	QWidget 							*editor)
{
	QString 							new_name;
	
	new_name = reinterpret_cast<QLineEdit*>(editor)->text();
	
	if(d_current_item_index == 1  && d_current_component_name != new_name && new_name.length())
	{
		d_ucbudget.Budget_Data()->Rename_Component(d_current_component_name,new_name);
	}
	else if(d_current_item_index == 2  && d_current_component_item_name != new_name && new_name.length())
	{
		d_ucbudget.Budget_Data()->Rename_Component_Item(d_current_component_name,d_current_component_item_name,new_name);
	}
}

void TUCBudgetEditor::BudgetTree_Add_Item(void)
{
	QList<QTreeWidgetItem*>				sel_items;
	TNameDialog							name_dialog(this);
	QString								name;
	
	name_dialog.Set_Title("New Budget Item");
	sel_items = d_budget_tree->selectedItems();

	if(sel_items.size() == 1)
	{
		if(name_dialog.exec() == QDialog::Accepted)
		{
			name = name_dialog.Name();

			if(d_current_item_index == 0)
			{
				if(d_ucbudget.Budget_Data()->Add_Component(name))
				{
					this->Update_BudgetTree();
					d_variable_widget->Update_Variable_List(d_ucbudget.Enum_Reportable_Components());
					d_reportgroup_widget->Update_Reportable_Components(d_ucbudget.Enum_Reportable_Components());
					
					this->Select_BudgetTree_Item(name);
					
					d_data_modified = true;
				}
			}
			else if(d_current_item_index > 0)
			{
				if(d_ucbudget.Budget_Data()->Add_Component_Item(d_current_component_name,name))
				{
					this->Update_BudgetTree();
					d_variable_widget->Update_Variable_List(d_ucbudget.Enum_Reportable_Components());
					d_reportgroup_widget->Update_Reportable_Components(d_ucbudget.Enum_Reportable_Components());
					
					this->Select_BudgetTree_Item(d_current_component_name,name);

					d_data_modified = true;
				}
			}
		}
	}
	
	d_save_action->setEnabled(d_data_modified);
	d_saveas_action->setEnabled(true);
}

void TUCBudgetEditor::BudgetTree_Remove_Item(void)
{
	QList<QTreeWidgetItem*>				sel_items;
	QList<QTreeWidgetItem*>::iterator	item_iter;
	QStringList							items;
	
	d_msg_box->setText("Delete Components");
	d_msg_box->setInformativeText("Are you sure you want to delete the selected components?");
	d_msg_box->setDetailedText(QString());
	d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
	d_msg_box->setDefaultButton(QMessageBox::No);
	d_msg_box->setIcon(QMessageBox::Question);
	
	if(d_msg_box->exec() == QMessageBox::No)
	{
		return;
	}
	
	d_data_modified = true;

	sel_items = d_budget_tree->selectedItems();
	
	if(d_current_item_index == 1)	// component
	{
		for(item_iter = sel_items.begin();item_iter != sel_items.end();++item_iter)
		{
			d_ucbudget.Budget_Data()->Remove_Component((*item_iter)->text(0));
		}
	}
	else if(d_current_item_index == 2) // component item
	{
		for(item_iter = sel_items.begin();item_iter != sel_items.end();++item_iter)
		{
			items.push_back((*item_iter)->text(0));
		}

		if(items.size())
		{
			d_ucbudget.Budget_Data()->Remove_Component_Items(d_current_component_name,items);
		}
	}
	
	this->Update_BudgetTree();
	d_variable_widget->Update_Variable_List(d_ucbudget.Enum_Reportable_Components());
	d_reportgroup_widget->Update_Reportable_Components(d_ucbudget.Enum_Reportable_Components());

	d_save_action->setEnabled(d_data_modified);
	d_saveas_action->setEnabled(true);
}

void TUCBudgetEditor::Component_Identifier_Selected(
	const QString						&identifier_name)
{
	TUCBudgetData::TComponentInstance	component_identifier;
	TUCBudgetData::TComponent			component;
	QString								default_expression;
	QString								replacement_expression;
	int									component_id;
	bool								use_default;

	component_id = d_ucbudget.Budget_Data()->Component_Index(d_current_component_name);
		
	if(!(component_id < 0))
	{
		component = d_ucbudget.Budget_Data()->Component(component_id);
		
		if(identifier_name.length())
		{
			d_current_component_identifier_name = identifier_name;

			component_identifier = d_ucbudget.Budget_Data()->Component_Instance(component_id,d_current_component_identifier_name);
			
			d_component_widget->Set_Dates(component_identifier.calibration_date,component_identifier.calibration_due_date);
			d_component_widget->Set_Expansion_Coefficient(component_identifier.expansion_coefficient);
			d_component_widget->Set_Length(component_identifier.length);
			
			d_component_item_widget->Set_Idenfitier_Name(d_current_component_identifier_name);
			d_component_item_widget->Enable_Option_Use_Default(true);
		}
		else
		{
			d_current_component_identifier_name = QString();

			d_component_widget->Set_Expansion_Coefficient(component.expansion_coefficient);
			d_component_widget->Set_Length(component.length);
			
			d_component_item_widget->Set_Idenfitier_Name(d_default_component_identifier_name);
			d_component_item_widget->Enable_Option_Use_Default(false);
		}
	}
	else
	{
		d_current_component_identifier_name = QString();
		d_component_item_widget->Set_Idenfitier_Name(d_default_component_identifier_name);
		d_component_item_widget->Set_Option_Use_Default(false);
		d_component_item_widget->Enable_Option_Use_Default(false);

		d_component_widget->Clear_Dates();
	}
	
		
	if(d_ucbudget.Component_Item_Default_Expression(d_current_component_name,
									   d_current_component_item_name,
									   &default_expression))
	{
		if(d_ucbudget.Component_Item_Expression(d_current_component_name,
													  d_current_component_item_name,
													  d_current_component_identifier_name,
													  &use_default,
													  &replacement_expression))
		{
			if(use_default)
			{
				d_component_item_widget->Set_Expression(default_expression);
			}
			else
			{
				d_component_item_widget->Set_Expression(replacement_expression);
			}
			
			d_component_item_widget->Set_Option_Use_Default(use_default);
		}
		else
		{
			d_component_item_widget->Set_Expression(default_expression);
			d_component_item_widget->Set_Option_Use_Default(false);
		}
	}
}

void TUCBudgetEditor::Component_Add_Identifier(
	const QString						&name)
{	
	d_ucbudget.Budget_Data()->Component_Add_Identifier(d_current_component_name,name);
	
	d_data_modified = true;
	
	d_save_action->setEnabled(d_data_modified);
	d_saveas_action->setEnabled(true);
}

void TUCBudgetEditor::Component_Rename_Identifier(
	const QString						&old_name,
	const QString						&new_name)
{
	d_ucbudget.Budget_Data()->Rename_Component_Identifier(d_current_component_name,old_name,new_name);
	
	d_data_modified = true;
	
	d_save_action->setEnabled(d_data_modified);
	d_saveas_action->setEnabled(true);
}

void TUCBudgetEditor::Component_Remove_Identifiers(
	const QStringList					&names)
{
	d_ucbudget.Budget_Data()->Component_Remove_Identifier(d_current_component_name,names);
	
	d_data_modified = true;
	
	d_save_action->setEnabled(d_data_modified);
	d_saveas_action->setEnabled(true);
}

void TUCBudgetEditor::Component_Data_Changed(void)
{
	TUCBudgetData::TComponent			component;
	TUCBudgetData::TComponentInstance	component_identifier;
	QString								component_identifier_name;
	int									component_id;
		
	component_id = d_ucbudget.Budget_Data()->Component_Index(d_current_component_name);
	component = d_ucbudget.Budget_Data()->Component(component_id);
		
	if(d_component_widget->Mode_Single())
	{
		component.mode = TUCBudgetData::INDIVIUAL_ITEM;
	}
	else
	{
		component.mode = TUCBudgetData::SUM_ITEMS;
	}
	
	component.export_equipment = d_component_widget->Mode_Export_Equipment();
	component.limit_uncertainty_length = d_component_widget->Mode_Limit_Uncertainty_Length();
	
	d_ucbudget.Budget_Data()->Update_Component(component);
	
	component_identifier_name = d_component_widget->Current_Identifier();
	
	if(component_identifier_name.length())
	{
		component_identifier = d_ucbudget.Budget_Data()->Component_Instance(component_id,component_identifier_name);
		component_identifier.calibration_date = d_component_widget->Calibration_Date();
		component_identifier.calibration_due_date = d_component_widget->Calibration_Due_Date();
	
		component_identifier.expansion_coefficient = d_component_widget->Expansion_Coefficient();
		component_identifier.length = d_component_widget->Length();

		d_ucbudget.Budget_Data()->Update_Component_Identifier(component_id,component_identifier);
	}
	
	d_variable_widget->Update_Variable_List(d_ucbudget.Enum_Reportable_Components());
	d_reportgroup_widget->Update_Reportable_Components(d_ucbudget.Enum_Reportable_Components());
	
	d_data_modified = true;
	
	d_save_action->setEnabled(d_data_modified);
	d_saveas_action->setEnabled(true);
}

void TUCBudgetEditor::Component_Item_Data_Changed(void)
{
	TUCBudgetData::TComponentItem		component_item;
	TComponentItemWidget::TRadioOptions	radio_options;
	QString								expression;
	int									component_id;
	bool								use_default;

	component_id = d_ucbudget.Budget_Data()->Component_Index(d_current_component_name);
	component_item = d_ucbudget.Budget_Data()->Component_Item(component_id,d_current_component_item_name);
	
	component_item.uctype = d_ucbudget.UCType(d_component_item_widget->UCType());
	component_item.distribution = d_ucbudget.Distribution_Type(d_component_item_widget->Distribution_Type());
	component_item.sensitivity = d_component_item_widget->Sensitivity();
	
	radio_options = d_component_item_widget->Radio_Options();
	
	if(radio_options == TComponentItemWidget::CONSTANT_LINEAR)
	{
		component_item.expression_type = TUCBudgetData::CONSTANT_LINEAR;
	}
	else if(radio_options == TComponentItemWidget::CONSTANT_TEMPERATURE)
	{
		component_item.expression_type = TUCBudgetData::CONSTANT_TEMPERATURE;
	}
	else if(radio_options == TComponentItemWidget::LENGTH_DEPENDENT_LINEAR)
	{
		component_item.expression_type = TUCBudgetData::LENGTH_DEPENDENT_LINEAR;
	}
	else if(radio_options == TComponentItemWidget::LENGTH_DEPENDENT_TEMPERATURE)
	{
		component_item.expression_type = TUCBudgetData::LENGTH_DEPENDENT_TEMPERATURE;
	}
	
	d_ucbudget.Budget_Data()->Update_Component_Item(component_id,component_item);
	
	expression = d_component_item_widget->Expression();
	use_default = d_component_item_widget->Use_Default_Expression();

	d_ucbudget.Update_Component_Item(d_current_component_name,d_current_component_item_name,d_current_component_identifier_name,expression,use_default);
	
	d_data_modified = true;
	
	d_save_action->setEnabled(d_data_modified);
	d_saveas_action->setEnabled(true);
}

void TUCBudgetEditor::Variable_Data_Chanaged(void)
{
	d_data_modified = true;
	
	d_save_action->setEnabled(d_data_modified);
	d_saveas_action->setEnabled(true);
}

void TUCBudgetEditor::Report_Groups_Changed(void)
{
	QStringList							list;
	
	list = d_reportgroup_widget->Group_Names();
	d_report_widget->Update_Report_Groups(list);
	
	d_data_modified = true;
	
	d_save_action->setEnabled(d_data_modified);
	d_saveas_action->setEnabled(true);
}

void TUCBudgetEditor::Update_Report_View(
	const QString						&group_name)
{
	double								length;
	
	d_log_edit->clear();
		
	length = d_variable_widget->End_Position();

	this->Create_Report_Data(group_name);
	
	d_report_widget->Clear_Report();
	
	if(this->Update_Report(length))
	{
		d_report_widget->Update_Status_Display(length);
	}
	else
	{
		this->Write_Log_Text(QString("ERR:  Cannot generate report data."));
		d_tab_widget->setCurrentWidget(d_log_edit);
		d_report_widget->Set_Default_Report_Group();
	}
}

void TUCBudgetEditor::Generate_Export_Report(
	const QString						&group_name)
{
	double								start_pos;
	double								end_pos;
	double								increment;
	double								pos;
	QString								file_name;
	QString								file_filter;
	QString								export_title;
	int									index;
	TReportWidget::TExportFormat		export_format;
	
	d_log_edit->clear();
	
	start_pos = d_variable_widget->Start_Position();
	end_pos = d_variable_widget->End_Position();
	increment = d_variable_widget->Increment();
	
	if(!(end_pos > start_pos))
	{
		this->Write_Log_Text(QString("ERR:  Start and End position invalid.  %1  to  %2").arg(start_pos,0,'f',6).arg(end_pos,0,'f',6));
		d_tab_widget->setCurrentWidget(d_log_edit);
		return;
	}
	
	if(!(increment > 0))
	{
		this->Write_Log_Text(QString("ERR:  Increment cannot be zero or negative.  %1").arg(increment,0,'f',6));
		d_tab_widget->setCurrentWidget(d_log_edit);
		return;
	}
	
	// check for excessive items.  No more than 10 pages
	index = static_cast<int>(0.5 + ((end_pos - start_pos) / increment));
	index /= 12;
	index++;
	
	if(index > 10)
	{
		this->Write_Log_Text(QString("ERR:  Range and increment will create a %1 page report (maximum is 10)").arg(index));
		d_tab_widget->setCurrentWidget(d_log_edit);
		return;
	}

	export_format = d_report_widget->Export_Format();
	
	switch(export_format)
	{
		case TReportWidget::EXPORT_PDF_FORMAT:
			file_filter = QStringLiteral("PDF Files (*.pdf)");
			export_title = QStringLiteral("Enter the name of the PDF to create");
			break;
			
		case TReportWidget::EXPORT_CSV_FORMAT:
			file_filter = QStringLiteral("CSV files (*.csv)");
			export_title = QStringLiteral("Enter the name of the CSV to create");
			break;
		
		default:
			file_filter = QStringLiteral("PDF Files (*.pdf);;CSV files (*.csv)");
			export_title = QStringLiteral("Enter the name of the file to create");
			break;
	}
	
	file_name = QFileDialog::getSaveFileName(this,export_title,d_data_path,file_filter);
	
	if(file_name.length() == 0)
	{
		return;
	}
	
	this->Create_Report_Data(group_name);
		
	d_report_widget->Clear_Export();
	
	for(pos = start_pos;pos <= (end_pos + (increment/2));pos += increment)
	{
		if(!this->Update_Export(pos))
		{
			this->Write_Log_Text(QString("ERR:  Cannot generate export data."));
			d_tab_widget->setCurrentWidget(d_log_edit);
			return;
		}
	}
	
	d_report_widget->Update_Export_Uncertainty_Data();
	d_report_widget->Format_Export_Data();
	
	switch(export_format)
	{
		case TReportWidget::EXPORT_PDF_FORMAT:
			d_report_widget->Write_Export_PDF_Data(file_name);
			break;
			
		case TReportWidget::EXPORT_CSV_FORMAT:
			d_report_widget->Write_Export_CSV_Data(file_name);
			break;
			
		default:
			break;
	}
}

void TUCBudgetEditor::New(void)
{
	if(d_data_modified)
	{
		d_msg_box->setText("Data Not Saved");
		d_msg_box->setInformativeText("Current data will be lost.  Do you want to continue?");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_msg_box->setDefaultButton(QMessageBox::No);
		d_msg_box->setIcon(QMessageBox::Question);
		
		if(d_msg_box->exec() == QMessageBox::No)
		{
			return;
		}
	}

	d_ucbudget.Budget_Data()->Clear();
	d_log_edit->clear();

	d_report_widget->Set_Default_Report_Group();
	d_report_widget->Clear_Report();
	d_report_widget->Update_Report_Groups(QStringList());
	d_reportgroup_widget->Clear_Groups();
	
	this->Update_BudgetTree();
	
	d_variable_widget->Update_Variable_List(d_ucbudget.Enum_Reportable_Components());
	d_reportgroup_widget->Update_Reportable_Components(d_ucbudget.Enum_Reportable_Components());
	
	d_data_modified = false;
	
	d_save_action->setEnabled(d_data_modified);
	d_saveas_action->setEnabled(false);
}

void TUCBudgetEditor::Open(void)
{
	QString								file_name;
	
	if(d_data_modified)
	{
		d_msg_box->setText("Data Not Saved");
		d_msg_box->setInformativeText("Current data will be lost.  Do you want to continue?");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_msg_box->setDefaultButton(QMessageBox::No);
		d_msg_box->setIcon(QMessageBox::Question);
		
		if(d_msg_box->exec() == QMessageBox::No)
		{
			return;
		}
	}

	
	file_name = QFileDialog::getOpenFileName(this,
											 QStringLiteral("Enter the name of the file to open"),
											 d_data_path,
											 QStringLiteral("UC Budget files (*.ucBudget)"));
	
	d_report_widget->Set_Default_Report_Group();
	d_report_widget->Clear_Report();
	d_report_widget->Update_Report_Groups(QStringList());
	d_reportgroup_widget->Clear_Groups();

	this->Load_Data(file_name);
}

void TUCBudgetEditor::Save(void)
{
	std::vector<TVariable>				variables;
	std::vector<TGroup>					groups;
	TInputData							input_data;
	int									start;
	int									var_count;

	if(d_save_file_name.length() == 0)
	{
		this->SaveAs();
		return;
	}
	
	start = d_variable_widget->Static_Variable_Count();
	var_count = d_variable_widget->Variable_Count();
	
	for(int i=start;i<var_count;++i)
	{
		variables.push_back(d_variable_widget->Variable(i));
	}
	
	groups = d_reportgroup_widget->Groups();
	
	input_data = d_variable_widget->Input_Data();

	d_ucbudget.Save(d_save_file_name,variables,groups,input_data);
	d_data_modified = false;
	
	d_save_action->setEnabled(false);
	d_saveas_action->setEnabled(true);
}

void TUCBudgetEditor::SaveAs(void)
{
	std::vector<TVariable>				variables;
	std::vector<TGroup>					groups;
	TInputData							input_data;
	int									start;
	int									var_count;
	int									index;
	QString								file_name;
	
	file_name = QFileDialog::getSaveFileName(this,
											 QStringLiteral("Enter the name of the file to save"),
											 d_data_path,
											 QStringLiteral("UC Budget files (*.ucBudget)"));
	
	if(file_name.length() == 0)
	{
		return;
	}
	
	if(!file_name.endsWith(".ucBudget",Qt::CaseInsensitive))
	{
		file_name += ".ucBudget";
	}
	
	// update data path
	d_data_path = file_name;
	
	index = d_data_path.lastIndexOf('/');
	d_data_path.truncate(index);
	
	start = d_variable_widget->Static_Variable_Count();
	var_count = d_variable_widget->Variable_Count();
	
	for(int i=start;i<var_count;++i)
	{
		variables.push_back(d_variable_widget->Variable(i));
	}
	
	groups = d_reportgroup_widget->Groups();
	
	input_data = d_variable_widget->Input_Data();
	
	d_ucbudget.Save(file_name,variables,groups,input_data);
	
	d_save_file_name = file_name;
	d_data_modified = false;
	
	d_save_action->setEnabled(false);
	d_saveas_action->setEnabled(true);
}

void TUCBudgetEditor::Exit(void)
{
	this->close();
}

void TUCBudgetEditor::View_Budget_Data(void)
{
	d_tab_widget->setCurrentIndex(0);
}

void TUCBudgetEditor::View_Variables(void)
{
	d_tab_widget->setCurrentIndex(1);

}

void TUCBudgetEditor::View_Groups(void)
{
	d_tab_widget->setCurrentIndex(2);

}

void TUCBudgetEditor::View_Report(void)
{
	d_tab_widget->setCurrentIndex(3);

}

void TUCBudgetEditor::View_Log(void)
{
	d_tab_widget->setCurrentIndex(4);

}

void TUCBudgetEditor::closeEvent(
	QCloseEvent							*event)
{
	if(d_data_modified)
	{
		d_msg_box->setText("Data Not Saved");
		d_msg_box->setInformativeText("Current data will be lost.  Are you sure you want to close?");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_msg_box->setDefaultButton(QMessageBox::No);
		d_msg_box->setIcon(QMessageBox::Question);
		
		if(d_msg_box->exec() == QMessageBox::No)
		{
			event->ignore();
			return;
		}
	}
	
	event->accept();
	
}

void TUCBudgetEditor::dragEnterEvent(
	QDragEnterEvent 					*drag_enter_event)
{
	if(drag_enter_event->mimeData()->hasFormat("text/uri-list"))
	{
		drag_enter_event->acceptProposedAction();
	}
}

void TUCBudgetEditor::dropEvent(
	QDropEvent 							*drop_event)
{
	QList<QUrl>							urls;
	
	urls = drop_event->mimeData()->urls();
	
	if(urls.size() == 1)
	{
		this->Load_Data((*urls.begin()).toLocalFile());
	}
}

void TUCBudgetEditor::Load_Data(
	const QString						&file_name)
{
	
	std::vector<TVariable>				variables;
	std::vector<TGroup>					groups;
	int									index;
	TInputData							input_data;
	
	if(file_name.length() == 0)
	{
		return;
	}
	
	
	if(d_data_modified)
	{
		d_msg_box->setText("Data Not Saved");
		d_msg_box->setInformativeText("Current data will be lost.  Do you want to continue?");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_msg_box->setDefaultButton(QMessageBox::No);
		d_msg_box->setIcon(QMessageBox::Question);
		
		if(d_msg_box->exec() == QMessageBox::No)
		{
			return;
		}
	}

	
	d_log_edit->clear();
	
	if(!d_ucbudget.Load(file_name,&variables,&groups,&input_data))
	{
		d_log_edit->appendPlainText(d_ucbudget.Get_Last_Error());
		d_tab_widget->setCurrentWidget(d_log_edit);
		
		return;
	}
	
	d_save_file_name = file_name;
	
	// update data path
	d_data_path = file_name;
	
	index = d_data_path.lastIndexOf('/');
	d_data_path.truncate(index);
	
	d_variable_widget->Set_Input_Data(input_data);
	
	d_variable_widget->Add_Optional_Variables(variables);
	d_reportgroup_widget->Set_Groups(groups);
	
	this->Report_Groups_Changed();
	
	this->Update_BudgetTree();
	
	d_data_modified = false;
	
	d_variable_widget->Update_Variable_List(d_ucbudget.Enum_Reportable_Components());
	d_reportgroup_widget->Update_Reportable_Components(d_ucbudget.Enum_Reportable_Components());
	
	d_save_action->setEnabled(d_data_modified);
	d_saveas_action->setEnabled(true);
	
	d_component_widget->setEnabled(false);
	d_component_item_widget->setEnabled(false);
}

void TUCBudgetEditor::Create_Report_Data(
	const QString						&group_name)
{
	TVariable							variable;
	std::vector<TVariable>::const_iterator var_iter;
	QString								component_name;
	QString								component_item_name;
	QString								description;
	int									cntr;
	int									count;
	double								value;
	bool								valid;
	QStringList							list;
	QStringList							variable_names;
	QStringList::ConstIterator			str_iter;
	QStringList::ConstIterator			str_list_iter;
	QStringList							reportable_component_list;
	TUCBudgetEditor::TReportData		report_data;
	
	this->Write_Log_Text(QString("Calculating data for group: %1").arg(group_name));
	
	// Generate base variable list
	
	this->Write_Log_Text("Global variables");

	d_base_variables.clear();
	count = d_variable_widget->Variable_Count();
	
	for(cntr = 0;cntr < count;++cntr)
	{
		variable = d_variable_widget->Variable(cntr);
		variable.value = d_variable_widget->Variable_Value(cntr,&valid);
		variable_names.push_back(variable.name);
	
		if(valid)
		{
			d_base_variables.push_back(variable);
			
			if(variable.name != QStringLiteral("CTEc")
			   && variable.name != QStringLiteral("CTEd")
			   && variable.name != QStringLiteral("Lc")
			   && variable.name != QStringLiteral("L"))		// These are updated locally.  Global value is zero.
			{
				this->Write_Log_Text(QString("Variable"),variable.name,QString("%1").arg(variable.value,0,'f',6));
			}
		}
	}
	
	// Generate items to be reported
	d_reporting_data.clear();
	d_reporting_variable_data.clear();
	
	list = d_reportgroup_widget->Group_List(group_name);
	
	for(str_iter = list.begin();str_iter != list.end();++str_iter)
	{
		this->Write_Log_Text(QString("Processing"),(*str_iter));

		if(d_ucbudget.Convert_From_Reportable((*str_iter),&component_name,&component_item_name))
		{
			report_data.component_name = component_name;
			report_data.component_item_name = component_item_name;
		}
		else
		{
			this->Write_Log_Text(QString("ERR:  Group item cannot be identified."));
			return;
		}
		
		report_data.identifier = d_report_widget->Current_Identifier(component_name);
		
		if(report_data.identifier.length())
		{
			this->Write_Log_Text(QString("Identifier"),report_data.identifier);
		}
		else
		{
			this->Write_Log_Text(QString("Identifier"),d_default_component_identifier_name);
		}
		
		report_data.length = d_ucbudget.Component_Length(component_name,report_data.identifier);
						
		// create local versions of 'CTEd', 'CTEc', and 'Lc' if required
		report_data.variables.clear();
		
		if(report_data.component_item_name.length())
		{
			variable_names = this->Variable_List(report_data.component_name,report_data.component_item_name,report_data.identifier);
		}
		else
		{
			variable_names = this->Variable_List(report_data.component_name,report_data.identifier);
		}
				
		if(variable_names.contains("CTEd"))
		{
			// add value for CTEd
			variable.name = "CTEd";
						
			value = d_ucbudget.Component_Expansion_Coefficient(component_name,report_data.identifier);
			value -= d_variable_widget->Machine_Expansion_Coefficient();
			
			variable.value = fabs(value);
			
			this->Write_Log_Text(QString("Local variable"),variable.name,QString("%1").arg(variable.value,0,'f',6));

			report_data.variables.push_back(variable);
		}
		
		if(variable_names.contains("CTEc"))
		{
			// add value for CTEc
			variable.name = "CTEc";
			
			variable.value = d_ucbudget.Component_Expansion_Coefficient(component_name,report_data.identifier);
			this->Write_Log_Text(QString("Local variable"),variable.name,QString("%1").arg(variable.value,0,'f',6));
			
			report_data.variables.push_back(variable);
		}

		if(variable_names.contains("Lc"))
		{
			// add value for Lc
			variable.name = "Lc";
			variable.value = 0.0;

			report_data.variables.push_back(variable);
		}
		
		if(variable_names.contains("L"))
		{
			// add value for L
			variable.name = "L";
			variable.value = 0.0;
			
			report_data.variables.push_back(variable);
		}
	
		d_reporting_data.push_back(report_data);
	}
	
	
	d_ucbudget.Reset_Expression_Variables();
	
	for(var_iter = d_base_variables.begin();var_iter != d_base_variables.end();++var_iter)
	{
		d_ucbudget.Add_Expression_Variable((*var_iter).name,(*var_iter).value);
	}

	// Get a list of non-global variables
	// Generate a list of reporting items required for variables
	list = this->Report_Data_Variable_List();

	for(str_iter = list.begin();str_iter != list.end();++str_iter)
	{
		report_data.variables.clear();
		
		description = d_variable_widget->Variable_Description((*str_iter));
		
		this->Write_Log_Text(QString("Processing"),QString("%1 (%2)").arg(description).arg(*str_iter));
		
		if(d_ucbudget.Convert_From_Reportable(description,&component_name,&component_item_name))
		{
			report_data.component_name = component_name;
			report_data.component_item_name = component_item_name;
		}
		else
		{
			this->Write_Log_Text(QString("ERR:  Group item cannot be identified."));
			return;
		}
		
		report_data.identifier = d_report_widget->Current_Identifier(component_name);
		
		if(report_data.identifier.length())
		{
			this->Write_Log_Text(QString("Identifier"),report_data.identifier);
		}
		else
		{
			this->Write_Log_Text(QString("Identifier"),d_default_component_identifier_name);
		}
		
		variable.name = (*str_iter);
		variable.description = description;
		variable.value = 0.0;

		report_data.variables.push_back(variable);

		d_reporting_variable_data.push_back(report_data);
	}
}

bool TUCBudgetEditor::Update_Report(
	const double						&length)
{
	std::vector<TVariable>::const_iterator var_iter;
	TReportWidget::TReportData			report_data;
	TReportWidget::TReportItemData		report_item_data;
	std::vector<TUCBudgetEditor::TReportData>::const_iterator rep_iter;
	QString								component_name;
	QString								component_item_name;
	QString								expression_type_text;
	bool								valid;
	double								dval;
	int									item_count;
	int									item_index;
	int									component_id;
	TUCBudgetData::TComponentItem		component_item;
	TUCBudgetData::TExpressionType		expression_type;
	bool								expression_type_init;
	

	d_report_widget->Start_Report_Update();
	
	this->Write_Log_Text(QString("Generating report"));
	this->Write_Log_Text(QString("Reporting length"),QString("%1").arg(length,0,'f',3));

	this->Write_Log_Text("Name","Unit Type","Expression Type","Value","Sensitivity","Distribution","Divisor","Source");

	// create variable report items first
	
	for(rep_iter = d_reporting_variable_data.begin();rep_iter != d_reporting_variable_data.end();++rep_iter)
	{
		report_data.report_items.clear();
				
		report_data.component_name = (*rep_iter).component_name;
		report_data.identifier = (*rep_iter).identifier;
		
		report_data.identifier_list.clear();
		report_data.identifier_list.push_back(d_default_component_identifier_name);
		report_data.identifier_list += d_ucbudget.Enum_Component_Identifiers((*rep_iter).component_name);
		
		report_data.is_variable = true;
		
		this->Write_Log_Text(report_data.component_name);
		
		for(var_iter = (*rep_iter).variables.begin();var_iter != (*rep_iter).variables.end();++var_iter)
		{
			dval = d_ucbudget.Evaluate_Reportable_Component((*var_iter).description,(*rep_iter).identifier,length,&valid);
			
			if(valid)
			{
				d_ucbudget.Add_Expression_Variable((*var_iter).name,dval);
				this->Write_Log_Text(QString(),QString("%1 = %2").arg((*var_iter).name).arg(dval,0,'f',6));
				
			}
			else
			{
				this->Write_Log_Text(QString("ERR:  Variable"),(*var_iter).name);
			}
		}

		
		component_id = d_ucbudget.Budget_Data()->Component_Index((*rep_iter).component_name);
		
		item_count = d_ucbudget.Budget_Data()->Component_Item_Count(component_id);
		
		for(item_index = 0;item_index < item_count;++item_index)
		{
			component_item = d_ucbudget.Budget_Data()->Component_Item(component_id,item_index);

			if((*rep_iter).component_item_name.length() != 0)
			{
				if((*rep_iter).component_item_name != component_item.name)
					continue;
			}
			
			expression_type = component_item.expression_type;
			
			if(expression_type == TUCBudgetData::CONSTANT_LINEAR || expression_type == TUCBudgetData::LENGTH_DEPENDENT_LINEAR)
			{
				expression_type_text = "Linear";
			}
			else if(expression_type == TUCBudgetData::CONSTANT_TEMPERATURE || expression_type == TUCBudgetData::LENGTH_DEPENDENT_TEMPERATURE)
			{
				expression_type_text = "Temperature";
			}
			
			report_item_data.component_item_name = component_item.name;
			
			if(component_item.expression_type == TUCBudgetData::CONSTANT_LINEAR ||
			   component_item.expression_type == TUCBudgetData::CONSTANT_TEMPERATURE)
			{
				report_item_data.value_type = "Constant";
			}
			else
			{
				report_item_data.value_type = "Length Dependent";
			}
			
			report_item_data.expression = d_ucbudget.Component_Item_Expression((*rep_iter).component_name,component_item.name,(*rep_iter).identifier);
			
			report_item_data.type = d_ucbudget.UCType(component_item.uctype);
			
			report_item_data.sensitivity = QString("%1").arg(component_item.sensitivity,0,'f',6);

			report_item_data.distribution = d_ucbudget.Distribution_Type(component_item.distribution);
			
			report_item_data.distribution_divisor = d_ucbudget.Distribution_Divisor(component_item.distribution);

			dval = d_ucbudget.Evaluate_Component_Item((*rep_iter).component_name,component_item.name,(*rep_iter).identifier,length,&valid);
			
			if(valid)
			{
				report_item_data.value_string = QString("%1").arg(dval,0,'f',6);
				report_item_data.value = dval;
			}
			else
			{
				report_item_data.value_string = "ERR";
				report_item_data.value = 0.0;
			}
			
			this->Write_Log_Text(QString(),
								 expression_type_text,
								 report_item_data.value_type,
								 report_item_data.value_string,
								 report_item_data.sensitivity,
								 report_item_data.distribution,
								 QString("%1").arg(report_item_data.distribution_divisor,0,'f',6),
								 component_item.name);
			
			report_data.report_items.push_back(report_item_data);
		}
		
		d_report_widget->Add_Report_Data(report_data);
	}
	
	// create report items
	expression_type_init = false;

	for(rep_iter = d_reporting_data.begin();rep_iter != d_reporting_data.end();++rep_iter)
	{
		report_data.report_items.clear();
		
		report_data.component_name = (*rep_iter).component_name;
		report_data.identifier = (*rep_iter).identifier;
		
		report_data.identifier_list.clear();
		report_data.identifier_list.push_back(d_default_component_identifier_name);
		report_data.identifier_list += d_ucbudget.Enum_Component_Identifiers((*rep_iter).component_name);
		
		report_data.is_variable = false;

		this->Write_Log_Text(report_data.component_name);
	
		for(var_iter = (*rep_iter).variables.begin();var_iter != (*rep_iter).variables.end();++var_iter)
		{
			
			// special case for 'Lc' variable
			if((*var_iter).name == QStringLiteral("Lc"))
			{
				if((*rep_iter).length > 0)
				{
					// find the number of times the item length will fit into the reported length
					// a staged step gauge of 600 mm should report a value of 1 with an end position of for Lc (zero based index)
					// the count can only increment when the measurement length exceeds the equipment length.  A value of 0.1% is assumed
					dval = static_cast<double>(static_cast<int>(length / ((*rep_iter).length * 1.001)));
				}
				else
				{
					dval = 0.0;
				}
				
				d_ucbudget.Add_Expression_Variable((*var_iter).name,dval);
				this->Write_Log_Text(QString(),QString("%1 = %2").arg((*var_iter).name).arg(dval,0,'f',6));
			}
			// special case for 'L' variable
			else if((*var_iter).name == QStringLiteral("L"))
			{
				d_ucbudget.Add_Expression_Variable((*var_iter).name,length);
				this->Write_Log_Text(QString(),QString("%1 = %2").arg((*var_iter).name).arg(length,0,'f',6));
			}
			else
			{
				d_ucbudget.Add_Expression_Variable((*var_iter).name,(*var_iter).value);
				this->Write_Log_Text(QString(),QString("%1 = %2").arg((*var_iter).name).arg((*var_iter).value,0,'f',6));
			}
		}

		
		component_id = d_ucbudget.Budget_Data()->Component_Index((*rep_iter).component_name);
		
		item_count = d_ucbudget.Budget_Data()->Component_Item_Count(component_id);
		
		for(item_index = 0;item_index < item_count;++item_index)
		{
			component_item = d_ucbudget.Budget_Data()->Component_Item(component_id,item_index);
			
			if((*rep_iter).component_item_name.length() != 0)
			{
				if((*rep_iter).component_item_name != component_item.name)
					continue;
			}
			
			if(!expression_type_init)
			{
				expression_type = component_item.expression_type;
				expression_type_init = true;
				
				if(expression_type == TUCBudgetData::CONSTANT_LINEAR || expression_type == TUCBudgetData::LENGTH_DEPENDENT_LINEAR)
				{
					expression_type_text = "Linear";
				}
				else if(expression_type == TUCBudgetData::CONSTANT_TEMPERATURE || expression_type == TUCBudgetData::LENGTH_DEPENDENT_TEMPERATURE)
				{
					expression_type_text = "Temperature";
				}
			}
			else
			{
				if(expression_type == TUCBudgetData::CONSTANT_LINEAR || expression_type == TUCBudgetData::LENGTH_DEPENDENT_LINEAR)
				{
					expression_type_text = "Linear";
					
					if(component_item.expression_type == TUCBudgetData::CONSTANT_TEMPERATURE || component_item.expression_type == TUCBudgetData::LENGTH_DEPENDENT_TEMPERATURE)
					{
						this->Write_Log_Text(QString("ERR:  Cannot mix linear and temperature expressions.  Use a variable to combine un-related items"));
						d_report_widget->End_Report_Update();
						d_report_widget->Clear_Report();
						return false;
					}
				}
				else if(expression_type == TUCBudgetData::CONSTANT_TEMPERATURE || expression_type == TUCBudgetData::LENGTH_DEPENDENT_TEMPERATURE)
				{
					expression_type_text = "Temperature";
					
					if(component_item.expression_type == TUCBudgetData::CONSTANT_LINEAR || component_item.expression_type == TUCBudgetData::LENGTH_DEPENDENT_LINEAR)
					{
						this->Write_Log_Text(QString("ERR:  Cannot mix linear and temperature expressions.  Use a variable to combine un-related items"));
						d_report_widget->End_Report_Update();
						d_report_widget->Clear_Report();
						return false;
					}
				}
			}
			
			report_item_data.component_item_name = component_item.name;
			
			if(component_item.expression_type == TUCBudgetData::CONSTANT_LINEAR ||
			   component_item.expression_type == TUCBudgetData::CONSTANT_TEMPERATURE)
			{
				report_item_data.value_type = "Constant";
			}
			else
			{
				report_item_data.value_type = "Length Dependent";
			}
			
			report_item_data.expression = d_ucbudget.Component_Item_Expression((*rep_iter).component_name,component_item.name,(*rep_iter).identifier);
			
			report_item_data.type = d_ucbudget.UCType(component_item.uctype);

			report_item_data.sensitivity = QString("%1").arg(component_item.sensitivity,0,'f',6);

			report_item_data.distribution = d_ucbudget.Distribution_Type(component_item.distribution);
	
			dval = d_ucbudget.Evaluate_Component_Item((*rep_iter).component_name,component_item.name,(*rep_iter).identifier,length,&valid);
			
			if(valid)
			{
				report_item_data.value_string = QString("%1").arg(dval,0,'f',6);
				report_item_data.value = dval;
			}
			else
			{
				report_item_data.value_string = "ERR";
				report_item_data.value = 0.0;
			}
			
			this->Write_Log_Text(QString(),
								 expression_type_text,
								 report_item_data.value_type,
								 report_item_data.value_string,
								 report_item_data.sensitivity,
								 report_item_data.distribution,
								 QString("%1").arg(report_item_data.distribution_divisor,0,'f',6),
								 component_item.name);

			report_data.report_items.push_back(report_item_data);
			
		}
		
		d_report_widget->Add_Report_Data(report_data);
	}
	
	d_report_widget->End_Report_Update();
	
	return true;
}


bool TUCBudgetEditor::Update_Export(
	const double						&length)
{
	std::vector<TVariable>::const_iterator var_iter;
	TReportWidget::TExportVariableData  export_variable_data;
	TReportWidget::TReportData			report_data;
	TReportWidget::TReportItemData		report_item_data;
	std::vector<TUCBudgetEditor::TReportData>::const_iterator rep_iter;
	QString								component_name;
	QString								component_item_name;
	QString								expression_type_text;
	bool								valid;
	double								dval;
	int									item_count;
	int									item_index;
	int									component_id;
	TUCBudgetData::TComponentItem		component_item;
	TUCBudgetData::TExpressionType		expression_type;
	bool								expression_type_init;
	
	// Reset expression variables and add defaults
	d_ucbudget.Reset_Expression_Variables();
	
	for(var_iter = d_base_variables.begin();var_iter != d_base_variables.end();++var_iter)
	{
		export_variable_data.variable_name = (*var_iter).name;
		
		export_variable_data.values.clear();
		export_variable_data.values.push_back((*var_iter).value);
		
		d_report_widget->Add_Export_Variable_Data(export_variable_data);
		
		d_ucbudget.Add_Expression_Variable((*var_iter).name,(*var_iter).value);
	}
		
	this->Write_Log_Text(QString());
	this->Write_Log_Text("Position","Value","Component","Component Item");

	// create variable report items first
	
	for(rep_iter = d_reporting_variable_data.begin();rep_iter != d_reporting_variable_data.end();++rep_iter)
	{
		report_data.report_items.clear();
		
		report_data.component_name = (*rep_iter).component_name;
		report_data.identifier = (*rep_iter).identifier;
		
		report_data.identifier_list.clear();
		report_data.identifier_list.push_back(d_default_component_identifier_name);
		report_data.identifier_list += d_ucbudget.Enum_Component_Identifiers((*rep_iter).component_name);
		
		report_data.is_variable = true;
			
		for(var_iter = (*rep_iter).variables.begin();var_iter != (*rep_iter).variables.end();++var_iter)
		{
			dval = d_ucbudget.Evaluate_Reportable_Component((*var_iter).description,(*rep_iter).identifier,length,&valid);
			
			if(valid)
			{
				export_variable_data.component_name = (*rep_iter).component_name;
				export_variable_data.variable_name = (*var_iter).name;
				
				export_variable_data.values.clear();
				export_variable_data.values.push_back(dval);
				
				d_report_widget->Add_Export_Variable_Data(export_variable_data);
				d_ucbudget.Add_Expression_Variable((*var_iter).name,dval);
				
			}
			else
			{
				this->Write_Log_Text(QString("ERR:  Variable"),(*var_iter).name);
				d_ucbudget.Add_Expression_Variable((*var_iter).name,0.0);
				
			}
		}
		
		component_id = d_ucbudget.Budget_Data()->Component_Index((*rep_iter).component_name);
		
		item_count = d_ucbudget.Budget_Data()->Component_Item_Count(component_id);
		
		for(item_index = 0;item_index < item_count;++item_index)
		{
			component_item = d_ucbudget.Budget_Data()->Component_Item(component_id,item_index);
			
			if((*rep_iter).component_item_name.length() != 0)
			{
				if((*rep_iter).component_item_name != component_item.name)
					continue;
			}
			
			expression_type = component_item.expression_type;
			
			if(expression_type == TUCBudgetData::CONSTANT_LINEAR || expression_type == TUCBudgetData::LENGTH_DEPENDENT_LINEAR)
			{
				expression_type_text = "Linear";
			}
			else if(expression_type == TUCBudgetData::CONSTANT_TEMPERATURE || expression_type == TUCBudgetData::LENGTH_DEPENDENT_TEMPERATURE)
			{
				expression_type_text = "Temperature";
			}
			
			report_item_data.component_item_name = component_item.name;
			
			if(component_item.expression_type == TUCBudgetData::CONSTANT_LINEAR ||
			   component_item.expression_type == TUCBudgetData::CONSTANT_TEMPERATURE)
			{
				report_item_data.value_type = "Constant";
			}
			else
			{
				report_item_data.value_type = "Length Dependent";
			}
			
			report_item_data.expression = d_ucbudget.Component_Item_Expression((*rep_iter).component_name,component_item.name,(*rep_iter).identifier);
			
			report_item_data.sensitivity = QString("%1").arg(component_item.sensitivity,0,'f',6);
			
			report_item_data.type = d_ucbudget.UCType(component_item.uctype);
			
			report_item_data.distribution = d_ucbudget.Distribution_Type(component_item.distribution);
			
			report_item_data.distribution_divisor = d_ucbudget.Distribution_Divisor(component_item.distribution);
			
			dval = d_ucbudget.Evaluate_Component_Item((*rep_iter).component_name,component_item.name,(*rep_iter).identifier,length,&valid);
			
			if(valid)
			{
				report_item_data.value = dval;
			}
			else
			{
				report_item_data.value = 0;
			}
			
			
			this->Write_Log_Text(QString("%1 mm").arg(length,0,'f',3),
								 QString("%1").arg(report_item_data.value,0,'f',6),
								 report_data.component_name,
								 component_item.name);

			report_data.report_items.push_back(report_item_data);
		}
		
		d_report_widget->Add_Export_Data(report_data,length);
	}
	
	// create report items
	expression_type_init = false;
	
	for(rep_iter = d_reporting_data.begin();rep_iter != d_reporting_data.end();++rep_iter)
	{
		report_data.report_items.clear();
		
		report_data.component_name = (*rep_iter).component_name;
		report_data.identifier = (*rep_iter).identifier;
		
		report_data.identifier_list.clear();
		report_data.identifier_list.push_back(d_default_component_identifier_name);
		report_data.identifier_list += d_ucbudget.Enum_Component_Identifiers((*rep_iter).component_name);
		
		report_data.is_variable = false;
		
		
		for(var_iter = (*rep_iter).variables.begin();var_iter != (*rep_iter).variables.end();++var_iter)
		{
			
			// special case for 'Lc' variable
			if((*var_iter).name == QStringLiteral("Lc"))
			{
				if((*rep_iter).length > 0)
				{
					// find the number of times the item length will fit into the reported length
					// a staged step gauge of 600 mm should report a value of 1 with an end position of for Lc (zero based index)
					// the count can only increment when the measurement length exceeds the equipment length.  A value of 0.1% is assumed
					dval = static_cast<double>(static_cast<int>(length / ((*rep_iter).length * 1.001)));
				}
				else
				{
					dval = 0.0;
				}
				
				export_variable_data.component_name = (*rep_iter).component_name;
				export_variable_data.variable_name = (*var_iter).name;
				
				export_variable_data.values.clear();
				export_variable_data.values.push_back(dval);
				
				d_report_widget->Add_Export_Variable_Data(export_variable_data);
				
				d_ucbudget.Add_Expression_Variable((*var_iter).name,dval);
			}
			// special case for 'L' variable
			else if((*var_iter).name == QStringLiteral("L"))
			{
				export_variable_data.component_name = (*rep_iter).component_name;
				export_variable_data.variable_name = (*var_iter).name;
				
				export_variable_data.values.clear();
				export_variable_data.values.push_back(length);
				
				d_report_widget->Add_Export_Variable_Data(export_variable_data);
				
				d_ucbudget.Add_Expression_Variable((*var_iter).name,length);
			}
			else
			{
				export_variable_data.component_name = (*rep_iter).component_name;
				export_variable_data.variable_name = (*var_iter).name;
				
				export_variable_data.values.clear();
				export_variable_data.values.push_back((*var_iter).value);
				
				d_report_widget->Add_Export_Variable_Data(export_variable_data);
								
				d_ucbudget.Add_Expression_Variable((*var_iter).name,(*var_iter).value);
			}
		}
		
		
		component_id = d_ucbudget.Budget_Data()->Component_Index((*rep_iter).component_name);
		
		item_count = d_ucbudget.Budget_Data()->Component_Item_Count(component_id);
		
		for(item_index = 0;item_index < item_count;++item_index)
		{
			component_item = d_ucbudget.Budget_Data()->Component_Item(component_id,item_index);
			
			if((*rep_iter).component_item_name.length() != 0)
			{
				if((*rep_iter).component_item_name != component_item.name)
					continue;
			}
			
			if(!expression_type_init)
			{
				expression_type = component_item.expression_type;
				expression_type_init = true;
				
				if(expression_type == TUCBudgetData::CONSTANT_LINEAR || expression_type == TUCBudgetData::LENGTH_DEPENDENT_LINEAR)
				{
					expression_type_text = "Linear";
				}
				else if(expression_type == TUCBudgetData::CONSTANT_TEMPERATURE || expression_type == TUCBudgetData::LENGTH_DEPENDENT_TEMPERATURE)
				{
					expression_type_text = "Temperature";
				}
			}
			else
			{
				if(expression_type == TUCBudgetData::CONSTANT_LINEAR || expression_type == TUCBudgetData::LENGTH_DEPENDENT_LINEAR)
				{
					expression_type_text = "Linear";

					if(component_item.expression_type == TUCBudgetData::CONSTANT_TEMPERATURE || component_item.expression_type == TUCBudgetData::LENGTH_DEPENDENT_TEMPERATURE)
					{
						this->Write_Log_Text(QString("ERR:  Cannot mix linear and temperature expressions.  Use a variable to combine un-related items"));

						return false;
					}
				}
				else if(expression_type == TUCBudgetData::CONSTANT_TEMPERATURE || expression_type == TUCBudgetData::LENGTH_DEPENDENT_TEMPERATURE)
				{
					expression_type_text = "Temperature";
					
					if(component_item.expression_type == TUCBudgetData::CONSTANT_LINEAR || component_item.expression_type == TUCBudgetData::LENGTH_DEPENDENT_LINEAR)
					{
						this->Write_Log_Text(QString("ERR:  Cannot mix linear and temperature expressions.  Use a variable to combine un-related items"));

						return false;
					}
				}
			}
						
			report_item_data.component_item_name = component_item.name;
			
			if(component_item.expression_type == TUCBudgetData::CONSTANT_LINEAR ||
			   component_item.expression_type == TUCBudgetData::CONSTANT_TEMPERATURE)
			{
				report_item_data.value_type = "Constant";
			}
			else
			{
				report_item_data.value_type = "Length Dependent";
			}
			
			report_item_data.expression = d_ucbudget.Component_Item_Expression((*rep_iter).component_name,component_item.name,(*rep_iter).identifier);
			
			report_item_data.sensitivity = QString("%1").arg(component_item.sensitivity,0,'f',6);
			
			report_item_data.type = d_ucbudget.UCType(component_item.uctype);
			
			report_item_data.distribution = d_ucbudget.Distribution_Type(component_item.distribution);
			
			report_item_data.distribution_divisor = d_ucbudget.Distribution_Divisor(component_item.distribution);
			
			dval = d_ucbudget.Evaluate_Component_Item((*rep_iter).component_name,component_item.name,(*rep_iter).identifier,length,&valid);
			
			if(valid)
			{
				report_item_data.value = dval;
			}
			else
			{
				report_item_data.value = 0;
			}
			
			
			this->Write_Log_Text(QString("%1 mm").arg(length,0,'f',3),
								 QString("%1").arg(report_item_data.value,0,'f',6),
								 report_data.component_name,
								 component_item.name);

			report_data.report_items.push_back(report_item_data);
			
		}
		
		d_report_widget->Add_Export_Data(report_data,length);
	}
	
	return true;
}

QStringList TUCBudgetEditor::Report_Data_Variable_List(void)
{
	std::vector<TUCBudgetEditor::TReportData>::const_iterator iter;
	std::vector<TVariable>::const_iterator var_iter;
	QStringList							list;
	
	for(iter = d_reporting_data.begin();iter != d_reporting_data.end();++iter)
	{
		if((*iter).component_item_name.length())
		{
			list += this->Variable_List((*iter).component_name,(*iter).component_item_name,(*iter).identifier);
		}
		else
		{
			list += this->Variable_List((*iter).component_name,(*iter).identifier);
		}
	}
	
	list.removeDuplicates();
	
	// remove all variables included from base variable list;
	for(var_iter = d_base_variables.begin();var_iter != d_base_variables.end();++var_iter)
	{
		this->Remove_From_List(&list,(*var_iter).name);
	}

	return list;
}

QStringList TUCBudgetEditor::Variable_List(
	const QString						&component_name,
	const QString						&identifier)
{
	int									component_id;
	int									item_count;
	int									item_cntr;
	QStringList							list;
	
	component_id = d_ucbudget.Budget_Data()->Component_Index(component_name);
	
	item_count = d_ucbudget.Budget_Data()->Component_Item_Count(component_id);
	
	for(item_cntr = 0;item_cntr < item_count;++item_cntr)
	{
		list += this->Variable_List(component_name,d_ucbudget.Budget_Data()->Component_Item(component_id,item_cntr).name,identifier);
	}
	
	list.removeDuplicates();
	
	return list;
}


QStringList TUCBudgetEditor::Variable_List(
	const QString						&component_name,
	const QString						&component_item_name,
	const QString						&identifier) const
{
	QStringList							init_list;
	QStringList							list;
	QStringList::ConstIterator			str_iter;
	QString								text;
	bool								result;
	QString								expression;
	
	expression = d_ucbudget.Component_Item_Expression(component_name,component_item_name,identifier,&result);
	
	// remove any function operators:  *+-/()^, sumsq sqrt
	// TODO.  should get operator list from expression evaluation routine.
	
	expression.replace("sumsq("," ",Qt::CaseInsensitive);
	expression.replace("sqrt("," ",Qt::CaseInsensitive);
	expression.replace('*',' ');
	expression.replace('+',' ');
	expression.replace('-',' ');
	expression.replace('/',' ');
	expression.replace('^',' ');
	expression.replace(',',' ');
	expression.replace('(',' ');
	expression.replace(')',' ');

	expression = expression.simplified();

	init_list = expression.split(' ');
	list.clear();
	
	for(str_iter = init_list.begin();str_iter != init_list.end();++str_iter)
	{
		if((*str_iter).length() > 0 && (*str_iter)[0].isLetter() == true)
		{
			list.push_back(*str_iter);
		}
	}

	list.removeDuplicates();
	
	return list;
}

void TUCBudgetEditor::Remove_From_List(
	QStringList							*list,
	const QString						&item) const
{
	QList<QString>::iterator			str_iter;
	
	for(str_iter = list->begin();str_iter != list->end();++str_iter)
	{
		if(0 == item.compare((*str_iter),Qt::CaseInsensitive))
		{
			list->erase(str_iter);
			return;
		}
	}
}

void TUCBudgetEditor::Write_Log_Text(
	const QString						&t1,
	const QString						&t2,
	const QString						&t3,
	const QString						&t4,
	const QString						&t5,
	const QString						&t6,
	const QString						&t7,
	const QString						&t8)
{
	QString								line;
	QString								text;
	int									width_adjustment(0);
	static const int					FIELD_WIDTH(20);
	
	text = QString("%1").arg(t1,-FIELD_WIDTH);
	line = text;
	text.length() > FIELD_WIDTH ? width_adjustment = text.length() - FIELD_WIDTH : width_adjustment = 0;
	
	if(t2.length())
	{
		text = QString("%1").arg(t2,-(FIELD_WIDTH - width_adjustment));
		line += text;
		text.length() > FIELD_WIDTH ? width_adjustment = text.length() - FIELD_WIDTH : width_adjustment = 0;
	}
	if(t3.length())
	{
		text = QString("%1").arg(t3,-(FIELD_WIDTH - width_adjustment));
		line += text;
		text.length() > FIELD_WIDTH ? width_adjustment = text.length() - FIELD_WIDTH : width_adjustment = 0;
	}
	if(t4.length())
	{
		text = QString("%1").arg(t4,-(FIELD_WIDTH - width_adjustment));
		line += text;
		text.length() > FIELD_WIDTH ? width_adjustment = text.length() - FIELD_WIDTH : width_adjustment = 0;
	}
	if(t5.length())
	{
		text = QString("%1").arg(t5,-(FIELD_WIDTH - width_adjustment));
		line += text;
		text.length() > FIELD_WIDTH ? width_adjustment = text.length() - FIELD_WIDTH : width_adjustment = 0;
	}
	if(t6.length())
	{
		text = QString("%1").arg(t6,-(FIELD_WIDTH - width_adjustment));
		line += text;
		text.length() > FIELD_WIDTH ? width_adjustment = text.length() - FIELD_WIDTH : width_adjustment = 0;
	}
	if(t7.length())
	{
		text = QString("%1").arg(t7,-(FIELD_WIDTH - width_adjustment));
		line += text;
		text.length() > FIELD_WIDTH ? width_adjustment = text.length() - FIELD_WIDTH : width_adjustment = 0;
	}
	if(t8.length())
	{
		text = QString("%1").arg(t8,-(FIELD_WIDTH - width_adjustment));
		line += text;
		text.length() > FIELD_WIDTH ? width_adjustment = text.length() - FIELD_WIDTH : width_adjustment = 0;
	}
	
	d_log_edit->appendPlainText(line);
}


void TUCBudgetEditor::Update_BudgetTree(void)
{
	QStringList							components;
	QStringList							component_items;
	QList<QString>::const_iterator		component_iter;
	QList<QString>::const_iterator		component_item_iter;
	QTreeWidgetItem						*root;
	QTreeWidgetItem						*tree_component;
	QTreeWidgetItem						*tree_component_item;

	d_budget_tree->clear();
	
	components = d_ucbudget.Enum_Components();
	
	root = new QTreeWidgetItem(d_budget_tree);
	root->setText(0,QStringLiteral("UC Budget Data"));
		
	for(component_iter = components.begin();component_iter != components.end();++component_iter)
	{
		tree_component = new QTreeWidgetItem(root);
		tree_component->setText(0,(*component_iter));
		tree_component->setFlags(tree_component->flags() | Qt::ItemIsEditable);

		component_items = d_ucbudget.Enum_Component_Items((*component_iter));
				
		for(component_item_iter = component_items.begin();
			component_item_iter != component_items.end();
			++component_item_iter)
		{
			tree_component_item = new QTreeWidgetItem(tree_component);
			tree_component_item->setText(0,(*component_item_iter));
			tree_component_item->setFlags(tree_component_item->flags() | Qt::ItemIsEditable);
		}
	}

	d_budget_tree->expandAll();
}

void TUCBudgetEditor::Select_BudgetTree_Item(
	const QString 						&component,
	const QString 						&component_item)
{
	QTreeWidgetItem						*root;
	QTreeWidgetItem						*tree_component;
	QTreeWidgetItem						*tree_component_item;
	int									count;
	int									cntr;
	
	d_current_item_index = -1;

	if(d_budget_tree->topLevelItemCount() == 1)
	{
		root = d_budget_tree->topLevelItem(0);
		
		if(component.length())
		{
			count = root->childCount();
			
			for(cntr = 0;cntr < count;++cntr)
			{
				tree_component = root->child(cntr);
				
				if(component.compare(tree_component->text(0)) == 0)
				{
					if(component_item.length())
					{
						count = tree_component->childCount();
						
						for(cntr = 0;cntr < count;++cntr)
						{
							tree_component_item = tree_component->child(cntr);
							
							if(component_item.compare(tree_component_item->text(0)) == 0)
							{
								d_current_item_index = 2;
								d_budget_tree->setCurrentItem(tree_component_item);
								return;
							}
						}
					}
					
					d_current_item_index = 1;
					d_budget_tree->setCurrentItem(tree_component);
					return;
				}
			}
		}
		d_current_item_index = 0;
		d_budget_tree->setCurrentItem(root);
	}
}
