/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////



#include <QVariant>
#include <QAction>
#include <QApplication>
#include <QButtonGroup>
#include <QComboBox>
#include <QFrame>
#include <QGridLayout>
#include <QHeaderView>
#include <QLabel>
#include <QLineEdit>
#include <QSpacerItem>
#include <QToolButton>
#include <QVBoxLayout>
#include <QFont>
#include <math.h>
#include <assert.h>

#include "../../core/optioneditor.h"

#include "variablewidget.h"
#include "namedialog.h"

TVariableWidget::TVariableWidget(
	const QWidget						*parent,
	Qt::WindowFlags						flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	QLabel								*input_label;
	QLabel								*machine_expansion_label;
	QLabel								*machine_expansion_unit_label;
	QLabel								*machine_scale_label;
	QLabel								*machine_scale_unit_label;
	QLabel								*minimum_temperature_label;
	QLabel								*minimum_temperature_unit_label;
	QLabel								*maximum_temperature_label;
	QLabel								*maximum_temperature_unit_label;
	QLabel								*gradient_temperature_label;
	QLabel								*gradient_temperature_unit_label;
	QFrame								*separator_line;
	QLabel								*variables_label;
	QLabel								*mean_stdev_label;
	QLabel								*mean_stdev_unit_label;
	QLabel								*start_position_label;
	QLabel								*end_position_label;
	QLabel								*increment_label;
	QLabel								*start_position_unit_label;
	QLabel								*end_position_unit_label;
	QLabel								*increment_unit_label;
	QWidget								*input_data_widget;
	QWidget								*variable_container_widget;
	QGridLayout							*input_data_layout;
	QVBoxLayout							*variable_container_vlayout;
	QHBoxLayout							*variable_button_hlayout;
	QSpacerItem							*input_vspacer;
	QSpacerItem							*variable_button_hspacer;	
	QFont								title_font;
	
	title_font.setPointSize(18);
	title_font.setUnderline(true);

	widget_layout = new QGridLayout(this);

	input_label = new QLabel(this);
	input_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	input_label->setFont(title_font);
	widget_layout->addWidget(input_label,0,0,1,1);
	
	input_data_widget = new QWidget(this);
	
	input_data_layout = new QGridLayout(input_data_widget);
	input_data_layout->setContentsMargins(0,0,0,0);
	
	machine_expansion_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(machine_expansion_label,0,0,1,1);
	
	d_machine_expansion_edit = new QLineEdit(input_data_widget);
	input_data_layout->addWidget(d_machine_expansion_edit,0,1,1,1);
	
	machine_expansion_unit_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(machine_expansion_unit_label,0,2,1,1);
		
	machine_scale_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(machine_scale_label,1,0,1,1);
	
	d_machine_scale_edit = new QLineEdit(input_data_widget);
	input_data_layout->addWidget(d_machine_scale_edit,1,1,1,1);
	
	machine_scale_unit_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(machine_scale_unit_label,1,2,1,1);
	
	minimum_temperature_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(minimum_temperature_label,2,0,1,1);
	
	d_minimum_temperature_edit = new QLineEdit(input_data_widget);
	input_data_layout->addWidget(d_minimum_temperature_edit,2,1,1,1);
	
	minimum_temperature_unit_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(minimum_temperature_unit_label,2,2,1,1);
	
	maximum_temperature_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(maximum_temperature_label,3,0,1,1);
	
	d_maximum_temperature_edit = new QLineEdit(input_data_widget);
	input_data_layout->addWidget(d_maximum_temperature_edit,3,1,1,1);
	
	maximum_temperature_unit_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(maximum_temperature_unit_label,3,2,1,1);

	gradient_temperature_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(gradient_temperature_label,4,0,1,1);

	d_gradient_temperature_edit = new QLineEdit(input_data_widget);
	input_data_layout->addWidget(d_gradient_temperature_edit,4,1,1,1);
	
	gradient_temperature_unit_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(gradient_temperature_unit_label,4,2,1,1);
	
	mean_stdev_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(mean_stdev_label,5,0,1,1);
	
	d_mean_stdev_edit = new QLineEdit(input_data_widget);
	input_data_layout->addWidget(d_mean_stdev_edit,5,1,1,1);
	
	mean_stdev_unit_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(mean_stdev_unit_label,5,2,1,1);
	
	start_position_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(start_position_label,6,0,1,1);
	
	d_start_position_edit = new QLineEdit(input_data_widget);
	d_start_position_edit->setEnabled(false);
	input_data_layout->addWidget(d_start_position_edit,6,1,1,1);
	
	start_position_unit_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(start_position_unit_label,6,2,1,1);
	
	end_position_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(end_position_label,7,0,1,1);
	
	d_end_position_edit = new QLineEdit(input_data_widget);
	input_data_layout->addWidget(d_end_position_edit,7,1,1,1);
	
	end_position_unit_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(end_position_unit_label,7,2,1,1);
	
	increment_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(increment_label,8,0,1,1);
	
	d_increment_edit = new QLineEdit(input_data_widget);
	input_data_layout->addWidget(d_increment_edit,8,1,1,1);
	
	increment_unit_label = new QLabel(input_data_widget);
	input_data_layout->addWidget(increment_unit_label,8,2,1,1);
	
	input_vspacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	input_data_layout->addItem(input_vspacer,9,0,1,3);

	widget_layout->addWidget(input_data_widget,1,0,1,1);

	separator_line = new QFrame(this);
	separator_line->setFrameShape(QFrame::VLine);
	separator_line->setFrameShadow(QFrame::Sunken);
	widget_layout->addWidget(separator_line,0,1,2,1);
	
	variables_label = new QLabel(this);
	variables_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	variables_label->setFont(title_font);
	widget_layout->addWidget(variables_label,0,2,1,1);
	
	variable_container_widget = new QWidget(this);
	variable_container_vlayout = new QVBoxLayout(variable_container_widget);
	
	variable_container_vlayout->setContentsMargins(0,0,0,0);
	variable_container_vlayout->setSpacing(4);

	d_variable_table = new TOptionEditorWidget(variable_container_widget);
	variable_container_vlayout->addWidget(d_variable_table);
	
	variable_button_hlayout = new QHBoxLayout;
	
	d_add_variable_button = new QToolButton(variable_container_widget);
	d_add_variable_button->setFixedSize(16,16);
	variable_button_hlayout->addWidget(d_add_variable_button);
		
	d_rename_variable_button = new QToolButton(variable_container_widget);
	d_rename_variable_button->setFixedSize(16,16);
	variable_button_hlayout->addWidget(d_rename_variable_button);

	d_remove_variable_button = new QToolButton(variable_container_widget);
	d_remove_variable_button->setFixedSize(16,16);
	variable_button_hlayout->addWidget(d_remove_variable_button);
	
	variable_button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	variable_button_hlayout->addItem(variable_button_hspacer);
	
	variable_container_vlayout->addLayout(variable_button_hlayout);

	widget_layout->addWidget(variable_container_widget,1,2,1,1);

	this->Initialize_Option_Editor();
	d_variable_list.push_back(QString("[no selection]"));
	
	d_add_variable_button->setIcon(QIcon(QStringLiteral(":/gui/add.png")));
	d_rename_variable_button->setIcon(QIcon(QStringLiteral(":/gui/edit.png")));
	d_remove_variable_button->setIcon(QIcon(QStringLiteral(":/gui/remove.png")));

	input_label->setText(QStringLiteral("Variable Values:"));
	machine_expansion_label->setText(QStringLiteral("Machine Expansion Coefficient:"));
	machine_expansion_unit_label->setText(QStringLiteral("um/m/˚C"));
	machine_scale_label->setText(QStringLiteral("Machine Scale Resolution:"));
	machine_scale_unit_label->setText(QStringLiteral("mm"));
	minimum_temperature_label->setText(QStringLiteral("Minimum Temperature:"));
	minimum_temperature_unit_label->setText(QStringLiteral("C"));
	maximum_temperature_label->setText(QStringLiteral("Maximum Temperature:"));
	maximum_temperature_unit_label->setText(QStringLiteral("C"));
	gradient_temperature_label->setText(QStringLiteral("Temperature Gradient:"));
	gradient_temperature_unit_label->setText(QStringLiteral("C"));
	variables_label->setText(QStringLiteral("Variable Entries:"));
	mean_stdev_label->setText(QStringLiteral("Mean Standard Deviation:"));
	start_position_label->setText(QStringLiteral("Start Position:"));
	end_position_label->setText(QStringLiteral("End Position:"));
	increment_label->setText(QStringLiteral("Increment:"));
	mean_stdev_unit_label->setText(QStringLiteral("mm"));
	start_position_unit_label->setText(QStringLiteral("mm"));
	end_position_unit_label->setText(QStringLiteral("mm"));
	increment_unit_label->setText(QStringLiteral("mm"));
	
	// defaults
	d_machine_expansion_edit->setText("11.5");
	d_machine_scale_edit->setText("0.001");
	d_minimum_temperature_edit->setText("20.0");
	d_maximum_temperature_edit->setText("20.0");
	d_gradient_temperature_edit->setText("0.0");
	d_mean_stdev_edit->setText("0.000");
	d_start_position_edit->setText("0.000");
	d_end_position_edit->setText("1000.000");
	d_increment_edit->setText("50.000");
	
	d_rename_variable_button->setEnabled(false);	
	d_remove_variable_button->setEnabled(false);

	connect(d_variable_table,&TOptionEditorWidget::itemSelectionChanged,this,&TVariableWidget::Variable_Selection_Changed);
	connect(d_variable_table,&TOptionEditorWidget::SelectionListItemChanged,this,&TVariableWidget::Variable_Selection_List_Changed);
	connect(d_variable_table,&TOptionEditorWidget::itemDoubleClicked,this,&TVariableWidget::Rename_Variable);

	connect(d_machine_expansion_edit,&QLineEdit::editingFinished,this,&TVariableWidget::Variable_Value_Changed);
	connect(d_machine_scale_edit,&QLineEdit::editingFinished,this,&TVariableWidget::Variable_Value_Changed);
	connect(d_minimum_temperature_edit,&QLineEdit::editingFinished,this,&TVariableWidget::Variable_Value_Changed);
	connect(d_maximum_temperature_edit,&QLineEdit::editingFinished,this,&TVariableWidget::Variable_Value_Changed);
	connect(d_gradient_temperature_edit,&QLineEdit::editingFinished,this,&TVariableWidget::Variable_Value_Changed);
	connect(d_mean_stdev_edit,&QLineEdit::editingFinished,this,&TVariableWidget::Variable_Value_Changed);
	connect(d_start_position_edit,&QLineEdit::editingFinished,this,&TVariableWidget::Variable_Value_Changed);
	connect(d_end_position_edit,&QLineEdit::editingFinished,this,&TVariableWidget::Variable_Value_Changed);
	connect(d_increment_edit,&QLineEdit::editingFinished,this,&TVariableWidget::Variable_Value_Changed);
	
	connect(d_add_variable_button,&QToolButton::clicked,this,&TVariableWidget::Add_Variable);
	connect(d_rename_variable_button,&QToolButton::clicked,this,&TVariableWidget::Rename_Variable);
	connect(d_remove_variable_button,&QToolButton::clicked,this,&TVariableWidget::Remove_Variable);

}

TVariableWidget::~TVariableWidget(void)
{
}

TInputData TVariableWidget::Input_Data(void) const
{
	TInputData			input_data;
	
	input_data.machine_expansion_coefficient = d_machine_expansion_edit->text().toDouble();
	input_data.machine_scale = d_machine_scale_edit->text().toDouble();
	input_data.temp_min = d_minimum_temperature_edit->text().toDouble();
	input_data.temp_max = d_maximum_temperature_edit->text().toDouble();
	input_data.temp_gradient = d_gradient_temperature_edit->text().toDouble();
	input_data.report_increment = d_increment_edit->text().toDouble();
	input_data.report_end = d_end_position_edit->text().toDouble();
	
	return input_data;
}

bool TVariableWidget::Variable_Exists(
	const QString						&name) const
{
	std::vector<TVariable>::const_iterator var_iter;
	
	for(var_iter = d_variables.begin();var_iter != d_variables.end();++var_iter)
	{
		if(0 == (*var_iter).name.compare(name,Qt::CaseInsensitive))
		{
			return true;
		}
	}
	
	return false;
}

TVariable TVariableWidget::Variable(
	const int							index) const
{
	assert(static_cast<unsigned int>(index) < d_variables.size());
	
	return d_variables[index];
}

QString TVariableWidget::Variable_Name(
	const int							index) const
{
	assert(static_cast<unsigned int>(index) < d_variables.size());
	
	return d_variables[index].name;
}

double TVariableWidget::Variable_Value(
	const int							index,
	bool								*valid) const
{
	TVariable							variable;
	double								dval;
	
	assert(static_cast<unsigned int>(index) < d_variables.size());
	assert(valid);
	
	variable = d_variables[index];
	*valid = false;
	
	if(variable.name == QStringLiteral("CTEm"))
	{
		*valid = true;
		return d_machine_expansion_edit->text().toDouble()/1000;	// mm/m
	}
	else if(variable.name == QStringLiteral("CTEd"))
	{
		*valid = true;
		return 0.0;		// value updated when used
	}
	else if(variable.name == QStringLiteral("CTEc"))
	{
		*valid = true;
		return 0.0;		// value updated when used
	}
	else if(variable.name == QStringLiteral("Sr"))
	{
		*valid = true;
		return d_machine_scale_edit->text().toDouble();
	}
	else if(variable.name == QStringLiteral("Td"))
	{
		dval = d_minimum_temperature_edit->text().toDouble();
		dval += d_maximum_temperature_edit->text().toDouble();
		dval /= 2;
		
		dval = fabs(dval - 20);	// absolute deviation from 20
		
		*valid = true;
		return dval;
	}
	else if(variable.name == QStringLiteral("Tr"))
	{
		dval = d_maximum_temperature_edit->text().toDouble();
		dval -= d_minimum_temperature_edit->text().toDouble();
		
		dval = fabs(dval);
		
		*valid = true;
		return dval;
	}
	else if(variable.name == QStringLiteral("Tg"))
	{
		dval = fabs(d_gradient_temperature_edit->text().toDouble());

		*valid = true;
		return dval;
	}
	else if(variable.name == QStringLiteral("Sdev"))
	{
		*valid = true;
		return fabs(d_mean_stdev_edit->text().toDouble());
	}
	else if(variable.name == QStringLiteral("Lc"))
	{
		*valid = true;
		return 0.0;		// value updated when used
	}
	else if(variable.name == QStringLiteral("L"))
	{
		*valid = true;
		return 0.0;		// value updated when used
	}
	
	return 0.0;
}

QString TVariableWidget::Variable_Description(
	const QString						&variable_name) const
{
	std::vector<TVariable>::const_iterator iter;
	
	for(iter = d_variables.begin();iter != d_variables.end();++iter)
	{
		if((*iter).name == variable_name)
		{
			return (*iter).description;
		}
	}
	
	return QString();
}

double TVariableWidget::Start_Position(void) const
{
	return d_start_position_edit->text().toDouble();
}

double TVariableWidget::End_Position(void) const
{
	return d_end_position_edit->text().toDouble();
}

double TVariableWidget::Increment(void) const
{
	return d_increment_edit->text().toDouble();
}

double TVariableWidget::Machine_Expansion_Coefficient(void) const
{
	return d_machine_expansion_edit->text().toDouble()/1000;
}

void TVariableWidget::Update_Variable_List(
	const QStringList					&list) 
{
	d_variable_list.clear();
	d_variable_list.push_back(QString("[no selection]"));

	d_variable_list += list;
	
	this->Update_Option_Editor();
}

void TVariableWidget::Add_Optional_Variables(
	const std::vector<TVariable>		&variables)
{
	d_variables.resize(d_static_variable_count);
	d_variables.insert(d_variables.end(),variables.begin(),variables.end());
	
	this->Update_Option_Editor();
}

void TVariableWidget::Set_Input_Data(
	const TInputData					&input_data)
{
	d_machine_expansion_edit->setText(QString("%1").arg(input_data.machine_expansion_coefficient,0,'f',2));
	d_machine_scale_edit->setText(QString("%1").arg(input_data.machine_scale,0,'f',6));
	d_minimum_temperature_edit->setText(QString("%1").arg(input_data.temp_min,0,'f',2));
	d_maximum_temperature_edit->setText(QString("%1").arg(input_data.temp_max,0,'f',2));
	d_gradient_temperature_edit->setText(QString("%1").arg(input_data.temp_gradient,0,'f',2));
	d_increment_edit->setText(QString("%1").arg(input_data.report_increment,0,'f',3));
	d_end_position_edit->setText(QString("%1").arg(input_data.report_end,0,'f',3));
}

void TVariableWidget::Variable_Selection_Changed(void)
{
	QList<QTreeWidgetItem*>				sel_items;
	int									sel_row;

	d_rename_variable_button->setEnabled(false);	
	d_remove_variable_button->setEnabled(false);	

	sel_items = d_variable_table->selectedItems();
	
	if(sel_items.size() == 1)
	{
		sel_row = d_variable_table->indexOfTopLevelItem(sel_items[0]);
				
		if(!(sel_row < d_static_variable_count))
		{
			d_rename_variable_button->setEnabled(true);	
			d_remove_variable_button->setEnabled(true);	
		}
	}
}
							  
void TVariableWidget::Variable_Selection_List_Changed(
	const int							index,
	const QString						&text)
{
	std::vector<TVariable>::iterator var_iter;
	
	for(var_iter = d_variables.begin();var_iter != d_variables.end();++var_iter)
	{
		if(index == (*var_iter).editor_index)
		{
			(*var_iter).description = text;
		}
	}
}

void TVariableWidget::Add_Variable(void)
{
	TNameDialog							name_dialog(this);
	TVariable							variable;
	
	name_dialog.Set_Title("New Variable");
	
	if(name_dialog.exec() == QDialog::Accepted)
	{
		variable.name = name_dialog.Name();
		
		variable.name.replace(' ','_');
		
		if(this->Variable_Exists(variable.name))
		{
			return;
		}
		
		variable.editor_index = d_variable_table->AddSelectionList(variable.name,d_variable_list);
		
		d_variables.push_back(variable);
		
		emit Variable_Changed();
	}
}

void TVariableWidget::Rename_Variable(void)
{
	TNameDialog							name_dialog(this);
	QTreeWidgetItem						*item;
	QTreeWidgetItem						*parent_item;
	int									index;
	QString								old_name;
	QString								new_name;
		
	item = d_variable_table->currentItem();
	
	if(item)
	{
		do
		{
			parent_item = item->parent();
			
			if(parent_item)
			{
				item = parent_item;
			}
			
		}while(parent_item);
		
		index = d_variable_table->indexOfTopLevelItem(item);
		
		if(!(index < d_static_variable_count))	// not allowed to change static variables
		{
			old_name = d_variable_table->OptionName(index);
			
			name_dialog.Set_Title("Rename Variable");
			name_dialog.Set_Name(old_name);

			if(name_dialog.exec() == QDialog::Accepted)
			{
				new_name = name_dialog.Name();
				new_name.replace(' ','_');
				
				if(old_name != new_name)
				{
					this->Rename_Variable_Entry(old_name,new_name);
					
					emit Variable_Changed();
				}
			}
		}
	}
}

void TVariableWidget::Remove_Variable(void)
{
	QList<QTreeWidgetItem*>				sel_items;
	QString								name;
	
	
	sel_items = d_variable_table->selectedItems();
	
	if(sel_items.size() == 1)
	{
		name = sel_items[0]->text(0);
		
		this->Remove_Variable_Entry(name);
		
		emit Variable_Changed();
	}
}

void TVariableWidget::Variable_Value_Changed(void)
{
	emit Variable_Changed();
}

void TVariableWidget::Rename_Variable_Entry(
	const QString						&old_name,
	const QString						&new_name)
{
	std::vector<TVariable>::iterator var_iter;

	for(var_iter = d_variables.begin();var_iter != d_variables.end();++var_iter)
	{
		if(0 == (*var_iter).name.compare(old_name,Qt::CaseInsensitive))
		{
			(*var_iter).name = new_name;
			
			this->Update_Option_Editor();
			return;
		}
	}
}

void TVariableWidget::Remove_Variable_Entry(
	const QString						&name)
{
	std::vector<TVariable>::iterator var_iter;
	
	for(var_iter = d_variables.begin();var_iter != d_variables.end();++var_iter)
	{
		if(0 == (*var_iter).name.compare(name,Qt::CaseInsensitive))
		{
			d_variables.erase(var_iter);
			
			this->Update_Option_Editor();
			return;
		}
	}
}

void TVariableWidget::Initialize_Option_Editor(void)
{
	QTreeWidgetItem						*item;
	TVariable							variable;
	std::vector<TVariable>::iterator	var_iter;
	
	d_variable_table->clear();
	d_variables.clear();
	
	d_variable_table->headerItem()->setText(0,QStringLiteral("Variable"));
	d_variable_table->headerItem()->setText(1,QStringLiteral("Description"));
	
	variable.editor_index = -1;
	
	variable.name = QStringLiteral("CTEm");
	variable.description = QStringLiteral("Machine Expansion Coefficient");
	d_variables.push_back(variable);
	
	variable.name = QStringLiteral("CTEc");
	variable.description = QStringLiteral("Component Expansion Coefficient");
	d_variables.push_back(variable);
	
	variable.name = QStringLiteral("CTEd");
	variable.description = QStringLiteral("Expansion Coefficient Difference");
	d_variables.push_back(variable);
	
	variable.name = QStringLiteral("Sr");
	variable.description = QStringLiteral("Scale Resolution");
	d_variables.push_back(variable);
	
	variable.name = QStringLiteral("Td");
	variable.description = QStringLiteral("Temperature Deviation from 20 C");
	d_variables.push_back(variable);
	
	variable.name = QStringLiteral("Tr");
	variable.description = QStringLiteral("Temperature Range");
	d_variables.push_back(variable);
	
	variable.name = QStringLiteral("Tg");
	variable.description = QStringLiteral("Temperature Gradient");
	d_variables.push_back(variable);
	
	variable.name = QStringLiteral("Sdev");
	variable.description = QStringLiteral("Standard Deviation");
	d_variables.push_back(variable);
	
	variable.name = QStringLiteral("Lc");
	variable.description = QStringLiteral("Length Count");
	d_variables.push_back(variable);
	
	variable.name = QStringLiteral("L");
	variable.description = QStringLiteral("Test Length");
	d_variables.push_back(variable);
	
	for(var_iter = d_variables.begin();var_iter != d_variables.end();++var_iter)
	{
		item = new QTreeWidgetItem(d_variable_table);
		item->setText(0,(*var_iter).name);
		item->setText(1,(*var_iter).description);
	}
	
	d_static_variable_count = d_variable_table->topLevelItemCount();
}

void TVariableWidget::Update_Option_Editor(void)
{
	std::vector<TVariable>				var_copy;
	std::vector<TVariable>::iterator	var_iter;
	int									cntr;
	int									index;
	
	var_copy = d_variables;
	
	this->Initialize_Option_Editor();
	
	for(cntr = static_cast<int>(d_variables.size());cntr < static_cast<int>(var_copy.size());++cntr)
	{
		index = d_variable_list.indexOf(var_copy[cntr].description);
		
		if(index < 0)
		{
			index = 0;
		}
				
		var_copy[cntr].editor_index = d_variable_table->AddSelectionList(var_copy[cntr].name,d_variable_list,index);
		
		d_variables.push_back(var_copy[cntr]);
	}
}

